@@ -1,4 +1,4 @@
-# Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+# Copyright (c), ETH Zurich and UNC Chapel Hill.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -28,12 +28,12 @@
 # POSSIBILITY OF SUCH DAMAGE.
 
 
-import re
 import argparse
+import re
+
 import requests
 from lxml.html import soupparser
 
-
 MAX_REQUEST_TRIALS = 10
 
 
@@ -48,7 +48,7 @@ def request_trial(func, *args, **kwargs):
     for i in range(MAX_REQUEST_TRIALS):
         try:
             response = func(*args, **kwargs)
-        except:
+        except:  # noqa E722
             continue
         else:
             return response
