@@ -1,4 +1,4 @@
-# Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+# Copyright (c), ETH Zurich and UNC Chapel Hill.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -30,14 +30,14 @@
 
 # This script is based on an original implementation by True Price.
 
-import sys
 import sqlite3
-import numpy as np
+import sys
 
+import numpy as np
 
 IS_PYTHON3 = sys.version_info[0] >= 3
 
-MAX_IMAGE_ID = 2 ** 31 - 1
+MAX_IMAGE_ID = 2**31 - 1
 
 CREATE_CAMERAS_TABLE = """CREATE TABLE IF NOT EXISTS cameras (
     camera_id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
@@ -60,14 +60,13 @@ CREATE_IMAGES_TABLE = """CREATE TABLE IF NOT EXISTS images (
     camera_id INTEGER NOT NULL,
     CONSTRAINT image_id_check CHECK(image_id >= 0 and image_id < {}),
     FOREIGN KEY(camera_id) REFERENCES cameras(camera_id))
-""".format(
-    MAX_IMAGE_ID
-)
+""".format(MAX_IMAGE_ID)
 
 CREATE_POSE_PRIORS_TABLE = """CREATE TABLE IF NOT EXISTS pose_priors (
     image_id INTEGER PRIMARY KEY NOT NULL,
     position BLOB,
     coordinate_system INTEGER NOT NULL,
+    position_covariance BLOB,
     FOREIGN KEY(image_id) REFERENCES images(image_id) ON DELETE CASCADE)"""
 
 CREATE_TWO_VIEW_GEOMETRIES_TABLE = """
@@ -201,32 +200,27 @@ class COLMAPDatabase(sqlite3.Connection):
         self,
         name,
         camera_id,
-        prior_q=np.full(4, np.NaN),
-        prior_t=np.full(3, np.NaN),
         image_id=None,
     ):
         cursor = self.execute(
-            "INSERT INTO images VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
-            (
-                image_id,
-                name,
-                camera_id,
-                prior_q[0],
-                prior_q[1],
-                prior_q[2],
-                prior_q[3],
-                prior_t[0],
-                prior_t[1],
-                prior_t[2],
-            ),
+            "INSERT INTO images VALUES (?, ?, ?)", (image_id, name, camera_id)
         )
         return cursor.lastrowid
 
-    def add_pose_prior(self, image_id, position, coordinate_system=-1):
+    def add_pose_prior(
+        self, image_id, position, coordinate_system=-1, position_covariance=None
+    ):
         position = np.asarray(position, dtype=np.float64)
+        if position_covariance is None:
+            position_covariance = np.full((3, 3), np.nan, dtype=np.float64)
         self.execute(
-            "INSERT INTO pose_priors VALUES (?, ?, ?)",
-            (image_id, array_to_blob(position), coordinate_system),
+            "INSERT INTO pose_priors VALUES (?, ?, ?, ?)",
+            (
+                image_id,
+                array_to_blob(position),
+                coordinate_system,
+                array_to_blob(position_covariance),
+            ),
         )
 
     def add_keypoints(self, image_id, keypoints):
@@ -302,8 +296,8 @@ class COLMAPDatabase(sqlite3.Connection):
 
 
 def example_usage():
-    import os
     import argparse
+    import os
 
     parser = argparse.ArgumentParser()
     parser.add_argument("--database_path", default="database.db")
@@ -375,6 +369,26 @@ def example_usage():
     db.add_matches(image_id2, image_id3, matches23)
     db.add_matches(image_id3, image_id4, matches34)
 
+    # Create dummy pose_priors.
+
+    pos1 = np.random.rand(3, 1) * np.random.randint(10)
+    pos2 = np.random.rand(3, 1) * np.random.randint(10)
+    pos3 = np.random.rand(3, 1) * np.random.randint(10)
+
+    cov3 = np.random.rand(3, 3) * np.random.randint(10)
+
+    pose_prior1 = [image_id1, pos1, 1, None]
+    pose_prior2 = [image_id2, pos2, -1, None]
+    pose_prior3 = [image_id3, pos3, 0, cov3]
+
+    db.add_pose_prior(*pose_prior1)
+    db.add_pose_prior(*pose_prior2)
+    db.add_pose_prior(*pose_prior3)
+
+    # Convert unset covariance to nan matrix for later check
+    pose_prior1[3] = np.full((3, 3), np.nan, dtype=np.float64)
+    pose_prior2[3] = np.full((3, 3), np.nan, dtype=np.float64)
+
     # Commit the data to the file.
 
     db.commit()
@@ -409,15 +423,6 @@ def example_usage():
 
     # Read and check matches.
 
-    pair_ids = [
-        image_ids_to_pair_id(*pair)
-        for pair in (
-            (image_id1, image_id2),
-            (image_id2, image_id3),
-            (image_id3, image_id4),
-        )
-    ]
-
     matches = dict(
         (pair_id_to_image_ids(pair_id), blob_to_array(data, np.uint32, (-1, 2)))
         for pair_id, data in db.execute("SELECT pair_id, data FROM matches")
@@ -427,6 +432,30 @@ def example_usage():
     assert np.all(matches[(image_id2, image_id3)] == matches23)
     assert np.all(matches[(image_id3, image_id4)] == matches34)
 
+    # Read and check pose_priors
+
+    rows = db.execute("SELECT * FROM pose_priors")
+
+    img_id1, pos1, coord_sys1, cov1 = next(rows)
+    img_id2, pos2, coord_sys2, cov2 = next(rows)
+    img_id3, pos3, coord_sys3, cov3 = next(rows)
+
+    assert pose_prior1[0] == img_id1
+    assert pose_prior2[0] == img_id2
+    assert pose_prior3[0] == img_id3
+
+    assert pose_prior1[1].all() == blob_to_array(pos1, np.float64, (3, 1)).all()
+    assert pose_prior2[1].all() == blob_to_array(pos2, np.float64, (3, 1)).all()
+    assert pose_prior3[1].all() == blob_to_array(pos3, np.float64, (3, 1)).all()
+
+    assert pose_prior1[2] == coord_sys1
+    assert pose_prior2[2] == coord_sys2
+    assert pose_prior3[2] == coord_sys3
+
+    assert pose_prior1[3].all() == blob_to_array(cov1, np.float64, (3, 3)).all()
+    assert pose_prior2[3].all() == blob_to_array(cov2, np.float64, (3, 3)).all()
+    assert pose_prior3[3].all() == blob_to_array(cov3, np.float64, (3, 3)).all()
+
     # Clean up.
 
     db.close()
