@@ -17,10 +17,9 @@
 #       <http://www.gnu.org/licenses/>.
 
 from itertools import islice as _islice
-
-import numpy as _np
 from sys import byteorder as _byteorder
 
+import numpy as _np
 
 try:
     _range = xrange
@@ -54,7 +53,7 @@ _data_type_reverse = dict((b, a) for (a, b) in _data_type_relation)
 
 _types_list = []
 _types_set = set()
-for (_a, _b) in _data_type_relation:
+for _a, _b in _data_type_relation:
     if _a not in _types_set:
         _types_list.append(_a)
         _types_set.add(_a)
@@ -103,9 +102,7 @@ def make2d(array, cols=None, dtype=None):
 
     """
     if (cols is None or dtype is None) and not len(array):
-        raise RuntimeError(
-            "cols and dtype must be specified for empty " "array"
-        )
+        raise RuntimeError("cols and dtype must be specified for empty array")
 
     if cols is None:
         cols = len(array[0])
@@ -117,7 +114,6 @@ def make2d(array, cols=None, dtype=None):
 
 
 class PlyParseError(Exception):
-
     """
     Raised when a PLY file cannot be parsed.
 
@@ -153,7 +149,6 @@ class PlyParseError(Exception):
 
 
 class PlyData(object):
-
     """
     PLY file header and data.
 
@@ -378,7 +373,6 @@ class PlyData(object):
 
 
 class PlyElement(object):
-
     """
     PLY file element.
 
@@ -536,7 +530,7 @@ class PlyElement(object):
             raise TypeError("only numpy arrays are supported")
 
         if len(data.shape) != 1:
-            raise ValueError("only one-dimensional arrays are " "supported")
+            raise ValueError("only one-dimensional arrays are supported")
 
         count = len(data)
 
@@ -557,7 +551,7 @@ class PlyElement(object):
                 if t[1][1] == "O":
                     if len(t) != 2:
                         raise ValueError(
-                            "non-scalar object fields not " "supported"
+                            "non-scalar object fields not supported"
                         )
 
                 len_str = _data_type_reverse[len_types.get(t[0], "u1")]
@@ -731,7 +725,6 @@ class PlyElement(object):
 
 
 class PlyProperty(object):
-
     """
     PLY property description.  This class is pure metadata; the data
     itself is contained in PlyElement instances.
@@ -766,17 +759,17 @@ class PlyProperty(object):
 
         if line[1] == "list":
             if len(line) > 5:
-                raise PlyParseError("too many fields after " "'property list'")
+                raise PlyParseError("too many fields after 'property list'")
             if len(line) < 5:
-                raise PlyParseError("too few fields after " "'property list'")
+                raise PlyParseError("too few fields after 'property list'")
 
             return PlyListProperty(line[4], line[2], line[3])
 
         else:
             if len(line) > 3:
-                raise PlyParseError("too many fields after " "'property'")
+                raise PlyParseError("too many fields after 'property'")
             if len(line) < 3:
-                raise PlyParseError("too few fields after " "'property'")
+                raise PlyParseError("too few fields after 'property'")
 
             return PlyProperty(line[2], line[1])
 
@@ -830,7 +823,6 @@ class PlyProperty(object):
 
 
 class PlyListProperty(PlyProperty):
-
     """
     PLY list property description.
 
