@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -73,10 +73,8 @@ void BundleAdjustmentController::Run() {
   Timer run_timer;
   run_timer.Start();
 
-  const std::vector<image_t>& reg_image_ids = reconstruction_->RegImageIds();
-
-  if (reg_image_ids.size() < 2) {
-    LOG(ERROR) << "Need at least two views.";
+  if (reconstruction_->NumRegFrames() == 0) {
+    LOG(ERROR) << "Need at least one registered frame.";
     return;
   }
 
@@ -90,15 +88,19 @@ void BundleAdjustmentController::Run() {
 
   // Configure bundle adjustment.
   BundleAdjustmentConfig ba_config;
-  for (const image_t image_id : reg_image_ids) {
+  for (const image_t image_id : reconstruction_->RegImageIds()) {
     ba_config.AddImage(image_id);
   }
-  ba_config.SetConstantCamPose(reg_image_ids[0]);
-  ba_config.SetConstantCamPositions(reg_image_ids[1], {0});
+  // Fixing the gauge with two cameras leads to a more stable optimization
+  // with fewer steps as compared to fixing three points.
+  // TODO(jsch): Investigate whether it is safe to not fix the gauge at all,
+  // as initial experiments show that it is even faster.
+  ba_config.FixGauge(BundleAdjustmentGauge::TWO_CAMS_FROM_WORLD);
 
   // Run bundle adjustment.
-  BundleAdjuster bundle_adjuster(ba_options, ba_config);
-  bundle_adjuster.Solve(reconstruction_.get());
+  std::unique_ptr<BundleAdjuster> bundle_adjuster = CreateDefaultBundleAdjuster(
+      std::move(ba_options), std::move(ba_config), *reconstruction_);
+  bundle_adjuster->Solve();
 
   run_timer.PrintMinutes();
 }
