@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -43,7 +43,7 @@ namespace colmap {
 FeatureMatcherWorker::FeatureMatcherWorker(
     const SiftMatchingOptions& matching_options,
     const TwoViewGeometryOptions& geometry_options,
-    FeatureMatcherCache* cache,
+    const std::shared_ptr<FeatureMatcherCache>& cache,
     JobQueue<Input>* input_queue,
     JobQueue<Output>* output_queue)
     : matching_options_(matching_options),
@@ -53,11 +53,6 @@ FeatureMatcherWorker::FeatureMatcherWorker(
       output_queue_(output_queue) {
   THROW_CHECK(matching_options_.Check());
 
-  prev_keypoints_image_ids_[0] = kInvalidImageId;
-  prev_keypoints_image_ids_[1] = kInvalidImageId;
-  prev_descriptors_image_ids_[0] = kInvalidImageId;
-  prev_descriptors_image_ids_[1] = kInvalidImageId;
-
   if (matching_options_.use_gpu) {
 #if !defined(COLMAP_CUDA_ENABLED)
     opengl_context_ = std::make_unique<OpenGLContextManager>();
@@ -77,6 +72,15 @@ void FeatureMatcherWorker::Run() {
 #endif
   }
 
+  matching_options_.cpu_descriptor_index_cache =
+      &cache_->GetFeatureDescriptorIndexCache();
+  THROW_CHECK_NOTNULL(matching_options_.cpu_descriptor_index_cache);
+
+  // Minimize the amount of allocated GPU memory by computing the maximum number
+  // of descriptors for any image over the whole database.
+  matching_options_.max_num_matches = std::min<int>(
+      matching_options_.max_num_matches, cache_->MaxNumKeypoints());
+
   std::unique_ptr<FeatureMatcher> matcher =
       CreateSiftFeatureMatcher(matching_options_);
   if (matcher == nullptr) {
@@ -104,15 +108,28 @@ void FeatureMatcherWorker::Run() {
 
       if (matching_options_.guided_matching) {
         matcher->MatchGuided(geometry_options_.ransac_options.max_error,
-                             GetKeypointsPtr(0, data.image_id1),
-                             GetKeypointsPtr(1, data.image_id2),
-                             GetDescriptorsPtr(0, data.image_id1),
-                             GetDescriptorsPtr(1, data.image_id2),
+                             {
+                                 data.image_id1,
+                                 cache_->GetDescriptors(data.image_id1),
+                                 cache_->GetKeypoints(data.image_id1),
+                             },
+                             {
+                                 data.image_id2,
+                                 cache_->GetDescriptors(data.image_id2),
+                                 cache_->GetKeypoints(data.image_id2),
+                             },
                              &data.two_view_geometry);
       } else {
-        matcher->Match(GetDescriptorsPtr(0, data.image_id1),
-                       GetDescriptorsPtr(1, data.image_id2),
-                       &data.matches);
+        matcher->Match(
+            {
+                data.image_id1,
+                cache_->GetDescriptors(data.image_id1),
+            },
+            {
+                data.image_id2,
+                cache_->GetDescriptors(data.image_id2),
+            },
+            &data.matches);
       }
 
       THROW_CHECK(output_queue_->Push(std::move(data)));
@@ -120,32 +137,6 @@ void FeatureMatcherWorker::Run() {
   }
 }
 
-std::shared_ptr<FeatureKeypoints> FeatureMatcherWorker::GetKeypointsPtr(
-    const int index, const image_t image_id) {
-  THROW_CHECK_GE(index, 0);
-  THROW_CHECK_LE(index, 1);
-  if (prev_keypoints_image_ids_[index] == image_id) {
-    return nullptr;
-  } else {
-    prev_keypoints_image_ids_[index] = image_id;
-    prev_keypoints_[index] = cache_->GetKeypoints(image_id);
-    return prev_keypoints_[index];
-  }
-}
-
-std::shared_ptr<FeatureDescriptors> FeatureMatcherWorker::GetDescriptorsPtr(
-    const int index, const image_t image_id) {
-  THROW_CHECK_GE(index, 0);
-  THROW_CHECK_LE(index, 1);
-  if (prev_descriptors_image_ids_[index] == image_id) {
-    return nullptr;
-  } else {
-    prev_descriptors_image_ids_[index] = image_id;
-    prev_descriptors_[index] = cache_->GetDescriptors(image_id);
-    return prev_descriptors_[index];
-  }
-}
-
 namespace {
 
 class VerifierWorker : public Thread {
@@ -154,11 +145,11 @@ class VerifierWorker : public Thread {
   typedef FeatureMatcherData Output;
 
   VerifierWorker(const TwoViewGeometryOptions& options,
-                 FeatureMatcherCache* cache,
+                 std::shared_ptr<FeatureMatcherCache> cache,
                  JobQueue<Input>* input_queue,
                  JobQueue<Output>* output_queue)
       : options_(options),
-        cache_(cache),
+        cache_(std::move(cache)),
         input_queue_(input_queue),
         output_queue_(output_queue) {
     THROW_CHECK(options_.Check());
@@ -202,7 +193,7 @@ class VerifierWorker : public Thread {
 
  private:
   const TwoViewGeometryOptions options_;
-  FeatureMatcherCache* cache_;
+  std::shared_ptr<FeatureMatcherCache> cache_;
   JobQueue<Input>* input_queue_;
   JobQueue<Output>* output_queue_;
 };
@@ -212,12 +203,10 @@ class VerifierWorker : public Thread {
 FeatureMatcherController::FeatureMatcherController(
     const SiftMatchingOptions& matching_options,
     const TwoViewGeometryOptions& geometry_options,
-    Database* database,
-    FeatureMatcherCache* cache)
+    std::shared_ptr<FeatureMatcherCache> cache)
     : matching_options_(matching_options),
       geometry_options_(geometry_options),
-      database_(database),
-      cache_(cache),
+      cache_(std::move(cache)),
       is_setup_(false) {
   THROW_CHECK(matching_options_.Check());
   THROW_CHECK(geometry_options_.Check());
@@ -248,7 +237,7 @@ FeatureMatcherController::FeatureMatcherController(
       matchers_.emplace_back(
           std::make_unique<FeatureMatcherWorker>(matching_options_copy,
                                                  geometry_options_,
-                                                 cache,
+                                                 cache_,
                                                  &matcher_queue_,
                                                  &verifier_queue_));
     }
@@ -261,7 +250,7 @@ FeatureMatcherController::FeatureMatcherController(
       matchers_.emplace_back(
           std::make_unique<FeatureMatcherWorker>(matching_options_copy,
                                                  geometry_options_,
-                                                 cache,
+                                                 cache_,
                                                  &matcher_queue_,
                                                  &verifier_queue_));
     }
@@ -272,7 +261,7 @@ FeatureMatcherController::FeatureMatcherController(
     // Redirect the verification output to final round of guided matching.
     for (int i = 0; i < num_threads; ++i) {
       verifiers_.emplace_back(std::make_unique<VerifierWorker>(
-          geometry_options_, cache, &verifier_queue_, &guided_matcher_queue_));
+          geometry_options_, cache_, &verifier_queue_, &guided_matcher_queue_));
     }
 
     if (matching_options_.use_gpu) {
@@ -283,7 +272,7 @@ FeatureMatcherController::FeatureMatcherController(
         guided_matchers_.emplace_back(
             std::make_unique<FeatureMatcherWorker>(matching_options_copy,
                                                    geometry_options_,
-                                                   cache,
+                                                   cache_,
                                                    &guided_matcher_queue_,
                                                    &output_queue_));
       }
@@ -293,7 +282,7 @@ FeatureMatcherController::FeatureMatcherController(
         guided_matchers_.emplace_back(
             std::make_unique<FeatureMatcherWorker>(matching_options_,
                                                    geometry_options_,
-                                                   cache,
+                                                   cache_,
                                                    &guided_matcher_queue_,
                                                    &output_queue_));
       }
@@ -301,7 +290,7 @@ FeatureMatcherController::FeatureMatcherController(
   } else {
     for (int i = 0; i < num_threads; ++i) {
       verifiers_.emplace_back(std::make_unique<VerifierWorker>(
-          geometry_options_, cache, &verifier_queue_, &output_queue_));
+          geometry_options_, cache_, &verifier_queue_, &output_queue_));
     }
   }
 }
@@ -345,8 +334,7 @@ FeatureMatcherController::~FeatureMatcherController() {
 bool FeatureMatcherController::Setup() {
   // Minimize the amount of allocated GPU memory by computing the maximum number
   // of descriptors for any image over the whole database.
-  const int max_num_features =
-      THROW_CHECK_NOTNULL(database_)->MaxNumKeypoints();
+  const int max_num_features = THROW_CHECK_NOTNULL(cache_)->MaxNumKeypoints();
   matching_options_.max_num_matches =
       std::min(matching_options_.max_num_matches, max_num_features);
 
@@ -383,7 +371,6 @@ bool FeatureMatcherController::Setup() {
 
 void FeatureMatcherController::Match(
     const std::vector<std::pair<image_t, image_t>>& image_pairs) {
-  THROW_CHECK_NOTNULL(database_);
   THROW_CHECK_NOTNULL(cache_);
   THROW_CHECK(is_setup_);
 
