@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -30,6 +30,7 @@
 #include "colmap/controllers/image_reader.h"
 
 #include "colmap/sensor/models.h"
+#include "colmap/util/file.h"
 #include "colmap/util/misc.h"
 
 namespace colmap {
@@ -52,19 +53,21 @@ ImageReader::ImageReader(const ImageReaderOptions& options, Database* database)
   // Ensure trailing slash, so that we can build the correct image name.
   options_.image_path =
       EnsureTrailingSlash(StringReplace(options_.image_path, "\\", "/"));
-  options_.mask_path =
-      EnsureTrailingSlash(StringReplace(options_.mask_path, "\\", "/"));
+  if (!options_.mask_path.empty()) {
+    options_.mask_path =
+        EnsureTrailingSlash(StringReplace(options_.mask_path, "\\", "/"));
+  }
 
   // Get a list of all files in the image path, sorted by image name.
-  if (options_.image_list.empty()) {
-    options_.image_list = GetRecursiveFileList(options_.image_path);
-    std::sort(options_.image_list.begin(), options_.image_list.end());
+  if (options_.image_names.empty()) {
+    options_.image_names = GetRecursiveFileList(options_.image_path);
+    std::sort(options_.image_names.begin(), options_.image_names.end());
   } else {
-    if (!std::is_sorted(options_.image_list.begin(),
-                        options_.image_list.end())) {
-      std::sort(options_.image_list.begin(), options_.image_list.end());
+    if (!std::is_sorted(options_.image_names.begin(),
+                        options_.image_names.end())) {
+      std::sort(options_.image_names.begin(), options_.image_names.end());
     }
-    for (auto& image_name : options_.image_list) {
+    for (auto& image_name : options_.image_names) {
       image_name = JoinPaths(options_.image_path, image_name);
     }
   }
@@ -72,6 +75,15 @@ ImageReader::ImageReader(const ImageReaderOptions& options, Database* database)
   if (static_cast<camera_t>(options_.existing_camera_id) != kInvalidCameraId) {
     THROW_CHECK(database->ExistsCamera(options_.existing_camera_id));
     prev_camera_ = database->ReadCamera(options_.existing_camera_id);
+    if (std::optional<Rig> rig =
+            database->ReadRigWithSensor(prev_camera_.SensorId());
+        rig.has_value()) {
+      prev_rig_ = std::move(*rig);
+    } else {
+      // For backwards compatibility with old databases without rigs.
+      prev_rig_.AddRefSensor(prev_camera_.SensorId());
+      prev_rig_.SetRigId(database_->WriteRig(prev_rig_));
+    }
   } else {
     // Set the manually specified camera parameters.
     prev_camera_.camera_id = kInvalidCameraId;
@@ -85,7 +97,8 @@ ImageReader::ImageReader(const ImageReaderOptions& options, Database* database)
   }
 }
 
-ImageReader::Status ImageReader::Next(Camera* camera,
+ImageReader::Status ImageReader::Next(Rig* rig,
+                                      Camera* camera,
                                       Image* image,
                                       PosePrior* pose_prior,
                                       Bitmap* bitmap,
@@ -95,9 +108,9 @@ ImageReader::Status ImageReader::Next(Camera* camera,
   THROW_CHECK_NOTNULL(bitmap);
 
   image_index_ += 1;
-  THROW_CHECK_LE(image_index_, options_.image_list.size());
+  THROW_CHECK_LE(image_index_, options_.image_names.size());
 
-  const std::string image_path = options_.image_list.at(image_index_ - 1);
+  const std::string image_path = options_.image_names.at(image_index_ - 1);
 
   DatabaseTransaction database_transaction(database_);
 
@@ -120,7 +133,7 @@ ImageReader::Status ImageReader::Next(Camera* camera,
   const bool exists_image = database_->ExistsImageWithName(image->Name());
 
   if (exists_image) {
-    *image = database_->ReadImageWithName(image->Name());
+    *image = database_->ReadImageWithName(image->Name()).value();
     const bool exists_keypoints = database_->ExistsKeypoints(image->ImageId());
     const bool exists_descriptors =
         database_->ExistsDescriptors(image->ImageId());
@@ -143,11 +156,26 @@ ImageReader::Status ImageReader::Next(Camera* camera,
   //////////////////////////////////////////////////////////////////////////////
 
   if (mask && !options_.mask_path.empty()) {
-    const std::string mask_path =
+    std::string mask_path =
         JoinPaths(options_.mask_path, image->Name() + ".png");
-    if (ExistsFile(mask_path) && !mask->Read(mask_path, false)) {
-      // NOTE: Maybe introduce a separate error type MASK_ERROR?
-      return Status::BITMAP_ERROR;
+    if (!ExistsFile(mask_path)) {
+      bool exists_mask = false;
+      if (HasFileExtension(image->Name(), ".png")) {
+        std::string alt_mask_path =
+            JoinPaths(options_.mask_path, image->Name());
+        if (ExistsFile(alt_mask_path)) {
+          mask_path = std::move(alt_mask_path);
+          exists_mask = true;
+        }
+      }
+      if (!exists_mask) {
+        LOG(ERROR) << "Mask at " << mask_path << " does not exist.";
+        return Status::MASK_ERROR;
+      }
+    }
+    if (!mask->Read(mask_path, false)) {
+      LOG(ERROR) << "Failed to read invalid mask file at: " << mask_path;
+      return Status::MASK_ERROR;
     }
   }
 
@@ -170,6 +198,16 @@ ImageReader::Status ImageReader::Next(Camera* camera,
     }
 
     prev_camera_ = std::move(current_camera);
+    if (std::optional<Rig> rig =
+            database_->ReadRigWithSensor(prev_camera_.SensorId());
+        rig.has_value()) {
+      prev_rig_ = std::move(rig.value());
+    } else {
+      // For backwards compatibility with old databases, we create a rig.
+      prev_rig_ = Rig();
+      prev_rig_.AddRefSensor(prev_camera_.SensorId());
+      prev_rig_.SetRigId(database_->WriteRig(prev_rig_));
+    }
 
   } else {
     //////////////////////////////////////////////////////////////////////////////
@@ -188,6 +226,7 @@ ImageReader::Status ImageReader::Next(Camera* camera,
     //////////////////////////////////////////////////////////////////////////////
     // Read camera model and check for consistency if it exists
     //////////////////////////////////////////////////////////////////////////////
+
     std::string camera_model;
     const bool valid_camera_model = bitmap->ExifCameraModel(&camera_model);
     if (camera_model_to_id_.count(camera_model) > 0) {
@@ -198,6 +237,16 @@ ImageReader::Status ImageReader::Next(Camera* camera,
         return Status::CAMERA_EXIST_DIM_ERROR;
       }
       prev_camera_ = std::move(camera);
+      if (std::optional<Rig> rig =
+              database_->ReadRigWithSensor(prev_camera_.SensorId());
+          rig.has_value()) {
+        prev_rig_ = std::move(rig.value());
+      } else {
+        // For backwards compatibility with old databases, we create a rig.
+        prev_rig_ = Rig();
+        prev_rig_.AddRefSensor(prev_camera_.SensorId());
+        prev_rig_.SetRigId(database_->WriteRig(prev_rig_));
+      }
     }
 
     //////////////////////////////////////////////////////////////////////////////
@@ -239,6 +288,16 @@ ImageReader::Status ImageReader::Next(Camera* camera,
       }
 
       prev_camera_.camera_id = database_->WriteCamera(prev_camera_);
+
+      // By default we create a separate rig per camera. Grouping of different
+      // cameras into the same rig is expected to be done with the
+      // "rig_configurator" after feature extraction.
+      if (!database_->ReadRigWithSensor(prev_camera_.SensorId()).has_value()) {
+        prev_rig_ = Rig();
+        prev_rig_.AddRefSensor(prev_camera_.SensorId());
+        prev_rig_.SetRigId(database_->WriteRig(prev_rig_));
+      }
+
       if (valid_camera_model) {
         camera_model_to_id_[camera_model] = prev_camera_.camera_id;
       }
@@ -260,6 +319,7 @@ ImageReader::Status ImageReader::Next(Camera* camera,
   }
 
   *camera = prev_camera_;
+  *rig = prev_rig_;
 
   image_folders_.insert(image_folder);
   prev_image_folder_ = image_folder;
@@ -269,6 +329,31 @@ ImageReader::Status ImageReader::Next(Camera* camera,
 
 size_t ImageReader::NextIndex() const { return image_index_; }
 
-size_t ImageReader::NumImages() const { return options_.image_list.size(); }
+size_t ImageReader::NumImages() const { return options_.image_names.size(); }
+
+std::string ImageReader::StatusToString(const ImageReader::Status status) {
+  switch (status) {
+    case ImageReader::Status::SUCCESS:
+      return "SUCCESS";
+    case ImageReader::Status::FAILURE:
+      return "FAILURE: Failed to process the image.";
+    case ImageReader::Status::IMAGE_EXISTS:
+      return "IMAGE_EXISTS: Features for image were already extracted.";
+    case ImageReader::Status::BITMAP_ERROR:
+      return "BITMAP_ERROR: Failed to read the image file format.";
+    case ImageReader::Status::MASK_ERROR:
+      return "MASK_ERROR: Failed to read the mask file.";
+    case ImageReader::Status::CAMERA_SINGLE_DIM_ERROR:
+      return "CAMERA_SINGLE_DIM_ERROR: Single camera specified, but images "
+             "have different dimensions.";
+    case ImageReader::Status::CAMERA_EXIST_DIM_ERROR:
+      return "CAMERA_EXIST_DIM_ERROR: Image previously processed, but current "
+             "image has different dimensions.";
+    case ImageReader::Status::CAMERA_PARAM_ERROR:
+      return "CAMERA_PARAM_ERROR: Camera has invalid parameters.";
+    default:
+      return "Unknown";
+  }
+}
 
 }  // namespace colmap
