@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -34,16 +34,15 @@
 #include "colmap/sensor/bitmap.h"
 #include "colmap/util/threading.h"
 
+#include <unordered_map>
 #include <unordered_set>
+#include <vector>
 
 namespace colmap {
 
 struct ImageReaderOptions {
-  // Path to database in which to store the extracted data.
-  std::string database_path = "";
-
   // Root path to folder which contains the images.
-  std::string image_path = "";
+  std::string image_path;
 
   // Optional root path to folder which contains image masks. For a given image,
   // the corresponding mask must have the same sub-path below this root as the
@@ -52,42 +51,39 @@ struct ImageReaderOptions {
   // mask would be mask_path/abc/012.jpg.png. No features will be extracted in
   // regions where the mask image is black (pixel intensity value 0 in
   // grayscale).
-  std::string mask_path = "";
+  std::string mask_path;
+
+  // Optional path to an image file specifying a mask for all images. No
+  // features will be extracted in regions where the mask is black (pixel
+  // intensity value 0 in grayscale).
+  std::string camera_mask_path;
 
   // Optional list of images to read. The list must contain the relative path
   // of the images with respect to the image_path.
-  std::vector<std::string> image_list;
-
+  std::vector<std::string> image_names;
   // Name of the camera model.
   std::string camera_model = "SIMPLE_RADIAL";
 
+  // Manual specification of camera parameters. If empty, camera parameters
+  // will be extracted from EXIF, i.e. principal point and focal length.
+  std::string camera_params;
+
   // Whether to use the same camera for all images.
   bool single_camera = false;
 
   // Whether to use the same camera for all images in the same sub-folder.
   bool single_camera_per_folder = false;
-
   // Whether to use a different camera for each image.
   bool single_camera_per_image = false;
-
   // Whether to explicitly use an existing camera for all images. Note that in
   // this case the specified camera model and parameters are ignored.
   int existing_camera_id = kInvalidCameraId;
 
-  // Manual specification of camera parameters. If empty, camera parameters
-  // will be extracted from EXIF, i.e. principal point and focal length.
-  std::string camera_params = "";
-
   // If camera parameters are not specified manually and the image does not
   // have focal length EXIF information, the focal length is set to the
   // value `default_focal_length_factor * max(width, height)`.
   double default_focal_length_factor = 1.2;
 
-  // Optional path to an image file specifying a mask for all images. No
-  // features will be extracted in regions where the mask is black (pixel
-  // intensity value 0 in grayscale).
-  std::string camera_mask_path = "";
-
   bool Check() const;
 };
 
@@ -101,6 +97,7 @@ class ImageReader {
     SUCCESS,
     IMAGE_EXISTS,
     BITMAP_ERROR,
+    MASK_ERROR,
     CAMERA_SINGLE_DIM_ERROR,
     CAMERA_EXIST_DIM_ERROR,
     CAMERA_PARAM_ERROR
@@ -108,7 +105,8 @@ class ImageReader {
 
   ImageReader(const ImageReaderOptions& options, Database* database);
 
-  Status Next(Camera* camera,
+  Status Next(Rig* rig,
+              Camera* camera,
               Image* image,
               PosePrior* pose_prior,
               Bitmap* bitmap,
@@ -116,13 +114,16 @@ class ImageReader {
   size_t NextIndex() const;
   size_t NumImages() const;
 
+  static std::string StatusToString(Status status);
+
  private:
   // Image reader options.
   ImageReaderOptions options_;
   Database* database_;
   // Index of previously processed image.
   size_t image_index_;
-  // Previously processed camera.
+  // Previously processed rig/camera.
+  Rig prev_rig_;
   Camera prev_camera_;
   std::unordered_map<std::string, camera_t> camera_model_to_id_;
   // Names of image sub-folders.
