@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -31,7 +31,9 @@
 
 #include "colmap/geometry/pose.h"
 #include "colmap/geometry/rigid3.h"
+#include "colmap/geometry/sim3.h"
 #include "colmap/math/math.h"
+#include "colmap/math/random.h"
 #include "colmap/sensor/models.h"
 
 #include <gtest/gtest.h>
@@ -39,10 +41,10 @@
 namespace colmap {
 namespace {
 
-TEST(BundleAdjustment, AbsolutePose) {
-  using CostFunction = ReprojErrorCostFunction<SimplePinholeCameraModel>;
+TEST(ReprojErrorCostFunctor, Nominal) {
+  using CostFunctor = ReprojErrorCostFunctor<SimplePinholeCameraModel>;
   std::unique_ptr<ceres::CostFunction> cost_function(
-      CostFunction::Create(Eigen::Vector2d::Zero()));
+      CostFunctor::Create(Eigen::Vector2d::Zero()));
   double cam_from_world_rotation[4] = {0, 0, 0, 1};
   double cam_from_world_translation[3] = {0, 0, 0};
   double point3D[3] = {0, 0, 1};
@@ -70,21 +72,13 @@ TEST(BundleAdjustment, AbsolutePose) {
   EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
   EXPECT_EQ(residuals[0], -2);
   EXPECT_EQ(residuals[1], 2);
-
-  std::unique_ptr<ceres::CostFunction> cost_function_with_noise(
-      IsotropicNoiseCostFunctionWrapper<CostFunction>::Create(
-          2.0, Eigen::Vector2d::Zero()));
-  EXPECT_TRUE(
-      cost_function_with_noise->Evaluate(parameters, residuals, nullptr));
-  EXPECT_EQ(residuals[0], -1);
-  EXPECT_EQ(residuals[1], 1);
 }
 
-TEST(BundleAdjustment, ConstantPoseAbsolutePose) {
+TEST(ReprojErrorConstantPoseCostFunctor, Nominal) {
   Rigid3d cam_from_world;
   std::unique_ptr<ceres::CostFunction> cost_function(
-      ReprojErrorConstantPoseCostFunction<SimplePinholeCameraModel>::Create(
-          cam_from_world, Eigen::Vector2d::Zero()));
+      ReprojErrorConstantPoseCostFunctor<SimplePinholeCameraModel>::Create(
+          Eigen::Vector2d::Zero(), cam_from_world));
   double point3D[3] = {0, 0, 1};
   double camera_params[3] = {1, 0, 0};
   double residuals[2];
@@ -109,7 +103,7 @@ TEST(BundleAdjustment, ConstantPoseAbsolutePose) {
   EXPECT_EQ(residuals[1], 2);
 }
 
-TEST(BundleAdjustment, ConstantPoint3DAbsolutePose) {
+TEST(ReprojErrorConstantPoint3DCostFunctor, Nominal) {
   Eigen::Vector2d point2D = Eigen::Vector2d::Zero();
   Eigen::Vector3d point3D;
   point3D << 0, 0, 1;
@@ -123,8 +117,8 @@ TEST(BundleAdjustment, ConstantPoint3DAbsolutePose) {
 
   {
     std::unique_ptr<ceres::CostFunction> cost_function(
-        ReprojErrorConstantPoint3DCostFunction<
-            SimplePinholeCameraModel>::Create(point2D, point3D));
+        ReprojErrorConstantPoint3DCostFunctor<SimplePinholeCameraModel>::Create(
+            point2D, point3D));
     EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
     EXPECT_EQ(residuals[0], 0);
     EXPECT_EQ(residuals[1], 0);
@@ -133,8 +127,8 @@ TEST(BundleAdjustment, ConstantPoint3DAbsolutePose) {
   {
     point3D[1] = 1;
     std::unique_ptr<ceres::CostFunction> cost_function(
-        ReprojErrorConstantPoint3DCostFunction<
-            SimplePinholeCameraModel>::Create(point2D, point3D));
+        ReprojErrorConstantPoint3DCostFunctor<SimplePinholeCameraModel>::Create(
+            point2D, point3D));
     EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
     EXPECT_EQ(residuals[0], 0);
     EXPECT_EQ(residuals[1], 1);
@@ -143,8 +137,8 @@ TEST(BundleAdjustment, ConstantPoint3DAbsolutePose) {
   {
     camera_params[0] = 2;
     std::unique_ptr<ceres::CostFunction> cost_function(
-        ReprojErrorConstantPoint3DCostFunction<
-            SimplePinholeCameraModel>::Create(point2D, point3D));
+        ReprojErrorConstantPoint3DCostFunctor<SimplePinholeCameraModel>::Create(
+            point2D, point3D));
     EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
     EXPECT_EQ(residuals[0], 0);
     EXPECT_EQ(residuals[1], 2);
@@ -153,17 +147,17 @@ TEST(BundleAdjustment, ConstantPoint3DAbsolutePose) {
   {
     point3D[0] = -1;
     std::unique_ptr<ceres::CostFunction> cost_function(
-        ReprojErrorConstantPoint3DCostFunction<
-            SimplePinholeCameraModel>::Create(point2D, point3D));
+        ReprojErrorConstantPoint3DCostFunctor<SimplePinholeCameraModel>::Create(
+            point2D, point3D));
     EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
     EXPECT_EQ(residuals[0], -2);
     EXPECT_EQ(residuals[1], 2);
   }
 }
 
-TEST(BundleAdjustment, Rig) {
+TEST(RigReprojErrorCostFunctor, Nominal) {
   std::unique_ptr<ceres::CostFunction> cost_function(
-      RigReprojErrorCostFunction<SimplePinholeCameraModel>::Create(
+      RigReprojErrorCostFunctor<SimplePinholeCameraModel>::Create(
           Eigen::Vector2d::Zero()));
   double cam_from_rig_rotation[4] = {0, 0, 0, 1};
   double cam_from_rig_translation[3] = {0, 0, -1};
@@ -198,12 +192,12 @@ TEST(BundleAdjustment, Rig) {
   EXPECT_EQ(residuals[1], 2);
 }
 
-TEST(BundleAdjustment, ConstantRig) {
+TEST(RigReprojErrorConstantRigCostFunctor, Nominal) {
   Rigid3d cam_from_rig;
   cam_from_rig.translation << 0, 0, -1;
   std::unique_ptr<ceres::CostFunction> cost_function(
-      RigReprojErrorConstantRigCostFunction<SimplePinholeCameraModel>::Create(
-          cam_from_rig, Eigen::Vector2d::Zero()));
+      RigReprojErrorConstantRigCostFunctor<SimplePinholeCameraModel>::Create(
+          Eigen::Vector2d::Zero(), cam_from_rig));
 
   double rig_from_world_rotation[4] = {0, 0, 0, 1};
   double rig_from_world_translation[3] = {0, 0, 1};
@@ -234,10 +228,10 @@ TEST(BundleAdjustment, ConstantRig) {
   EXPECT_EQ(residuals[1], 2);
 }
 
-TEST(BundleAdjustment, RelativePose) {
+TEST(SampsonErrorCostFunctor, Nominal) {
   std::unique_ptr<ceres::CostFunction> cost_function(
-      SampsonErrorCostFunction::Create(Eigen::Vector2d(0, 0),
-                                       Eigen::Vector2d(0, 0)));
+      SampsonErrorCostFunctor::Create(Eigen::Vector3d(0, 0, 1),
+                                      Eigen::Vector3d(0, 0, 1)));
   double cam_from_world_rotation[4] = {1, 0, 0, 0};
   double cam_from_world_translation[3] = {0, 1, 0};
   double residuals[1];
@@ -246,24 +240,57 @@ TEST(BundleAdjustment, RelativePose) {
   EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
   EXPECT_EQ(residuals[0], 0);
 
-  cost_function.reset(SampsonErrorCostFunction::Create(Eigen::Vector2d(0, 0),
-                                                       Eigen::Vector2d(1, 0)));
+  cost_function.reset(SampsonErrorCostFunctor::Create(
+      Eigen::Vector3d(0, 0, 1), Eigen::Vector3d(1, 0, 1)));
   EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
-  EXPECT_EQ(residuals[0], 0.5);
+  EXPECT_NEAR(residuals[0] * residuals[0], 0.5, 1e-6);
 
-  cost_function.reset(SampsonErrorCostFunction::Create(Eigen::Vector2d(0, 0),
-                                                       Eigen::Vector2d(1, 1)));
+  cost_function.reset(SampsonErrorCostFunctor::Create(
+      Eigen::Vector3d(0, 0, 1), Eigen::Vector3d(1, 1, 1)));
   EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
-  EXPECT_EQ(residuals[0], 0.5);
+  EXPECT_NEAR(residuals[0] * residuals[0], 0.5, 1e-6);
 }
 
-TEST(PoseGraphOptimization, AbsolutePose) {
-  const Rigid3d mes_cam_from_world;
-  EigenMatrix6d covariance_cam = EigenMatrix6d::Identity();
-  covariance_cam(5, 5) = 4;
+TEST(AbsolutePosePositionPriorCostFunctor, Nominal) {
   std::unique_ptr<ceres::CostFunction> cost_function(
-      AbsolutePoseErrorCostFunction::Create(mes_cam_from_world,
-                                            covariance_cam));
+      AbsolutePosePositionPriorCostFunctor::Create(Eigen::Vector3d(0, 0, 0)));
+
+  double cam_from_world_rotation[4] = {0, 0, 0, 1};
+  double cam_from_world_translation[3] = {0, 0, 0};
+
+  double residuals[3];
+  const double* parameters[2] = {cam_from_world_rotation,
+                                 cam_from_world_translation};
+  EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
+  EXPECT_EQ(residuals[0], 0);
+  EXPECT_EQ(residuals[1], 0);
+  EXPECT_EQ(residuals[2], 0);
+
+  const Rigid3d cam_from_world(Eigen::Quaterniond::UnitRandom(),
+                               Eigen::Vector3d::Random());
+  const Rigid3d world_from_cam = Inverse(cam_from_world);
+
+  cost_function.reset(
+      AbsolutePosePositionPriorCostFunctor::Create(Eigen::Vector3d::Zero()));
+  parameters[0] = cam_from_world.rotation.coeffs().data();
+  parameters[1] = cam_from_world.translation.data();
+  EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
+  EXPECT_NEAR(residuals[0], -world_from_cam.translation[0], 1e-6);
+  EXPECT_NEAR(residuals[1], -world_from_cam.translation[1], 1e-6);
+  EXPECT_NEAR(residuals[2], -world_from_cam.translation[2], 1e-6);
+
+  cost_function.reset(
+      AbsolutePosePositionPriorCostFunctor::Create(world_from_cam.translation));
+  EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
+  EXPECT_EQ(residuals[0], 0);
+  EXPECT_EQ(residuals[1], 0);
+  EXPECT_EQ(residuals[2], 0);
+}
+
+TEST(AbsolutePosePriorCostFunctor, Nominal) {
+  const Rigid3d cam_from_world_prior;
+  std::unique_ptr<ceres::CostFunction> cost_function(
+      AbsolutePosePriorCostFunctor::Create(cam_from_world_prior));
 
   double cam_from_world_rotation[4] = {0, 0, 0, 1};
   double cam_from_world_translation[3] = {0, 0, 0};
@@ -300,16 +327,14 @@ TEST(PoseGraphOptimization, AbsolutePose) {
   EXPECT_NEAR(residuals[2], 0, 1e-6);
   EXPECT_NEAR(residuals[3], 1, 1e-6);
   EXPECT_NEAR(residuals[4], 2, 1e-6);
-  EXPECT_NEAR(residuals[5], 1.5, 1e-6);
+  EXPECT_NEAR(residuals[5], 3, 1e-6);
 }
 
-TEST(PoseGraphOptimization, RelativePose) {
-  Rigid3d i_from_j;
-  i_from_j.translation << 0, 0, -1;
-  EigenMatrix6d covariance_j = EigenMatrix6d::Identity();
-  covariance_j(5, 5) = 4;
+TEST(RelativePosePriorCostFunctor, Nominal) {
+  Rigid3d i_from_j_prior(Eigen::Quaterniond::Identity(),
+                         Eigen::Vector3d(0, 0, -1));
   std::unique_ptr<ceres::CostFunction> cost_function(
-      MetricRelativePoseErrorCostFunction::Create(i_from_j, covariance_j));
+      RelativePosePriorCostFunctor::Create(i_from_j_prior));
 
   double i_from_world_rotation[4] = {0, 0, 0, 1};
   double i_from_world_translation[3] = {0, 0, 0};
@@ -335,16 +360,16 @@ TEST(PoseGraphOptimization, RelativePose) {
   EXPECT_EQ(residuals[2], 0);
   EXPECT_EQ(residuals[3], 0);
   EXPECT_EQ(residuals[4], 0);
-  EXPECT_EQ(residuals[5], -2);
+  EXPECT_EQ(residuals[5], 4);
 
   j_from_world_translation[0] = 2;
   EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
   EXPECT_EQ(residuals[0], 0);
   EXPECT_EQ(residuals[1], 0);
   EXPECT_EQ(residuals[2], 0);
-  EXPECT_EQ(residuals[3], 2);
+  EXPECT_EQ(residuals[3], -2);
   EXPECT_EQ(residuals[4], 0);
-  EXPECT_EQ(residuals[5], -2);
+  EXPECT_EQ(residuals[5], 4);
 
   // Rotation by 90 degrees around the Y axis.
   Eigen::Matrix3d rotation_matrix;
@@ -353,11 +378,102 @@ TEST(PoseGraphOptimization, RelativePose) {
       rotation_matrix;
   EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
   EXPECT_NEAR(residuals[0], 0, 1e-6);
-  EXPECT_NEAR(residuals[1], DegToRad(90.0), 1e-6);
+  EXPECT_NEAR(residuals[1], DegToRad(-90.0), 1e-6);
   EXPECT_NEAR(residuals[2], 0, 1e-6);
-  EXPECT_NEAR(residuals[3], -3, 1e-6);
+  EXPECT_NEAR(residuals[3], 0, 1e-6);
   EXPECT_NEAR(residuals[4], 0, 1e-6);
-  EXPECT_NEAR(residuals[5], 0.5, 1e-6);
+  EXPECT_NEAR(residuals[5], 2, 1e-6);
+}
+
+TEST(Point3DAlignmentCostFunctor, Nominal) {
+  // generate a test transformation
+  Sim3d tform = Sim3d(RandomUniformReal<double>(0.1, 10),
+                      Eigen::Quaterniond::UnitRandom(),
+                      Eigen::Vector3d::Random());
+  // construct cost function and evaluate
+  // use log scale
+  Eigen::Vector3d point_in_b_prior(1., 1., 1.);
+  Eigen::Vector3d point(0., 0., 0.);
+  std::unique_ptr<ceres::CostFunction> cost_function_log_scale(
+      Point3DAlignmentCostFunctor::Create(point_in_b_prior,
+                                          /*use_log_scale=*/true));
+  double log_scale = std::log(tform.scale);
+  const double* parameters_log_scale[4] = {point.data(),
+                                           tform.rotation.coeffs().data(),
+                                           tform.translation.data(),
+                                           &log_scale};
+  double residuals_log_scale[3];
+  EXPECT_TRUE(cost_function_log_scale->Evaluate(
+      parameters_log_scale, residuals_log_scale, nullptr));
+
+  // use direct scale
+  std::unique_ptr<ceres::CostFunction> cost_function_direct_scale(
+      Point3DAlignmentCostFunctor::Create(point_in_b_prior,
+                                          /*use_log_scale=*/false));
+  const double* parameters_direct_scale[4] = {point.data(),
+                                              tform.rotation.coeffs().data(),
+                                              tform.translation.data(),
+                                              &tform.scale};
+  double residuals_direct_scale[3];
+  EXPECT_TRUE(cost_function_direct_scale->Evaluate(
+      parameters_direct_scale, residuals_direct_scale, nullptr));
+
+  // test with reference computation from C++
+  Eigen::Vector3d error = tform * point - point_in_b_prior;
+  EXPECT_NEAR(residuals_log_scale[0], error[0], 1e-6);
+  EXPECT_NEAR(residuals_log_scale[1], error[1], 1e-6);
+  EXPECT_NEAR(residuals_log_scale[2], error[2], 1e-6);
+  EXPECT_NEAR(residuals_direct_scale[0], error[0], 1e-6);
+  EXPECT_NEAR(residuals_direct_scale[1], error[1], 1e-6);
+  EXPECT_NEAR(residuals_direct_scale[2], error[2], 1e-6);
+}
+
+TEST(CovarianceWeightedCostFunctor, ReprojErrorCostFunctor) {
+  using CostFunctor = ReprojErrorCostFunctor<SimplePinholeCameraModel>;
+  double cam_from_world_rotation[4] = {0, 0, 0, 1};
+  double cam_from_world_translation[3] = {0, 0, 0};
+  double point3D[3] = {-1, 1, 1};
+  double camera_params[3] = {2, 0, 0};
+  double residuals[2];
+  const double* parameters[4] = {cam_from_world_rotation,
+                                 cam_from_world_translation,
+                                 point3D,
+                                 camera_params};
+
+  std::unique_ptr<ceres::CostFunction> cost_function1(
+      CovarianceWeightedCostFunctor<CostFunctor>::Create(
+          Eigen::Matrix2d::Identity(), Eigen::Vector2d::Zero()));
+  EXPECT_TRUE(cost_function1->Evaluate(parameters, residuals, nullptr));
+  EXPECT_EQ(residuals[0], -2);
+  EXPECT_EQ(residuals[1], 2);
+
+  std::unique_ptr<ceres::CostFunction> cost_function2(
+      CovarianceWeightedCostFunctor<CostFunctor>::Create(
+          4 * Eigen::Matrix2d::Identity(), Eigen::Vector2d::Zero()));
+  EXPECT_TRUE(cost_function2->Evaluate(parameters, residuals, nullptr));
+  EXPECT_EQ(residuals[0], -1);
+  EXPECT_EQ(residuals[1], 1);
+}
+
+TEST(CovarianceWeightedCostFunctor, AbsolutePosePositionPriorCostFunctor) {
+  const Rigid3d cam_from_world(Eigen::Quaterniond::UnitRandom(),
+                               Eigen::Vector3d::Random());
+  const Rigid3d world_from_cam = Inverse(cam_from_world);
+
+  double residuals[3];
+  const double* parameters[2] = {cam_from_world.rotation.coeffs().data(),
+                                 cam_from_world.translation.data()};
+
+  std::unique_ptr<ceres::CostFunction> cost_function(
+      CovarianceWeightedCostFunctor<AbsolutePosePositionPriorCostFunctor>::
+          Create(2 * Eigen::Matrix3d::Identity(), Eigen::Vector3d(0, 0, 0)));
+  EXPECT_TRUE(cost_function->Evaluate(parameters, residuals, nullptr));
+  EXPECT_NEAR(
+      residuals[0], -0.5 * std::sqrt(2) * world_from_cam.translation[0], 1e-6);
+  EXPECT_NEAR(
+      residuals[1], -0.5 * std::sqrt(2) * world_from_cam.translation[1], 1e-6);
+  EXPECT_NEAR(
+      residuals[2], -0.5 * std::sqrt(2) * world_from_cam.translation[2], 1e-6);
 }
 
 }  // namespace
