@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -33,45 +33,14 @@
 #include "colmap/estimators/bundle_adjustment.h"
 #include "colmap/estimators/cost_functions.h"
 #include "colmap/estimators/essential_matrix.h"
+#include "colmap/estimators/manifold.h"
 #include "colmap/geometry/essential_matrix.h"
 #include "colmap/geometry/pose.h"
-#include "colmap/math/matrix.h"
+#include "colmap/geometry/triangulation.h"
 #include "colmap/sensor/models.h"
 #include "colmap/util/logging.h"
-#include "colmap/util/threading.h"
 
 namespace colmap {
-namespace {
-
-typedef LORANSAC<P3PEstimator, EPNPEstimator> AbsolutePoseRANSAC;
-
-void EstimateAbsolutePoseKernel(const Camera& camera,
-                                const double focal_length_factor,
-                                const std::vector<Eigen::Vector2d>& points2D,
-                                const std::vector<Eigen::Vector3d>& points3D,
-                                const RANSACOptions& options,
-                                AbsolutePoseRANSAC::Report* report) {
-  // Scale the focal length by the given factor.
-  Camera scaled_camera = camera;
-  for (const size_t idx : camera.FocalLengthIdxs()) {
-    scaled_camera.params[idx] *= focal_length_factor;
-  }
-
-  // Normalize image coordinates with current camera hypothesis.
-  std::vector<Eigen::Vector2d> points2D_in_cam(points2D.size());
-  for (size_t i = 0; i < points2D.size(); ++i) {
-    points2D_in_cam[i] = scaled_camera.CamFromImg(points2D[i]);
-  }
-
-  // Estimate pose for given focal length.
-  auto custom_options = options;
-  custom_options.max_error =
-      scaled_camera.CamFromImgThreshold(options.max_error);
-  AbsolutePoseRANSAC ransac(custom_options);
-  *report = ransac.Estimate(points2D_in_cam, points3D);
-}
-
-}  // namespace
 
 bool EstimateAbsolutePose(const AbsolutePoseEstimationOptions& options,
                           const std::vector<Eigen::Vector2d>& points2D,
@@ -83,125 +52,109 @@ bool EstimateAbsolutePose(const AbsolutePoseEstimationOptions& options,
   THROW_CHECK_EQ(points2D.size(), points3D.size());
   options.Check();
 
-  std::vector<double> focal_length_factors;
-  if (options.estimate_focal_length) {
-    // Generate focal length factors using a quadratic function,
-    // such that more samples are drawn for small focal lengths
-    focal_length_factors.reserve(options.num_focal_length_samples + 1);
-    const double fstep = 1.0 / options.num_focal_length_samples;
-    const double fscale =
-        options.max_focal_length_ratio - options.min_focal_length_ratio;
-    double focal = 0.;
-    for (size_t i = 0; i <= options.num_focal_length_samples;
-         ++i, focal += fstep) {
-      focal_length_factors.push_back(options.min_focal_length_ratio +
-                                     fscale * focal * focal);
-    }
-  } else {
-    focal_length_factors.reserve(1);
-    focal_length_factors.push_back(1);
-  }
-
-  std::vector<std::future<void>> futures;
-  futures.resize(focal_length_factors.size());
-  std::vector<typename AbsolutePoseRANSAC::Report,
-              Eigen::aligned_allocator<typename AbsolutePoseRANSAC::Report>>
-      reports;
-  reports.resize(focal_length_factors.size());
-
-  ThreadPool thread_pool(std::min(
-      options.num_threads, static_cast<int>(focal_length_factors.size())));
-
-  for (size_t i = 0; i < focal_length_factors.size(); ++i) {
-    futures[i] = thread_pool.AddTask(EstimateAbsolutePoseKernel,
-                                     *camera,
-                                     focal_length_factors[i],
-                                     points2D,
-                                     points3D,
-                                     options.ransac_options,
-                                     &reports[i]);
-  }
-
-  double focal_length_factor = 0;
-  Eigen::Matrix3x4d cam_from_world_matrix;
   *num_inliers = 0;
   inlier_mask->clear();
 
-  // Find best model among all focal lengths.
-  for (size_t i = 0; i < focal_length_factors.size(); ++i) {
-    futures[i].get();
-    const auto report = reports[i];
-    if (report.success && report.support.num_inliers > *num_inliers) {
+  if (options.estimate_focal_length) {
+    // TODO(jsch): Implement non-minimal solver for LORANSAC refinement.
+    // Experiments showed marginal difference between RANSAC/LORANSAC for PNPF
+    // after refining the estimates of this function using RefineAbsolutePose.
+    const Eigen::Vector2d principal_point(camera->PrincipalPointX(),
+                                          camera->PrincipalPointY());
+    std::vector<Eigen::Vector2d> points2D_centered(points2D.size());
+    for (size_t i = 0; i < points2D.size(); ++i) {
+      points2D_centered[i] = points2D[i] - principal_point;
+    }
+    RANSAC<P4PFEstimator> ransac(options.ransac_options);
+    auto report = ransac.Estimate(points2D_centered, points3D);
+    if (report.success) {
+      *cam_from_world =
+          Rigid3d(Eigen::Quaterniond(report.model.cam_from_world.leftCols<3>()),
+                  report.model.cam_from_world.col(3));
+      for (const size_t idx : camera->FocalLengthIdxs()) {
+        camera->params[idx] = report.model.focal_length;
+      }
       *num_inliers = report.support.num_inliers;
-      cam_from_world_matrix = report.model;
-      *inlier_mask = report.inlier_mask;
-      focal_length_factor = focal_length_factors[i];
+      *inlier_mask = std::move(report.inlier_mask);
+      return true;
+    }
+  } else {
+    std::vector<P3PEstimator::X_t> points2D_with_rays(points2D.size());
+    for (size_t i = 0; i < points2D.size(); ++i) {
+      points2D_with_rays[i].image_point = points2D[i];
+      if (const std::optional<Eigen::Vector2d> cam_point =
+              camera->CamFromImg(points2D[i]);
+          cam_point) {
+        points2D_with_rays[i].camera_ray =
+            cam_point->homogeneous().normalized();
+      } else {
+        points2D_with_rays[i].camera_ray.setZero();
+      }
+    }
+
+    ImgFromCamFunc img_from_cam_func =
+        std::bind(&Camera::ImgFromCam, camera, std::placeholders::_1);
+    LORANSAC<P3PEstimator, EPNPEstimator> ransac(
+        options.ransac_options,
+        P3PEstimator(img_from_cam_func),
+        EPNPEstimator(img_from_cam_func));
+    auto report = ransac.Estimate(points2D_with_rays, points3D);
+    if (report.success) {
+      *cam_from_world = Rigid3d(Eigen::Quaterniond(report.model.leftCols<3>()),
+                                report.model.col(3));
+      *num_inliers = report.support.num_inliers;
+      *inlier_mask = std::move(report.inlier_mask);
+      return true;
     }
   }
 
-  if (*num_inliers == 0) {
-    return false;
-  }
-
-  // Scale output camera with best estimated focal length.
-  if (options.estimate_focal_length && *num_inliers > 0) {
-    for (const size_t idx : camera->FocalLengthIdxs()) {
-      camera->params[idx] *= focal_length_factor;
-    }
-  }
-
-  *cam_from_world =
-      Rigid3d(Eigen::Quaterniond(cam_from_world_matrix.leftCols<3>()),
-              cam_from_world_matrix.col(3));
-
-  if (cam_from_world->rotation.coeffs().array().isNaN().any() ||
-      cam_from_world->translation.array().isNaN().any()) {
-    return false;
-  }
-
-  return true;
+  return false;
 }
 
-size_t EstimateRelativePose(const RANSACOptions& ransac_options,
-                            const std::vector<Eigen::Vector2d>& points1,
-                            const std::vector<Eigen::Vector2d>& points2,
-                            Rigid3d* cam2_from_cam1) {
-  RANSAC<EssentialMatrixFivePointEstimator> ransac(ransac_options);
-  const auto report = ransac.Estimate(points1, points2);
+bool EstimateRelativePose(const RANSACOptions& ransac_options,
+                          const std::vector<Eigen::Vector3d>& cam_rays1,
+                          const std::vector<Eigen::Vector3d>& cam_rays2,
+                          Rigid3d* cam2_from_cam1,
+                          size_t* num_inliers,
+                          std::vector<char>* inlier_mask) {
+  THROW_CHECK_EQ(cam_rays1.size(), cam_rays2.size());
+
+  LORANSAC<EssentialMatrixFivePointEstimator, EssentialMatrixFivePointEstimator>
+      ransac(ransac_options);
+  auto report = ransac.Estimate(cam_rays1, cam_rays2);
 
   if (!report.success) {
-    return 0;
+    return false;
   }
 
-  std::vector<Eigen::Vector2d> inliers1(report.support.num_inliers);
-  std::vector<Eigen::Vector2d> inliers2(report.support.num_inliers);
+  std::vector<Eigen::Vector3d> inlier_cam_rays1(report.support.num_inliers);
+  std::vector<Eigen::Vector3d> inlier_cam_rays2(report.support.num_inliers);
 
   size_t j = 0;
-  for (size_t i = 0; i < points1.size(); ++i) {
+  for (size_t i = 0; i < cam_rays1.size(); ++i) {
     if (report.inlier_mask[i]) {
-      inliers1[j] = points1[i];
-      inliers2[j] = points2[i];
+      inlier_cam_rays1[j] = cam_rays1[i];
+      inlier_cam_rays2[j] = cam_rays2[i];
       j += 1;
     }
   }
 
-  Eigen::Matrix3d cam2_from_cam1_rot_mat;
   std::vector<Eigen::Vector3d> points3D;
   PoseFromEssentialMatrix(report.model,
-                          inliers1,
-                          inliers2,
-                          &cam2_from_cam1_rot_mat,
-                          &cam2_from_cam1->translation,
+                          inlier_cam_rays1,
+                          inlier_cam_rays2,
+                          cam2_from_cam1,
                           &points3D);
 
-  cam2_from_cam1->rotation = Eigen::Quaterniond(cam2_from_cam1_rot_mat);
-
   if (cam2_from_cam1->rotation.coeffs().array().isNaN().any() ||
       cam2_from_cam1->translation.array().isNaN().any()) {
-    return 0;
+    return false;
   }
 
-  return points3D.size();
+  *num_inliers = report.support.num_inliers;
+  *inlier_mask = std::move(report.inlier_mask);
+
+  return !points3D.empty();
 }
 
 bool RefineAbsolutePose(const AbsolutePoseRefinementOptions& options,
@@ -219,8 +172,11 @@ bool RefineAbsolutePose(const AbsolutePoseRefinementOptions& options,
       std::make_unique<ceres::CauchyLoss>(options.loss_function_scale);
 
   double* camera_params = camera->params.data();
-  double* rig_from_world_rotation = cam_from_world->rotation.coeffs().data();
-  double* rig_from_world_translation = cam_from_world->translation.data();
+  double* cam_from_world_rotation = cam_from_world->rotation.coeffs().data();
+  double* cam_from_world_translation = cam_from_world->translation.data();
+
+  // CostFunction assumes unit quaternions.
+  cam_from_world->rotation.normalize();
 
   ceres::Problem::Options problem_options;
   problem_options.loss_function_ownership = ceres::DO_NOT_TAKE_OWNERSHIP;
@@ -232,16 +188,16 @@ bool RefineAbsolutePose(const AbsolutePoseRefinementOptions& options,
       continue;
     }
     problem.AddResidualBlock(
-        CameraCostFunction<ReprojErrorConstantPoint3DCostFunction>(
+        CreateCameraCostFunction<ReprojErrorConstantPoint3DCostFunctor>(
             camera->model_id, points2D[i], points3D[i]),
         loss_function.get(),
-        rig_from_world_rotation,
-        rig_from_world_translation,
+        cam_from_world_rotation,
+        cam_from_world_translation,
         camera_params);
   }
 
   if (problem.NumResiduals() > 0) {
-    SetQuaternionManifold(&problem, rig_from_world_rotation);
+    SetQuaternionManifold(&problem, cam_from_world_rotation);
 
     // Camera parameterization.
     if (!options.refine_focal_length && !options.refine_extra_params) {
@@ -299,11 +255,15 @@ bool RefineAbsolutePose(const AbsolutePoseRefinementOptions& options,
     PrintSolverSummary(summary, "Pose refinement report");
   }
 
+  if (!summary.IsSolutionUsable()) {
+    return false;
+  }
+
   if (problem.NumResiduals() > 0 && cam_from_world_cov != nullptr) {
     ceres::Covariance::Options options;
     ceres::Covariance covariance(options);
-    std::vector<const double*> parameter_blocks = {rig_from_world_rotation,
-                                                   rig_from_world_translation};
+    std::vector<const double*> parameter_blocks = {cam_from_world_rotation,
+                                                   cam_from_world_translation};
     if (!covariance.Compute(parameter_blocks, &problem)) {
       return false;
     }
@@ -314,14 +274,16 @@ bool RefineAbsolutePose(const AbsolutePoseRefinementOptions& options,
                                                  cam_from_world_cov->data());
   }
 
-  return summary.IsSolutionUsable();
+  return true;
 }
 
 bool RefineRelativePose(const ceres::Solver::Options& options,
-                        const std::vector<Eigen::Vector2d>& points1,
-                        const std::vector<Eigen::Vector2d>& points2,
+                        const std::vector<char>& inlier_mask,
+                        const std::vector<Eigen::Vector3d>& cam_rays1,
+                        const std::vector<Eigen::Vector3d>& cam_rays2,
                         Rigid3d* cam2_from_cam1) {
-  THROW_CHECK_EQ(points1.size(), points2.size());
+  THROW_CHECK_EQ(cam_rays1.size(), cam_rays2.size());
+  THROW_CHECK_EQ(cam_rays1.size(), inlier_mask.size());
 
   // CostFunction assumes unit quaternions.
   cam2_from_cam1->rotation.normalize();
@@ -329,16 +291,22 @@ bool RefineRelativePose(const ceres::Solver::Options& options,
   double* cam2_from_cam1_rotation = cam2_from_cam1->rotation.coeffs().data();
   double* cam2_from_cam1_translation = cam2_from_cam1->translation.data();
 
-  const double kMaxL2Error = 1.0;
-  ceres::LossFunction* loss_function = new ceres::CauchyLoss(kMaxL2Error);
+  constexpr double kMaxL2Error = 1.0;
+  const auto loss_function = std::make_unique<ceres::CauchyLoss>(kMaxL2Error);
 
-  ceres::Problem problem;
+  ceres::Problem::Options problem_options;
+  problem_options.loss_function_ownership = ceres::DO_NOT_TAKE_OWNERSHIP;
+  ceres::Problem problem(problem_options);
 
-  for (size_t i = 0; i < points1.size(); ++i) {
+  for (size_t i = 0; i < cam_rays1.size(); ++i) {
+    // Skip outlier observations
+    if (!inlier_mask[i]) {
+      continue;
+    }
     ceres::CostFunction* cost_function =
-        SampsonErrorCostFunction::Create(points1[i], points2[i]);
+        SampsonErrorCostFunctor::Create(cam_rays1[i], cam_rays2[i]);
     problem.AddResidualBlock(cost_function,
-                             loss_function,
+                             loss_function.get(),
                              cam2_from_cam1_rotation,
                              cam2_from_cam1_translation);
   }
@@ -353,12 +321,12 @@ bool RefineRelativePose(const ceres::Solver::Options& options,
 }
 
 bool RefineEssentialMatrix(const ceres::Solver::Options& options,
-                           const std::vector<Eigen::Vector2d>& points1,
-                           const std::vector<Eigen::Vector2d>& points2,
+                           const std::vector<Eigen::Vector3d>& cam_rays1,
+                           const std::vector<Eigen::Vector3d>& cam_rays2,
                            const std::vector<char>& inlier_mask,
                            Eigen::Matrix3d* E) {
-  THROW_CHECK_EQ(points1.size(), points2.size());
-  THROW_CHECK_EQ(points1.size(), inlier_mask.size());
+  THROW_CHECK_EQ(cam_rays1.size(), cam_rays2.size());
+  THROW_CHECK_EQ(cam_rays1.size(), inlier_mask.size());
 
   // Extract inlier points for decomposing the essential matrix into
   // rotation and translation components.
@@ -370,29 +338,22 @@ bool RefineEssentialMatrix(const ceres::Solver::Options& options,
     }
   }
 
-  std::vector<Eigen::Vector2d> inlier_points1(num_inliers);
-  std::vector<Eigen::Vector2d> inlier_points2(num_inliers);
+  std::vector<Eigen::Vector3d> inlier_cam_rays1(num_inliers);
+  std::vector<Eigen::Vector3d> inlier_cam_rays2(num_inliers);
   size_t j = 0;
   for (size_t i = 0; i < inlier_mask.size(); ++i) {
     if (inlier_mask[i]) {
-      inlier_points1[j] = points1[i];
-      inlier_points2[j] = points2[i];
+      inlier_cam_rays1[j] = cam_rays1[i];
+      inlier_cam_rays2[j] = cam_rays2[i];
       j += 1;
     }
   }
 
   // Extract relative pose from essential matrix.
-
   Rigid3d cam2_from_cam1;
-  Eigen::Matrix3d cam2_from_cam1_rot_mat;
   std::vector<Eigen::Vector3d> points3D;
-  PoseFromEssentialMatrix(*E,
-                          inlier_points1,
-                          inlier_points2,
-                          &cam2_from_cam1_rot_mat,
-                          &cam2_from_cam1.translation,
-                          &points3D);
-  cam2_from_cam1.rotation = Eigen::Quaterniond(cam2_from_cam1_rot_mat);
+  PoseFromEssentialMatrix(
+      *E, inlier_cam_rays1, inlier_cam_rays2, &cam2_from_cam1, &points3D);
 
   if (points3D.size() == 0) {
     return false;
@@ -401,8 +362,12 @@ bool RefineEssentialMatrix(const ceres::Solver::Options& options,
   // Refine essential matrix, use all points so that refinement is able to
   // consider points as inliers that were originally outliers.
 
-  const bool refinement_success = RefineRelativePose(
-      options, inlier_points1, inlier_points2, &cam2_from_cam1);
+  const bool refinement_success =
+      RefineRelativePose(options,
+                         std::vector<char>(num_inliers, true),
+                         inlier_cam_rays1,
+                         inlier_cam_rays2,
+                         &cam2_from_cam1);
 
   if (!refinement_success) {
     return false;
