@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -31,12 +31,14 @@
 
 #include "colmap/controllers/automatic_reconstruction.h"
 #include "colmap/controllers/bundle_adjustment.h"
-#include "colmap/controllers/hierarchical_mapper.h"
+#include "colmap/controllers/hierarchical_pipeline.h"
 #include "colmap/controllers/option_manager.h"
 #include "colmap/estimators/similarity_transform.h"
 #include "colmap/exe/gui.h"
 #include "colmap/scene/reconstruction.h"
+#include "colmap/scene/rig.h"
 #include "colmap/sfm/observation_manager.h"
+#include "colmap/util/file.h"
 #include "colmap/util/misc.h"
 #include "colmap/util/opengl_utils.h"
 
@@ -44,9 +46,43 @@
 #include <boost/property_tree/ptree.hpp>
 
 namespace colmap {
+namespace {
+
+std::pair<std::vector<image_t>, std::vector<Eigen::Vector3d>>
+ExtractExistingImages(const Reconstruction& reconstruction) {
+  std::vector<image_t> fixed_image_ids = reconstruction.RegImageIds();
+  std::vector<Eigen::Vector3d> orig_fixed_image_positions;
+  orig_fixed_image_positions.reserve(fixed_image_ids.size());
+  for (const image_t image_id : fixed_image_ids) {
+    orig_fixed_image_positions.push_back(
+        reconstruction.Image(image_id).ProjectionCenter());
+  }
+  return {std::move(fixed_image_ids), std::move(orig_fixed_image_positions)};
+}
+
+void UpdateDatabasePosePriorsCovariance(const std::string& database_path,
+                                        const Eigen::Matrix3d& covariance) {
+  Database database(database_path);
+  DatabaseTransaction database_transaction(&database);
+
+  LOG(INFO)
+      << "Setting up database pose priors with the same covariance matrix: \n"
+      << covariance << '\n';
+
+  for (const auto& image : database.ReadAllImages()) {
+    if (database.ExistsPosePrior(image.ImageId())) {
+      PosePrior prior = database.ReadPosePrior(image.ImageId());
+      prior.position_covariance = covariance;
+      database.UpdatePosePrior(image.ImageId(), prior);
+    }
+  }
+}
+
+}  // namespace
 
 int RunAutomaticReconstructor(int argc, char** argv) {
   AutomaticReconstructionController::Options reconstruction_options;
+  std::string image_list_path;
   std::string data_type = "individual";
   std::string quality = "high";
   std::string mesher = "poisson";
@@ -55,6 +91,7 @@ int RunAutomaticReconstructor(int argc, char** argv) {
   options.AddRequiredOption("workspace_path",
                             &reconstruction_options.workspace_path);
   options.AddRequiredOption("image_path", &reconstruction_options.image_path);
+  options.AddDefaultOption("image_list_path", &image_list_path);
   options.AddDefaultOption("mask_path", &reconstruction_options.mask_path);
   options.AddDefaultOption("vocab_tree_path",
                            &reconstruction_options.vocab_tree_path);
@@ -65,8 +102,12 @@ int RunAutomaticReconstructor(int argc, char** argv) {
                            &reconstruction_options.camera_model);
   options.AddDefaultOption("single_camera",
                            &reconstruction_options.single_camera);
+  options.AddDefaultOption("single_camera_per_folder",
+                           &reconstruction_options.single_camera_per_folder);
   options.AddDefaultOption("camera_params",
                            &reconstruction_options.camera_params);
+  options.AddDefaultOption("extraction", &reconstruction_options.extraction);
+  options.AddDefaultOption("matching", &reconstruction_options.matching);
   options.AddDefaultOption("sparse", &reconstruction_options.sparse);
   options.AddDefaultOption("dense", &reconstruction_options.dense);
   options.AddDefaultOption("mesher", &mesher, "{poisson, delaunay}");
@@ -75,51 +116,26 @@ int RunAutomaticReconstructor(int argc, char** argv) {
   options.AddDefaultOption("gpu_index", &reconstruction_options.gpu_index);
   options.Parse(argc, argv);
 
-  StringToLower(&data_type);
-  if (data_type == "individual") {
-    reconstruction_options.data_type =
-        AutomaticReconstructionController::DataType::INDIVIDUAL;
-  } else if (data_type == "video") {
-    reconstruction_options.data_type =
-        AutomaticReconstructionController::DataType::VIDEO;
-  } else if (data_type == "internet") {
-    reconstruction_options.data_type =
-        AutomaticReconstructionController::DataType::INTERNET;
-  } else {
-    LOG(FATAL_THROW) << "Invalid data type provided";
+  if (!image_list_path.empty()) {
+    reconstruction_options.image_names = ReadTextFileLines(image_list_path);
   }
 
-  StringToLower(&quality);
-  if (quality == "low") {
-    reconstruction_options.quality =
-        AutomaticReconstructionController::Quality::LOW;
-  } else if (quality == "medium") {
-    reconstruction_options.quality =
-        AutomaticReconstructionController::Quality::MEDIUM;
-  } else if (quality == "high") {
-    reconstruction_options.quality =
-        AutomaticReconstructionController::Quality::HIGH;
-  } else if (quality == "extreme") {
-    reconstruction_options.quality =
-        AutomaticReconstructionController::Quality::EXTREME;
-  } else {
-    LOG(FATAL_THROW) << "Invalid quality provided";
-  }
+  StringToUpper(&data_type);
+  reconstruction_options.data_type =
+      AutomaticReconstructionController::DataTypeFromString(data_type);
 
-  StringToLower(&mesher);
-  if (mesher == "poisson") {
-    reconstruction_options.mesher =
-        AutomaticReconstructionController::Mesher::POISSON;
-  } else if (mesher == "delaunay") {
-    reconstruction_options.mesher =
-        AutomaticReconstructionController::Mesher::DELAUNAY;
-  } else {
-    LOG(FATAL_THROW) << "Invalid mesher provided";
-  }
+  StringToUpper(&quality);
+  reconstruction_options.quality =
+      AutomaticReconstructionController::QualityFromString(quality);
+
+  StringToUpper(&mesher);
+  reconstruction_options.mesher =
+      AutomaticReconstructionController::MesherFromString(mesher);
 
   auto reconstruction_manager = std::make_shared<ReconstructionManager>();
 
-  if (reconstruction_options.use_gpu && kUseOpenGL) {
+  if (reconstruction_options.use_gpu && kUseOpenGL &&
+      (reconstruction_options.extraction || reconstruction_options.matching)) {
     QApplication app(argc, argv);
     AutomaticReconstructionController controller(reconstruction_options,
                                                  reconstruction_manager);
@@ -203,9 +219,7 @@ int RunMapper(int argc, char** argv) {
   }
 
   if (!image_list_path.empty()) {
-    const auto image_names = ReadTextFileLines(image_list_path);
-    options.mapper->image_names =
-        std::unordered_set<std::string>(image_names.begin(), image_names.end());
+    options.mapper->image_names = ReadTextFileLines(image_list_path);
   }
 
   auto reconstruction_manager = std::make_shared<ReconstructionManager>();
@@ -217,47 +231,41 @@ int RunMapper(int argc, char** argv) {
     reconstruction_manager->Read(input_path);
   }
 
-  // If fix_existing_images is enabled, we store the initial positions of
+  // If fix_existing_frames is enabled, we store the initial positions of
   // existing images in order to transform them back to the original coordinate
   // frame, as the reconstruction is normalized multiple times for numerical
   // stability.
   std::vector<Eigen::Vector3d> orig_fixed_image_positions;
   std::vector<image_t> fixed_image_ids;
-  if (options.mapper->fix_existing_images &&
+  if (options.mapper->fix_existing_frames &&
       reconstruction_manager->Size() > 0) {
-    const auto& reconstruction = reconstruction_manager->Get(0);
-    fixed_image_ids = reconstruction->RegImageIds();
-    orig_fixed_image_positions.reserve(fixed_image_ids.size());
-    for (const image_t image_id : fixed_image_ids) {
-      orig_fixed_image_positions.push_back(
-          reconstruction->Image(image_id).ProjectionCenter());
-    }
+    std::tie(fixed_image_ids, orig_fixed_image_positions) =
+        ExtractExistingImages(*reconstruction_manager->Get(0));
   }
 
-  IncrementalMapperController mapper(options.mapper,
-                                     *options.image_path,
-                                     *options.database_path,
-                                     reconstruction_manager);
+  IncrementalPipeline mapper(options.mapper,
+                             *options.image_path,
+                             *options.database_path,
+                             reconstruction_manager);
 
   // In case a new reconstruction is started, write results of individual sub-
   // models to as their reconstruction finishes instead of writing all results
   // after all reconstructions finished.
   size_t prev_num_reconstructions = 0;
   if (input_path == "") {
-    mapper.AddCallback(
-        IncrementalMapperController::LAST_IMAGE_REG_CALLBACK, [&]() {
-          // If the number of reconstructions has not changed, the last model
-          // was discarded for some reason.
-          if (reconstruction_manager->Size() > prev_num_reconstructions) {
-            const std::string reconstruction_path = JoinPaths(
-                output_path, std::to_string(prev_num_reconstructions));
-            CreateDirIfNotExists(reconstruction_path);
-            reconstruction_manager->Get(prev_num_reconstructions)
-                ->Write(reconstruction_path);
-            options.Write(JoinPaths(reconstruction_path, "project.ini"));
-            prev_num_reconstructions = reconstruction_manager->Size();
-          }
-        });
+    mapper.AddCallback(IncrementalPipeline::LAST_IMAGE_REG_CALLBACK, [&]() {
+      // If the number of reconstructions has not changed, the last model
+      // was discarded for some reason.
+      if (reconstruction_manager->Size() > prev_num_reconstructions) {
+        const std::string reconstruction_path =
+            JoinPaths(output_path, std::to_string(prev_num_reconstructions));
+        CreateDirIfNotExists(reconstruction_path);
+        reconstruction_manager->Get(prev_num_reconstructions)
+            ->Write(reconstruction_path);
+        options.Write(JoinPaths(reconstruction_path, "project.ini"));
+        prev_num_reconstructions = reconstruction_manager->Size();
+      }
+    });
   }
 
   mapper.Run();
@@ -273,7 +281,7 @@ int RunMapper(int argc, char** argv) {
     const auto& reconstruction = reconstruction_manager->Get(0);
 
     // Transform the final reconstruction back to the original coordinate frame.
-    if (options.mapper->fix_existing_images) {
+    if (options.mapper->fix_existing_frames) {
       if (fixed_image_ids.size() < 3) {
         LOG(WARNING) << "Too few images to transform the reconstruction.";
       } else {
@@ -302,7 +310,7 @@ int RunMapper(int argc, char** argv) {
 }
 
 int RunHierarchicalMapper(int argc, char** argv) {
-  HierarchicalMapperController::Options mapper_options;
+  HierarchicalPipeline::Options mapper_options;
   std::string output_path;
 
   OptionManager options;
@@ -325,8 +333,8 @@ int RunHierarchicalMapper(int argc, char** argv) {
 
   mapper_options.incremental_options = *options.mapper;
   auto reconstruction_manager = std::make_shared<ReconstructionManager>();
-  HierarchicalMapperController hierarchical_mapper(mapper_options,
-                                                   reconstruction_manager);
+  HierarchicalPipeline hierarchical_mapper(mapper_options,
+                                           reconstruction_manager);
   hierarchical_mapper.Run();
 
   if (reconstruction_manager->Size() == 0) {
@@ -340,6 +348,139 @@ int RunHierarchicalMapper(int argc, char** argv) {
   return EXIT_SUCCESS;
 }
 
+int RunPosePriorMapper(int argc, char** argv) {
+  std::string input_path;
+  std::string output_path;
+
+  bool overwrite_priors_covariance = false;
+  double prior_position_std_x = 1.;
+  double prior_position_std_y = 1.;
+  double prior_position_std_z = 1.;
+
+  OptionManager options;
+  options.AddDatabaseOptions();
+  options.AddImageOptions();
+  options.AddDefaultOption("input_path", &input_path);
+  options.AddRequiredOption("output_path", &output_path);
+  options.AddMapperOptions();
+
+  options.mapper->use_prior_position = true;
+
+  options.AddDefaultOption(
+      "overwrite_priors_covariance",
+      &overwrite_priors_covariance,
+      "Priors covariance read from database. If true, overwrite the priors "
+      "covariance using the follwoing prior_position_std_... options");
+  options.AddDefaultOption("prior_position_std_x", &prior_position_std_x);
+  options.AddDefaultOption("prior_position_std_y", &prior_position_std_y);
+  options.AddDefaultOption("prior_position_std_z", &prior_position_std_z);
+  options.AddDefaultOption("use_robust_loss_on_prior_position",
+                           &options.mapper->use_robust_loss_on_prior_position);
+  options.AddDefaultOption("prior_position_loss_scale",
+                           &options.mapper->prior_position_loss_scale);
+  options.Parse(argc, argv);
+
+  if (!ExistsDir(output_path)) {
+    LOG(ERROR) << "`output_path` is not a directory.";
+    return EXIT_FAILURE;
+  }
+
+  if (overwrite_priors_covariance) {
+    const Eigen::Matrix3d covariance =
+        Eigen::Vector3d(
+            prior_position_std_x, prior_position_std_y, prior_position_std_z)
+            .cwiseAbs2()
+            .asDiagonal();
+    UpdateDatabasePosePriorsCovariance(*options.database_path, covariance);
+  }
+
+  auto reconstruction_manager = std::make_shared<ReconstructionManager>();
+  if (input_path != "") {
+    if (!ExistsDir(input_path)) {
+      LOG(ERROR) << "`input_path` is not a directory.";
+      return EXIT_FAILURE;
+    }
+    reconstruction_manager->Read(input_path);
+  }
+
+  // If fix_existing_frames is enabled, we store the initial positions of
+  // existing images in order to transform them back to the original coordinate
+  // frame, as the reconstruction is normalized multiple times for numerical
+  // stability.
+  std::vector<Eigen::Vector3d> orig_fixed_image_positions;
+  std::vector<image_t> fixed_image_ids;
+  if (options.mapper->fix_existing_frames &&
+      reconstruction_manager->Size() > 0) {
+    std::tie(fixed_image_ids, orig_fixed_image_positions) =
+        ExtractExistingImages(*reconstruction_manager->Get(0));
+  }
+
+  IncrementalPipeline mapper(options.mapper,
+                             *options.image_path,
+                             *options.database_path,
+                             reconstruction_manager);
+
+  // In case a new reconstruction is started, write results of individual sub-
+  // models to as their reconstruction finishes instead of writing all results
+  // after all reconstructions finished.
+  size_t prev_num_reconstructions = 0;
+  if (input_path == "") {
+    mapper.AddCallback(IncrementalPipeline::LAST_IMAGE_REG_CALLBACK, [&]() {
+      // If the number of reconstructions has not changed, the last model
+      // was discarded for some reason.
+      if (reconstruction_manager->Size() > prev_num_reconstructions) {
+        const std::string reconstruction_path =
+            JoinPaths(output_path, std::to_string(prev_num_reconstructions));
+        CreateDirIfNotExists(reconstruction_path);
+        reconstruction_manager->Get(prev_num_reconstructions)
+            ->Write(reconstruction_path);
+        options.Write(JoinPaths(reconstruction_path, "project.ini"));
+        prev_num_reconstructions = reconstruction_manager->Size();
+      }
+    });
+  }
+
+  mapper.Run();
+
+  if (reconstruction_manager->Size() == 0) {
+    LOG(ERROR) << "failed to create sparse model";
+    return EXIT_FAILURE;
+  }
+
+  // In case the reconstruction is continued from an existing reconstruction, do
+  // not create sub-folders but directly write the results.
+  if (input_path != "") {
+    const auto& reconstruction = reconstruction_manager->Get(0);
+
+    // Transform the final reconstruction back to the original coordinate frame.
+    if (options.mapper->fix_existing_frames) {
+      if (fixed_image_ids.size() < 3) {
+        LOG(WARNING) << "Too few images to transform the reconstruction.";
+      } else {
+        std::vector<Eigen::Vector3d> new_fixed_image_positions;
+        new_fixed_image_positions.reserve(fixed_image_ids.size());
+        for (const image_t image_id : fixed_image_ids) {
+          new_fixed_image_positions.push_back(
+              reconstruction->Image(image_id).ProjectionCenter());
+        }
+        Sim3d orig_from_new;
+        if (EstimateSim3d(new_fixed_image_positions,
+                          orig_fixed_image_positions,
+                          orig_from_new)) {
+          reconstruction->Transform(orig_from_new);
+        } else {
+          LOG(WARNING) << "Failed to transform the reconstruction back "
+                          "to the input coordinate frame.";
+        }
+      }
+    }
+
+    reconstruction->Write(output_path);
+  }
+
+  return EXIT_SUCCESS;
+}
+
 int RunPointFiltering(int argc, char** argv) {
   std::string input_path;
   std::string output_path;
@@ -431,7 +572,7 @@ void RunPointTriangulatorImpl(
     const std::string& database_path,
     const std::string& image_path,
     const std::string& output_path,
-    const IncrementalMapperOptions& options,
+    const IncrementalPipelineOptions& options,
     const bool clear_points,
     const bool refine_intrinsics) {
   THROW_CHECK_GE(reconstruction->NumRegImages(), 2)
@@ -442,238 +583,60 @@ void RunPointTriangulatorImpl(
     reconstruction->TranscribeImageIdsToDatabase(database);
   }
 
-  auto options_tmp = std::make_shared<IncrementalMapperOptions>(options);
-  options_tmp->fix_existing_images = true;
+  auto options_tmp = std::make_shared<IncrementalPipelineOptions>(options);
+  options_tmp->fix_existing_frames = true;
   options_tmp->ba_refine_focal_length = refine_intrinsics;
   options_tmp->ba_refine_principal_point = false;
   options_tmp->ba_refine_extra_params = refine_intrinsics;
 
   auto reconstruction_manager = std::make_shared<ReconstructionManager>();
-  IncrementalMapperController mapper(
+  IncrementalPipeline mapper(
       options_tmp, image_path, database_path, reconstruction_manager);
   mapper.TriangulateReconstruction(reconstruction);
   reconstruction->Write(output_path);
 }
 
-namespace {
-
-// Read the configuration of the camera rigs from a JSON file. The input images
-// of a camera rig must be named consistently to assign them to the appropriate
-// camera rig and the respective snapshots.
-//
-// An example configuration of a single camera rig:
-// [
-//   {
-//     "ref_camera_id": 1,
-//     "cameras":
-//     [
-//       {
-//           "camera_id": 1,
-//           "image_prefix": "left1_image"
-//           "cam_from_rig_rotation": [1, 0, 0, 0],
-//           "cam_from_rig_translation": [0, 0, 0]
-//       },
-//       {
-//           "camera_id": 2,
-//           "image_prefix": "left2_image"
-//           "cam_from_rig_rotation": [1, 0, 0, 0],
-//           "cam_from_rig_translation": [0, 0, 1]
-//       },
-//       {
-//           "camera_id": 3,
-//           "image_prefix": "right1_image"
-//           "cam_from_rig_rotation": [1, 0, 0, 0],
-//           "cam_from_rig_translation": [0, 0, 2]
-//       },
-//       {
-//           "camera_id": 4,
-//           "image_prefix": "right2_image"
-//           "cam_from_rig_rotation": [1, 0, 0, 0],
-//           "cam_from_rig_translation": [0, 0, 3]
-//       }
-//     ]
-//   }
-// ]
-//
-// The "camera_id" and "image_prefix" fields are required, whereas the
-// "cam_from_rig_rotation" and "cam_from_rig_translation" fields optionally
-// specify the relative extrinsics of the camera rig in the form of a
-// translation vector and a rotation quaternion (w, x, y, z). If the relative
-// extrinsics are not provided then they are automatically inferred from the
-// reconstruction.
-//
-// This file specifies the configuration for a single camera rig and that you
-// could potentially define multiple camera rigs. The rig is composed of 4
-// cameras: all images of the first camera must have "left1_image" as a name
-// prefix, e.g., "left1_image_frame000.png" or "left1_image/frame000.png".
-// Images with the same suffix ("_frame000.png" and "/frame000.png") are
-// assigned to the same snapshot, i.e., they are assumed to be captured at the
-// same time. Only snapshots with the reference image registered will be added
-// to the bundle adjustment problem. The remaining images will be added with
-// independent poses to the bundle adjustment problem. The above configuration
-// could have the following input image file structure:
-//
-//    /path/to/images/...
-//        left1_image/...
-//            frame000.png
-//            frame001.png
-//            frame002.png
-//            ...
-//        left2_image/...
-//            frame000.png
-//            frame001.png
-//            frame002.png
-//            ...
-//        right1_image/...
-//            frame000.png
-//            frame001.png
-//            frame002.png
-//            ...
-//        right2_image/...
-//            frame000.png
-//            frame001.png
-//            frame002.png
-//            ...
-//
-std::vector<CameraRig> ReadCameraRigConfig(const std::string& rig_config_path,
-                                           const Reconstruction& reconstruction,
-                                           bool estimate_rig_relative_poses) {
-  boost::property_tree::ptree pt;
-  boost::property_tree::read_json(rig_config_path.c_str(), pt);
-
-  std::vector<CameraRig> camera_rigs;
-  for (const auto& rig_config : pt) {
-    CameraRig camera_rig;
-
-    std::vector<std::string> image_prefixes;
-    for (const auto& camera : rig_config.second.get_child("cameras")) {
-      const int camera_id = camera.second.get<int>("camera_id");
-      image_prefixes.push_back(camera.second.get<std::string>("image_prefix"));
-
-      Rigid3d cam_from_rig;
-
-      auto cam_from_rig_rotation_node =
-          camera.second.get_child_optional("cam_from_rig_rotation");
-      if (cam_from_rig_rotation_node) {
-        int index = 0;
-        Eigen::Vector4d cam_from_rig_wxyz;
-        for (const auto& node : cam_from_rig_rotation_node.get()) {
-          cam_from_rig_wxyz[index++] = node.second.get_value<double>();
-        }
-        cam_from_rig.rotation = Eigen::Quaterniond(cam_from_rig_wxyz(0),
-                                                   cam_from_rig_wxyz(1),
-                                                   cam_from_rig_wxyz(2),
-                                                   cam_from_rig_wxyz(3));
-      } else {
-        estimate_rig_relative_poses = true;
-      }
-
-      auto cam_from_rig_translation_node =
-          camera.second.get_child_optional("cam_from_rig_translation");
-      if (cam_from_rig_translation_node) {
-        int index = 0;
-        for (const auto& node : cam_from_rig_translation_node.get()) {
-          cam_from_rig.translation(index++) = node.second.get_value<double>();
-        }
-      } else {
-        estimate_rig_relative_poses = true;
-      }
-
-      camera_rig.AddCamera(camera_id, cam_from_rig);
-    }
-
-    camera_rig.SetRefCameraId(rig_config.second.get<int>("ref_camera_id"));
-
-    std::unordered_map<std::string, std::vector<image_t>> snapshots;
-    for (const auto image_id : reconstruction.RegImageIds()) {
-      const auto& image = reconstruction.Image(image_id);
-      for (const auto& image_prefix : image_prefixes) {
-        if (StringContains(image.Name(), image_prefix)) {
-          const std::string image_suffix =
-              StringGetAfter(image.Name(), image_prefix);
-          snapshots[image_suffix].push_back(image_id);
-        }
-      }
-    }
-
-    for (const auto& snapshot : snapshots) {
-      bool has_ref_camera = false;
-      for (const auto image_id : snapshot.second) {
-        const auto& image = reconstruction.Image(image_id);
-        if (image.CameraId() == camera_rig.RefCameraId()) {
-          has_ref_camera = true;
-          break;
-        }
-      }
-
-      if (has_ref_camera) {
-        camera_rig.AddSnapshot(snapshot.second);
-      }
-    }
-
-    camera_rig.Check(reconstruction);
-    if (estimate_rig_relative_poses) {
-      PrintHeading2("Estimating relative rig poses");
-      if (!camera_rig.ComputeCamsFromRigs(reconstruction)) {
-        LOG(WARNING) << "Failed to estimate rig poses from reconstruction; "
-                        "cannot use rig BA";
-        return std::vector<CameraRig>();
-      }
-    }
-
-    camera_rigs.push_back(camera_rig);
-  }
-
-  return camera_rigs;
-}
-
-}  // namespace
-
+// TODO: Remove once version 3.12 is released.
 int RunRigBundleAdjuster(int argc, char** argv) {
   std::string input_path;
   std::string output_path;
   std::string rig_config_path;
-  bool estimate_rig_relative_poses = true;
-
-  RigBundleAdjuster::Options rig_ba_options;
 
   OptionManager options;
   options.AddRequiredOption("input_path", &input_path);
   options.AddRequiredOption("output_path", &output_path);
   options.AddRequiredOption("rig_config_path", &rig_config_path);
-  options.AddDefaultOption("estimate_rig_relative_poses",
-                           &estimate_rig_relative_poses);
-  options.AddDefaultOption("RigBundleAdjustment.refine_relative_poses",
-                           &rig_ba_options.refine_relative_poses);
   options.AddBundleAdjustmentOptions();
   options.Parse(argc, argv);
 
+  LOG(WARNING)
+      << "rig_bundle_adjuster is deprecated and will be removed in the next "
+         "version, run rig_configurator and bundle_adjuster instead.";
+
   Reconstruction reconstruction;
   reconstruction.Read(input_path);
 
-  PrintHeading1("Camera rig configuration");
-
-  auto camera_rigs = ReadCameraRigConfig(
-      rig_config_path, reconstruction, estimate_rig_relative_poses);
-
   BundleAdjustmentConfig config;
-  for (size_t i = 0; i < camera_rigs.size(); ++i) {
-    const auto& camera_rig = camera_rigs[i];
-    PrintHeading2(StringPrintf("Camera Rig %d", i + 1));
-    LOG(INFO) << StringPrintf("Cameras: %d", camera_rig.NumCameras());
-    LOG(INFO) << StringPrintf("Snapshots: %d", camera_rig.NumSnapshots());
-
-    // Add all registered images to the bundle adjustment configuration.
-    for (const auto image_id : reconstruction.RegImageIds()) {
-      config.AddImage(image_id);
-    }
+  for (const image_t image_id : reconstruction.RegImageIds()) {
+    config.AddImage(image_id);
   }
 
-  PrintHeading1("Rig bundle adjustment");
+  Database database(Database::kInMemoryDatabasePath);
+  for (const auto& [_, camera] : reconstruction.Cameras()) {
+    database.WriteCamera(camera, /*use_camera_id=*/true);
+  }
+  for (const auto& [image_id, image] : reconstruction.Images()) {
+    database.WriteImage(image, /*use_image_id=*/true);
+    config.AddImage(image_id);
+  }
+  ApplyRigConfig(ReadRigConfig(rig_config_path), database, &reconstruction);
 
-  BundleAdjustmentOptions ba_options = *options.bundle_adjustment;
-  RigBundleAdjuster bundle_adjuster(ba_options, rig_ba_options, config);
-  THROW_CHECK(bundle_adjuster.Solve(&reconstruction, &camera_rigs));
+  std::unique_ptr<BundleAdjuster> bundle_adjuster = CreateDefaultBundleAdjuster(
+      *options.bundle_adjustment, std::move(config), reconstruction);
+  if (bundle_adjuster->Solve().termination_type == ceres::FAILURE) {
+    LOG(ERROR) << "Failed to solve rig bundle adjustment";
+    return EXIT_FAILURE;
+  }
 
   reconstruction.Write(output_path);
 
