@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -34,9 +34,10 @@
 #include "colmap/feature/sift.h"
 #include "colmap/feature/utils.h"
 #include "colmap/optim/random_sampler.h"
+#include "colmap/retrieval/resources.h"
 #include "colmap/retrieval/visual_index.h"
 #include "colmap/scene/database.h"
-#include "colmap/util/misc.h"
+#include "colmap/util/file.h"
 #include "colmap/util/opengl_utils.h"
 
 #include <numeric>
@@ -47,14 +48,14 @@ namespace {
 // Loads descriptors for training from the database. Loads all descriptors from
 // the database if max_num_images < 0, otherwise the descriptors of a random
 // subset of images are selected.
-FeatureDescriptors LoadRandomDatabaseDescriptors(
+retrieval::VisualIndex::Descriptors LoadRandomDatabaseDescriptors(
     const std::string& database_path, const int max_num_images) {
   Database database(database_path);
   DatabaseTransaction database_transaction(&database);
 
   const std::vector<Image> images = database.ReadAllImages();
 
-  FeatureDescriptors descriptors;
+  retrieval::VisualIndex::Descriptors descriptors;
 
   std::vector<size_t> image_idxs;
   size_t num_descriptors = 0;
@@ -83,7 +84,7 @@ FeatureDescriptors LoadRandomDatabaseDescriptors(
     const FeatureDescriptors image_descriptors =
         database.ReadDescriptors(image.ImageId());
     descriptors.block(descriptor_row, 0, image_descriptors.rows(), 128) =
-        image_descriptors;
+        image_descriptors.cast<float>();
     descriptor_row += image_descriptors.rows();
   }
 
@@ -103,12 +104,12 @@ std::vector<Image> ReadVocabTreeRetrievalImageList(const std::string& path,
   } else {
     DatabaseTransaction database_transaction(database);
 
-    const auto image_names = ReadTextFileLines(path);
+    const std::vector<std::string> image_names = ReadTextFileLines(path);
     images.reserve(image_names.size());
     for (const auto& image_name : image_names) {
-      const auto image = database->ReadImageWithName(image_name);
+      Image image = database->ReadImageWithName(image_name).value();
       THROW_CHECK_NE(image.ImageId(), kInvalidImageId);
-      images.push_back(image);
+      images.push_back(std::move(image));
     }
   }
   return images;
@@ -117,17 +118,18 @@ std::vector<Image> ReadVocabTreeRetrievalImageList(const std::string& path,
 }  // namespace
 
 int RunVocabTreeBuilder(int argc, char** argv) {
-  std::string vocab_tree_path;
-  retrieval::VisualIndex<>::BuildOptions build_options;
+  std::string vocab_tree_path = kDefaultVocabTreeUri;
+  retrieval::VisualIndex::BuildOptions build_options;
   int max_num_images = -1;
 
   OptionManager options;
   options.AddDatabaseOptions();
   options.AddRequiredOption("vocab_tree_path", &vocab_tree_path);
   options.AddDefaultOption("num_visual_words", &build_options.num_visual_words);
-  options.AddDefaultOption("num_checks", &build_options.num_checks);
-  options.AddDefaultOption("branching", &build_options.branching);
   options.AddDefaultOption("num_iterations", &build_options.num_iterations);
+  options.AddDefaultOption("num_checks", &build_options.num_checks);
+  options.AddDefaultOption("num_threads", &build_options.num_threads);
+  options.AddDefaultOption("num_rounds", &build_options.num_rounds);
   options.AddDefaultOption("max_num_images", &max_num_images);
   options.Parse(argc, argv);
 
@@ -137,26 +139,27 @@ int RunVocabTreeBuilder(int argc, char** argv) {
   LOG(INFO) << "=> Loaded a total of " << descriptors.rows() << " descriptors";
   THROW_CHECK_GT(descriptors.size(), 0);
 
-  retrieval::VisualIndex<> visual_index;
+  auto visual_index = retrieval::VisualIndex::Create();
 
   LOG(INFO) << "Building index for visual words...";
   // NOLINTNEXTLINE(clang-analyzer-optin.cplusplus.VirtualCall)
-  visual_index.Build(build_options, descriptors);
+  visual_index->Build(build_options, descriptors);
   LOG(INFO) << "=> Quantized descriptor space using "
-            << visual_index.NumVisualWords() << " visual words";
+            << visual_index->NumVisualWords() << " visual words";
 
   LOG(INFO) << "Saving index to file...";
-  visual_index.Write(vocab_tree_path);
+  visual_index->Write(vocab_tree_path);
 
   return EXIT_SUCCESS;
 }
 
 int RunVocabTreeRetriever(int argc, char** argv) {
-  std::string vocab_tree_path;
+  std::string vocab_tree_path = kDefaultVocabTreeUri;
   std::string database_image_list_path;
   std::string query_image_list_path;
   std::string output_index_path;
-  retrieval::VisualIndex<>::QueryOptions query_options;
+  retrieval::VisualIndex::QueryOptions query_options;
+  retrieval::VisualIndex::IndexOptions index_options;
   int max_num_features = -1;
 
   OptionManager options;
@@ -169,13 +172,15 @@ int RunVocabTreeRetriever(int argc, char** argv) {
   options.AddDefaultOption("num_images", &query_options.max_num_images);
   options.AddDefaultOption("num_neighbors", &query_options.num_neighbors);
   options.AddDefaultOption("num_checks", &query_options.num_checks);
+  options.AddDefaultOption("num_threads", &query_options.num_threads);
   options.AddDefaultOption("num_images_after_verification",
                            &query_options.num_images_after_verification);
   options.AddDefaultOption("max_num_features", &max_num_features);
   options.Parse(argc, argv);
 
-  retrieval::VisualIndex<> visual_index;
-  visual_index.Read(vocab_tree_path);
+  index_options.num_threads = query_options.num_threads;
+
+  auto visual_index = retrieval::VisualIndex::Read(vocab_tree_path);
 
   Database database(*options.database_path);
 
@@ -198,31 +203,33 @@ int RunVocabTreeRetriever(int argc, char** argv) {
                      "Indexing image [%d/%d]", i + 1, database_images.size())
               << std::flush;
 
-    if (visual_index.ImageIndexed(database_images[i].ImageId())) {
+    if (visual_index->IsImageIndexed(database_images[i].ImageId())) {
       continue;
     }
 
-    auto keypoints = database.ReadKeypoints(database_images[i].ImageId());
-    auto descriptors = database.ReadDescriptors(database_images[i].ImageId());
+    FeatureKeypoints keypoints =
+        database.ReadKeypoints(database_images[i].ImageId());
+    FeatureDescriptors descriptors =
+        database.ReadDescriptors(database_images[i].ImageId());
     if (max_num_features > 0 && descriptors.rows() > max_num_features) {
       ExtractTopScaleFeatures(&keypoints, &descriptors, max_num_features);
     }
 
-    visual_index.Add(retrieval::VisualIndex<>::IndexOptions(),
-                     database_images[i].ImageId(),
-                     keypoints,
-                     descriptors);
+    visual_index->Add(index_options,
+                      database_images[i].ImageId(),
+                      keypoints,
+                      descriptors.cast<float>());
 
     LOG(INFO) << StringPrintf(" in %.3fs", timer.ElapsedSeconds());
   }
 
   // Compute the TF-IDF weights, etc.
-  visual_index.Prepare();
+  visual_index->Prepare();
 
   // Optionally save the indexing data for the database images (as well as the
   // original vocabulary tree data) to speed up future indexing.
   if (!output_index_path.empty()) {
-    visual_index.Write(output_index_path);
+    visual_index->Write(output_index_path);
   }
 
   if (query_images.empty()) {
@@ -256,7 +263,8 @@ int RunVocabTreeRetriever(int argc, char** argv) {
     }
 
     std::vector<retrieval::ImageScore> image_scores;
-    visual_index.Query(query_options, keypoints, descriptors, &image_scores);
+    visual_index->Query(
+        query_options, keypoints, descriptors.cast<float>(), &image_scores);
 
     LOG(INFO) << StringPrintf(" in %.3fs", timer.ElapsedSeconds());
     for (const auto& image_score : image_scores) {
