@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -31,6 +31,7 @@
 
 #include "colmap/feature/utils.h"
 #include "colmap/geometry/gps.h"
+#include "colmap/util/file.h"
 #include "colmap/util/logging.h"
 #include "colmap/util/misc.h"
 #include "colmap/util/timer.h"
@@ -41,6 +42,9 @@
 #include <unordered_set>
 #include <vector>
 
+#include <faiss/IndexFlat.h>
+#include <omp.h>
+
 namespace colmap {
 namespace {
 
@@ -123,6 +127,7 @@ VocabTreeMatchingOptions SequentialMatchingOptions::VocabTreeOptions() const {
       loop_detection_num_images_after_verification;
   options.max_num_features = loop_detection_max_num_features;
   options.vocab_tree_path = vocab_tree_path;
+  options.num_threads = num_threads;
   return options;
 }
 
@@ -175,9 +180,8 @@ ExhaustivePairGenerator::ExhaustivePairGenerator(
     const std::shared_ptr<Database>& database)
     : ExhaustivePairGenerator(
           options,
-          std::make_shared<FeatureMatcherCache>(CacheSize(options),
-                                                THROW_CHECK_NOTNULL(database),
-                                                /*do_setup=*/true)) {}
+          std::make_shared<FeatureMatcherCache>(
+              options.CacheSize(), THROW_CHECK_NOTNULL(database))) {}
 
 void ExhaustivePairGenerator::Reset() {
   start_idx1_ = 0;
@@ -225,17 +229,17 @@ std::vector<std::pair<image_t, image_t>> ExhaustivePairGenerator::Next() {
 
 VocabTreePairGenerator::VocabTreePairGenerator(
     const VocabTreeMatchingOptions& options,
-    std::shared_ptr<FeatureMatcherCache> cache,
+    const std::shared_ptr<FeatureMatcherCache>& cache,
     const std::vector<image_t>& query_image_ids)
     : options_(options),
-      cache_(std::move(THROW_CHECK_NOTNULL(cache))),
-      thread_pool(options_.num_threads),
-      queue(options_.num_threads) {
+      cache_(THROW_CHECK_NOTNULL(cache)),
+      thread_pool_(options_.num_threads),
+      queue_(options_.num_threads) {
   THROW_CHECK(options.Check());
   LOG(INFO) << "Generating image pairs with vocabulary tree...";
 
   // Read the pre-trained vocabulary tree from disk.
-  visual_index_.Read(options_.vocab_tree_path);
+  visual_index_ = retrieval::VisualIndex::Read(options_.vocab_tree_path);
 
   const std::vector<image_t> all_image_ids = cache_->GetImageIds();
   if (query_image_ids.size() > 0) {
@@ -272,6 +276,9 @@ VocabTreePairGenerator::VocabTreePairGenerator(
 
   IndexImages(all_image_ids);
 
+  // Since we parallelize over the query images, there is no need to parallelize
+  // the nearest neighbor search over the query descriptors.
+  query_options_.num_threads = 1;
   query_options_.max_num_images = options_.num_images;
   query_options_.num_neighbors = options_.num_nearest_neighbors;
   query_options_.num_checks = options_.num_checks;
@@ -285,9 +292,8 @@ VocabTreePairGenerator::VocabTreePairGenerator(
     const std::vector<image_t>& query_image_ids)
     : VocabTreePairGenerator(
           options,
-          std::make_shared<FeatureMatcherCache>(CacheSize(options),
-                                                THROW_CHECK_NOTNULL(database),
-                                                /*do_setup=*/true),
+          std::make_shared<FeatureMatcherCache>(options.CacheSize(),
+                                                THROW_CHECK_NOTNULL(database)),
           query_image_ids) {}
 
 void VocabTreePairGenerator::Reset() {
@@ -302,15 +308,15 @@ bool VocabTreePairGenerator::HasFinished() const {
 std::vector<std::pair<image_t, image_t>> VocabTreePairGenerator::Next() {
   image_pairs_.clear();
   if (HasFinished()) {
-    return image_pairs_;
+    return {};
   }
   if (query_idx_ == 0) {
     // Initially, make all retrieval threads busy and continue with the
     // matching.
     const size_t init_num_tasks =
-        std::min(query_image_ids_.size(), 2 * thread_pool.NumThreads());
+        std::min(query_image_ids_.size(), 2 * thread_pool_.NumThreads());
     for (; query_idx_ < init_num_tasks; ++query_idx_) {
-      thread_pool.AddTask(
+      thread_pool_.AddTask(
           &VocabTreePairGenerator::Query, this, query_image_ids_[query_idx_]);
     }
   }
@@ -320,12 +326,12 @@ std::vector<std::pair<image_t, image_t>> VocabTreePairGenerator::Next() {
 
   // Push the next image to the retrieval queue.
   if (query_idx_ < query_image_ids_.size()) {
-    thread_pool.AddTask(
+    thread_pool_.AddTask(
         &VocabTreePairGenerator::Query, this, query_image_ids_[query_idx_++]);
   }
 
   // Pop the next results from the retrieval queue.
-  auto retrieval = queue.Pop();
+  auto retrieval = queue_.Pop();
   THROW_CHECK(retrieval.IsValid());
 
   const auto& image_id = retrieval.Data().image_id;
@@ -342,9 +348,14 @@ std::vector<std::pair<image_t, image_t>> VocabTreePairGenerator::Next() {
 
 void VocabTreePairGenerator::IndexImages(
     const std::vector<image_t>& image_ids) {
-  retrieval::VisualIndex<>::IndexOptions index_options;
-  index_options.num_threads = options_.num_threads;
+  retrieval::VisualIndex::IndexOptions index_options;
+  // We only assign each feature to a single visual word in the indexing phase.
+  // During the query phase, we check for overlap in possibly multiple nearest
+  // neighbor visual words. We could do it symmetrically but experiments showed
+  // only marginal improvements that do not justify the memory/compute increase.
+  index_options.num_neighbors = 1;
   index_options.num_checks = options_.num_checks;
+  index_options.num_threads = options_.num_threads;
 
   for (size_t i = 0; i < image_ids.size(); ++i) {
     Timer timer;
@@ -358,12 +369,13 @@ void VocabTreePairGenerator::IndexImages(
       ExtractTopScaleFeatures(
           &keypoints, &descriptors, options_.max_num_features);
     }
-    visual_index_.Add(index_options, image_ids[i], keypoints, descriptors);
+    visual_index_->Add(
+        index_options, image_ids[i], keypoints, descriptors.cast<float>());
     LOG(INFO) << StringPrintf(" in %.3fs", timer.ElapsedSeconds());
   }
 
   // Compute the TF-IDF weights, etc.
-  visual_index_.Prepare();
+  visual_index_->Prepare();
 }
 
 void VocabTreePairGenerator::Query(const image_t image_id) {
@@ -377,16 +389,18 @@ void VocabTreePairGenerator::Query(const image_t image_id) {
 
   Retrieval retrieval;
   retrieval.image_id = image_id;
-  visual_index_.Query(
-      query_options_, keypoints, descriptors, &retrieval.image_scores);
+  visual_index_->Query(query_options_,
+                       keypoints,
+                       descriptors.cast<float>(),
+                       &retrieval.image_scores);
 
-  THROW_CHECK(queue.Push(std::move(retrieval)));
+  THROW_CHECK(queue_.Push(std::move(retrieval)));
 }
 
 SequentialPairGenerator::SequentialPairGenerator(
     const SequentialMatchingOptions& options,
-    std::shared_ptr<FeatureMatcherCache> cache)
-    : options_(options), cache_(std::move(THROW_CHECK_NOTNULL(cache))) {
+    const std::shared_ptr<FeatureMatcherCache>& cache)
+    : options_(options), cache_(THROW_CHECK_NOTNULL(cache)) {
   THROW_CHECK(options.Check());
   LOG(INFO) << "Generating sequential image pairs...";
   image_ids_ = GetOrderedImageIds();
@@ -401,6 +415,20 @@ SequentialPairGenerator::SequentialPairGenerator(
     vocab_tree_pair_generator_ = std::make_unique<VocabTreePairGenerator>(
         options_.VocabTreeOptions(), cache_, query_image_ids);
   }
+
+  if (options_.expand_rig_images) {
+    const std::vector<frame_t> frame_ids = cache_->GetFrameIds();
+    frame_to_image_ids_.reserve(frame_ids.size());
+    image_to_frame_ids_.reserve(image_ids_.size());
+    for (const frame_t frame_id : frame_ids) {
+      const Frame& frame = cache_->GetFrame(frame_id);
+      auto& frame_image_ids = frame_to_image_ids_[frame_id];
+      for (const data_t& data_id : frame.ImageIds()) {
+        frame_image_ids.push_back(data_id.id);
+        image_to_frame_ids_[data_id.id] = frame_id;
+      }
+    }
+  }
 }
 
 SequentialPairGenerator::SequentialPairGenerator(
@@ -408,9 +436,8 @@ SequentialPairGenerator::SequentialPairGenerator(
     const std::shared_ptr<Database>& database)
     : SequentialPairGenerator(
           options,
-          std::make_shared<FeatureMatcherCache>(CacheSize(options),
-                                                THROW_CHECK_NOTNULL(database),
-                                                /*do_setup=*/true)) {}
+          std::make_shared<FeatureMatcherCache>(
+              options.CacheSize(), THROW_CHECK_NOTNULL(database))) {}
 
 void SequentialPairGenerator::Reset() {
   image_idx_ = 0;
@@ -437,19 +464,55 @@ std::vector<std::pair<image_t, image_t>> SequentialPairGenerator::Next() {
       "Matching image [%d/%d]", image_idx_ + 1, image_ids_.size());
 
   const auto image_id1 = image_ids_.at(image_idx_);
-  for (int i = 0; i < options_.overlap; ++i) {
-    const size_t image_idx_2 = image_idx_ + i;
-    if (image_idx_2 < image_ids_.size()) {
-      image_pairs_.emplace_back(image_id1, image_ids_.at(image_idx_2));
-      if (options_.quadratic_overlap) {
-        const size_t image_idx_2_quadratic = image_idx_ + (1ull << i);
-        if (image_idx_2_quadratic < image_ids_.size()) {
-          image_pairs_.emplace_back(image_id1,
-                                    image_ids_.at(image_idx_2_quadratic));
+
+  // If image is part of a rig, then pair the other images in the same frame.
+  if (options_.expand_rig_images) {
+    if (const auto frame_id1_it = image_to_frame_ids_.find(image_id1);
+        frame_id1_it != image_to_frame_ids_.end()) {
+      for (const image_t frame_image_id2 :
+           frame_to_image_ids_.at(frame_id1_it->second)) {
+        if (image_id1 != frame_image_id2) {
+          image_pairs_.emplace_back(image_id1, frame_image_id2);
         }
       }
+    }
+  }
+
+  auto MaybeExpandRigImages = [this](image_t image_id1, image_t image_id2) {
+    if (!options_.expand_rig_images) {
+      return;
+    }
+    const auto frame_id2_it = image_to_frame_ids_.find(image_id2);
+    if (frame_id2_it != image_to_frame_ids_.end()) {
+      // Pair with all images in second frame.
+      for (const image_t frame_image_id2 :
+           frame_to_image_ids_.at(frame_id2_it->second)) {
+        if (image_id1 != frame_image_id2 && image_id2 != frame_image_id2) {
+          image_pairs_.emplace_back(image_id1, frame_image_id2);
+        }
+      }
+    }
+  };
+
+  for (int i = 0; i < options_.overlap; ++i) {
+    if (options_.quadratic_overlap) {
+      const size_t image_idx_2_quadratic = image_idx_ + (1ull << i);
+      if (image_idx_2_quadratic < image_ids_.size()) {
+        const image_t image_id2 = image_ids_.at(image_idx_2_quadratic);
+        image_pairs_.emplace_back(image_id1, image_id2);
+        MaybeExpandRigImages(image_id1, image_id2);
+      } else {
+        break;
+      }
     } else {
-      break;
+      const size_t image_idx_2 = image_idx_ + i + 1;
+      if (image_idx_2 < image_ids_.size()) {
+        const image_t image_id2 = image_ids_.at(image_idx_2);
+        image_pairs_.emplace_back(image_id1, image_id2);
+        MaybeExpandRigImages(image_id1, image_id2);
+      } else {
+        break;
+      }
     }
   }
   ++image_idx_;
@@ -483,7 +546,7 @@ std::vector<image_t> SequentialPairGenerator::GetOrderedImageIds() const {
 SpatialPairGenerator::SpatialPairGenerator(
     const SpatialMatchingOptions& options,
     const std::shared_ptr<FeatureMatcherCache>& cache)
-    : options_(options), image_ids_(cache->GetImageIds()) {
+    : options_(options), image_ids_(THROW_CHECK_NOTNULL(cache)->GetImageIds()) {
   LOG(INFO) << "Generating spatial image pairs...";
   THROW_CHECK(options.Check());
 
@@ -491,12 +554,11 @@ SpatialPairGenerator::SpatialPairGenerator(
   timer.Start();
   LOG(INFO) << "Indexing images...";
 
-  Eigen::Matrix<float, Eigen::Dynamic, 3, Eigen::RowMajor> location_matrix =
-      ReadLocationData(*cache);
-  size_t num_locations = location_idxs_.size();
+  Eigen::RowMajorMatrixXf position_matrix = ReadPositionPriorData(*cache);
+  const size_t num_positions = position_idxs_.size();
 
   LOG(INFO) << StringPrintf(" in %.3fs", timer.ElapsedSeconds());
-  if (num_locations == 0) {
+  if (num_positions == 0) {
     LOG(INFO) << "=> No images with location data.";
     return;
   }
@@ -504,38 +566,27 @@ SpatialPairGenerator::SpatialPairGenerator(
   timer.Restart();
   LOG(INFO) << "Building search index...";
 
-  flann::Matrix<float> locations(
-      location_matrix.data(), num_locations, location_matrix.cols());
-
-  flann::LinearIndexParams index_params;
-  flann::LinearIndex<flann::L2<float>> search_index(index_params);
-  search_index.buildIndex(locations);
+  faiss::IndexFlatL2 search_index(/*d=*/3);
+  search_index.add(position_matrix.rows(), position_matrix.data());
 
   LOG(INFO) << StringPrintf(" in %.3fs", timer.ElapsedSeconds());
 
   timer.Restart();
   LOG(INFO) << "Searching for nearest neighbors...";
 
-  const int knn = std::min<int>(options_.max_num_neighbors, num_locations);
-  image_pairs_.reserve(knn);
+  knn_ = std::min<int>(options_.max_num_neighbors + 1, num_positions);
+  image_pairs_.reserve(knn_);
 
-  index_matrix_.resize(num_locations, knn);
-  flann::Matrix<size_t> indices(index_matrix_.data(), num_locations, knn);
+  index_matrix_.resize(num_positions, knn_);
+  distance_matrix_.resize(num_positions, knn_);
 
-  distance_matrix_.resize(num_locations, knn);
-  flann::Matrix<float> distances(distance_matrix_.data(), num_locations, knn);
+  omp_set_num_threads(GetEffectiveNumThreads(options_.num_threads));
 
-  flann::SearchParams search_params(flann::FLANN_CHECKS_AUTOTUNED);
-  if (options_.num_threads == ThreadPool::kMaxNumThreads) {
-    search_params.cores = std::thread::hardware_concurrency();
-  } else {
-    search_params.cores = options_.num_threads;
-  }
-  if (search_params.cores <= 0) {
-    search_params.cores = 1;
-  }
-
-  search_index.knnSearch(locations, indices, distances, knn, search_params);
+  search_index.search(position_matrix.rows(),
+                      position_matrix.data(),
+                      knn_,
+                      distance_matrix_.data(),
+                      index_matrix_.data());
 
   LOG(INFO) << StringPrintf(" in %.3fs", timer.ElapsedSeconds());
 }
@@ -545,14 +596,13 @@ SpatialPairGenerator::SpatialPairGenerator(
     const std::shared_ptr<Database>& database)
     : SpatialPairGenerator(
           options,
-          std::make_shared<FeatureMatcherCache>(CacheSize(options),
-                                                THROW_CHECK_NOTNULL(database),
-                                                /*do_setup=*/true)) {}
+          std::make_shared<FeatureMatcherCache>(
+              options.CacheSize(), THROW_CHECK_NOTNULL(database))) {}
 
 void SpatialPairGenerator::Reset() { current_idx_ = 0; }
 
 bool SpatialPairGenerator::HasFinished() const {
-  return current_idx_ >= location_idxs_.size();
+  return current_idx_ >= position_idxs_.size();
 }
 
 std::vector<std::pair<image_t, image_t>> SpatialPairGenerator::Next() {
@@ -562,14 +612,12 @@ std::vector<std::pair<image_t, image_t>> SpatialPairGenerator::Next() {
   }
 
   LOG(INFO) << StringPrintf(
-      "Matching image [%d/%d]", current_idx_ + 1, location_idxs_.size());
-  const int knn =
-      std::min<int>(options_.max_num_neighbors, location_idxs_.size());
+      "Matching image [%d/%d]", current_idx_ + 1, position_idxs_.size());
   const float max_distance =
       static_cast<float>(options_.max_distance * options_.max_distance);
-  for (int j = 0; j < knn; ++j) {
+  for (int j = 0; j < knn_; ++j) {
     // Check if query equals result.
-    if (index_matrix_(current_idx_, j) == current_idx_) {
+    if (index_matrix_(current_idx_, j) == static_cast<int>(current_idx_)) {
       continue;
     }
 
@@ -578,8 +626,8 @@ std::vector<std::pair<image_t, image_t>> SpatialPairGenerator::Next() {
       break;
     }
 
-    const image_t image_id = image_ids_.at(location_idxs_[current_idx_]);
-    const size_t nn_idx = location_idxs_.at(index_matrix_(current_idx_, j));
+    const image_t image_id = image_ids_.at(position_idxs_[current_idx_]);
+    const size_t nn_idx = position_idxs_.at(index_matrix_(current_idx_, j));
     const image_t nn_image_id = image_ids_.at(nn_idx);
     image_pairs_.emplace_back(image_id, nn_image_id);
   }
@@ -587,68 +635,164 @@ std::vector<std::pair<image_t, image_t>> SpatialPairGenerator::Next() {
   return image_pairs_;
 }
 
-Eigen::Matrix<float, Eigen::Dynamic, 3, Eigen::RowMajor>
-SpatialPairGenerator::ReadLocationData(const FeatureMatcherCache& cache) {
+Eigen::RowMajorMatrixXf SpatialPairGenerator::ReadPositionPriorData(
+    FeatureMatcherCache& cache) {
   GPSTransform gps_transform;
   std::vector<Eigen::Vector3d> ells(1);
 
-  size_t num_locations = 0;
-  location_idxs_.clear();
-  location_idxs_.reserve(image_ids_.size());
-  Eigen::Matrix<float, Eigen::Dynamic, 3, Eigen::RowMajor> location_matrix(
-      image_ids_.size(), 3);
+  Eigen::RowMajorMatrixXd position_matrix(image_ids_.size(), 3);
+  position_idxs_.clear();
+  position_idxs_.reserve(image_ids_.size());
 
   for (size_t i = 0; i < image_ids_.size(); ++i) {
-    if (!cache.ExistsPosePrior(image_ids_[i])) {
-      continue;
-    }
-    const auto& pose_prior = cache.GetPosePrior(image_ids_[i]);
-    const Eigen::Vector3d& translation_prior = pose_prior.position;
-    if ((translation_prior(0) == 0 && translation_prior(1) == 0 &&
-         options_.ignore_z) ||
-        (translation_prior(0) == 0 && translation_prior(1) == 0 &&
-         translation_prior(2) == 0 && !options_.ignore_z)) {
+    const PosePrior* pose_prior = cache.GetPosePriorOrNull(image_ids_[i]);
+    if (pose_prior == nullptr) {
       continue;
     }
 
-    location_idxs_.push_back(i);
+    const Eigen::Vector3d& position_prior = pose_prior->position;
+    if ((position_prior(0) == 0 && position_prior(1) == 0 &&
+         position_prior(2) == 0) ||
+        (options_.ignore_z && position_prior(0) == 0 &&
+         position_prior(1) == 0)) {
+      continue;
+    }
 
-    switch (pose_prior.coordinate_system) {
+    const size_t position_idx = position_idxs_.size();
+    position_idxs_.push_back(i);
+
+    switch (pose_prior->coordinate_system) {
       case PosePrior::CoordinateSystem::WGS84: {
-        ells[0](0) = translation_prior(0);
-        ells[0](1) = translation_prior(1);
-        ells[0](2) = options_.ignore_z ? 0 : translation_prior(2);
+        ells[0](0) = position_prior(0);
+        ells[0](1) = position_prior(1);
+        ells[0](2) = options_.ignore_z ? 0 : position_prior(2);
 
-        const auto xyzs = gps_transform.EllToXYZ(ells);
-        location_matrix(num_locations, 0) = static_cast<float>(xyzs[0](0));
-        location_matrix(num_locations, 1) = static_cast<float>(xyzs[0](1));
-        location_matrix(num_locations, 2) = static_cast<float>(xyzs[0](2));
+        const std::vector<Eigen::Vector3d> xyzs =
+            gps_transform.EllipsoidToECEF(ells);
+        position_matrix(position_idx, 0) = xyzs[0](0);
+        position_matrix(position_idx, 1) = xyzs[0](1);
+        position_matrix(position_idx, 2) = xyzs[0](2);
       } break;
       case PosePrior::CoordinateSystem::UNDEFINED:
-        LOG(INFO) << "Unknown coordinate system for image " << image_ids_[i]
-                  << ", assuming cartesian.";
-      case PosePrior::CoordinateSystem::CARTESIAN:
       default:
-        location_matrix(num_locations, 0) =
-            static_cast<float>(translation_prior(0));
-        location_matrix(num_locations, 1) =
-            static_cast<float>(translation_prior(1));
-        location_matrix(num_locations, 2) =
-            static_cast<float>(options_.ignore_z ? 0 : translation_prior(2));
+        LOG(WARNING) << "Unknown coordinate system for image " << image_ids_[i]
+                     << ", assuming cartesian.";
+      case PosePrior::CoordinateSystem::CARTESIAN:
+        position_matrix(position_idx, 0) = position_prior(0);
+        position_matrix(position_idx, 1) = position_prior(1);
+        position_matrix(position_idx, 2) =
+            options_.ignore_z ? 0 : position_prior(2);
     }
-
-    num_locations += 1;
   }
-  return location_matrix;
+
+  // Subtract the mean coordinate before casting to float for better numerical
+  // precision when dealing with large coordinates (e.g. GPS). For even better
+  // precision, we could also rescale the coordinates.
+  position_matrix.rowwise() -= position_matrix.colwise().mean();
+  return position_matrix.topRows(position_idxs_.size()).cast<float>();
+}
+
+TransitivePairGenerator::TransitivePairGenerator(
+    const TransitiveMatchingOptions& options,
+    const std::shared_ptr<FeatureMatcherCache>& cache)
+    : options_(options), cache_(cache) {
+  THROW_CHECK(options.Check());
+}
+
+TransitivePairGenerator::TransitivePairGenerator(
+    const TransitiveMatchingOptions& options,
+    const std::shared_ptr<Database>& database)
+    : TransitivePairGenerator(
+          options,
+          std::make_shared<FeatureMatcherCache>(
+              options.CacheSize(), THROW_CHECK_NOTNULL(database))) {}
+
+void TransitivePairGenerator::Reset() {
+  current_iteration_ = 0;
+  current_batch_idx_ = 0;
+  image_pairs_.clear();
+  image_pair_ids_.clear();
+}
+
+bool TransitivePairGenerator::HasFinished() const {
+  return current_iteration_ >= options_.num_iterations && image_pairs_.empty();
+}
+
+std::vector<std::pair<image_t, image_t>> TransitivePairGenerator::Next() {
+  if (!image_pairs_.empty()) {
+    current_batch_idx_++;
+    std::vector<std::pair<image_t, image_t>> batch;
+    while (!image_pairs_.empty() &&
+           static_cast<int>(batch.size()) < options_.batch_size) {
+      batch.push_back(image_pairs_.back());
+      image_pairs_.pop_back();
+    }
+    LOG(INFO) << StringPrintf(
+        "Matching batch [%d/%d]", current_batch_idx_, current_num_batches_);
+    return batch;
+  }
+
+  if (current_iteration_ >= options_.num_iterations) {
+    return {};
+  }
+
+  current_batch_idx_ = 0;
+  current_num_batches_ = 0;
+  current_iteration_++;
+
+  LOG(INFO) << StringPrintf(
+      "Iteration [%d/%d]", current_iteration_, options_.num_iterations);
+
+  std::vector<std::pair<image_pair_t, int>> existing_pair_ids_and_num_inliers;
+  cache_->AccessDatabase(
+      [&existing_pair_ids_and_num_inliers](Database& database) {
+        existing_pair_ids_and_num_inliers =
+            database.ReadTwoViewGeometryNumInliers();
+      });
+
+  std::map<image_t, std::vector<image_t>> adjacency;
+  for (const auto& [pair_id, _] : existing_pair_ids_and_num_inliers) {
+    const auto [image_id1, image_id2] = Database::PairIdToImagePair(pair_id);
+    adjacency[image_id1].push_back(image_id2);
+    adjacency[image_id2].push_back(image_id1);
+    image_pair_ids_.insert(pair_id);
+  }
+
+  for (const auto& image : adjacency) {
+    const auto image_id1 = image.first;
+    for (const auto& image_id2 : image.second) {
+      const auto it = adjacency.find(image_id2);
+      if (it == adjacency.end()) {
+        continue;
+      }
+      for (const auto& image_id3 : it->second) {
+        if (image_id1 == image_id3) {
+          continue;
+        }
+        const auto image_pair_id =
+            Database::ImagePairToPairId(image_id1, image_id3);
+        if (image_pair_ids_.count(image_pair_id) != 0) {
+          continue;
+        }
+        image_pairs_.emplace_back(std::minmax(image_id1, image_id3));
+        image_pair_ids_.insert(image_pair_id);
+      }
+    }
+  }
+
+  current_num_batches_ =
+      std::ceil(static_cast<double>(image_pairs_.size()) / options_.batch_size);
+
+  return Next();
 }
 
 ImportedPairGenerator::ImportedPairGenerator(
     const ImagePairsMatchingOptions& options,
     const std::shared_ptr<FeatureMatcherCache>& cache)
     : options_(options) {
-  LOG(INFO) << "Importing image pairs...";
   THROW_CHECK(options.Check());
 
+  LOG(INFO) << "Importing image pairs...";
   const std::vector<image_t> image_ids = cache->GetImageIds();
   std::unordered_map<std::string, image_t> image_name_to_image_id;
   image_name_to_image_id.reserve(image_ids.size());
@@ -666,9 +810,8 @@ ImportedPairGenerator::ImportedPairGenerator(
     const std::shared_ptr<Database>& database)
     : ImportedPairGenerator(
           options,
-          std::make_shared<FeatureMatcherCache>(CacheSize(options),
-                                                THROW_CHECK_NOTNULL(database),
-                                                /*do_setup=*/true)) {}
+          std::make_shared<FeatureMatcherCache>(
+              options.CacheSize(), THROW_CHECK_NOTNULL(database))) {}
 
 void ImportedPairGenerator::Reset() { pair_idx_ = 0; }
 
