@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -135,6 +135,39 @@ TEST(ExtractCovariantAffineSiftFeaturesCPU, Nominal) {
   FeatureDescriptors descriptors;
   EXPECT_TRUE(extractor->Extract(bitmap, &keypoints, &descriptors));
 
+  EXPECT_EQ(keypoints.size(), 22);
+  for (size_t i = 0; i < keypoints.size(); ++i) {
+    EXPECT_GE(keypoints[i].x, 0);
+    EXPECT_GE(keypoints[i].y, 0);
+    EXPECT_LE(keypoints[i].x, bitmap.Width());
+    EXPECT_LE(keypoints[i].y, bitmap.Height());
+    EXPECT_GT(keypoints[i].ComputeScale(), 0);
+    EXPECT_GT(keypoints[i].ComputeOrientation(), -M_PI);
+    EXPECT_LT(keypoints[i].ComputeOrientation(), M_PI);
+  }
+
+  EXPECT_EQ(descriptors.rows(), 22);
+  for (FeatureDescriptors::Index i = 0; i < descriptors.rows(); ++i) {
+    EXPECT_LT(std::abs(descriptors.row(i).cast<float>().norm() - 512), 1);
+  }
+}
+
+TEST(ExtractCovariantAffineSiftFeaturesCPU, Upright) {
+  Bitmap bitmap;
+  CreateImageWithSquare(256, &bitmap);
+
+  SiftExtractionOptions options;
+  options.use_gpu = false;
+  options.estimate_affine_shape = true;
+  options.upright = true;
+  options.domain_size_pooling = false;
+  options.force_covariant_extractor = false;
+  auto extractor = CreateSiftFeatureExtractor(options);
+
+  FeatureKeypoints keypoints;
+  FeatureDescriptors descriptors;
+  EXPECT_TRUE(extractor->Extract(bitmap, &keypoints, &descriptors));
+
   EXPECT_EQ(keypoints.size(), 10);
   for (size_t i = 0; i < keypoints.size(); ++i) {
     EXPECT_GE(keypoints[i].x, 0);
@@ -199,7 +232,7 @@ TEST(ExtractCovariantAffineDSPSiftFeaturesCPU, Nominal) {
   FeatureDescriptors descriptors;
   EXPECT_TRUE(extractor->Extract(bitmap, &keypoints, &descriptors));
 
-  EXPECT_EQ(keypoints.size(), 10);
+  EXPECT_EQ(keypoints.size(), 22);
   for (size_t i = 0; i < keypoints.size(); ++i) {
     EXPECT_GE(keypoints[i].x, 0);
     EXPECT_GE(keypoints[i].y, 0);
@@ -210,7 +243,7 @@ TEST(ExtractCovariantAffineDSPSiftFeaturesCPU, Nominal) {
     EXPECT_LT(keypoints[i].ComputeOrientation(), M_PI);
   }
 
-  EXPECT_EQ(descriptors.rows(), 10);
+  EXPECT_EQ(descriptors.rows(), 22);
   for (FeatureDescriptors::Index i = 0; i < descriptors.rows(); ++i) {
     EXPECT_LT(std::abs(descriptors.row(i).cast<float>().norm() - 512), 1);
   }
@@ -266,14 +299,18 @@ TEST(ExtractSiftFeaturesGPU, Nominal) {
 
 FeatureDescriptors CreateRandomFeatureDescriptors(const size_t num_features) {
   SetPRNGSeed(0);
-  FeatureDescriptorsFloat descriptors(num_features, 128);
+  FeatureDescriptorsFloat descriptors_float =
+      FeatureDescriptorsFloat::Zero(num_features, 128);
+  std::vector<int> dims(128);
+  std::iota(dims.begin(), dims.end(), 0);
   for (size_t i = 0; i < num_features; ++i) {
-    for (size_t j = 0; j < 128; ++j) {
-      descriptors(i, j) = std::pow(RandomUniformReal(0.0f, 1.0f), 2);
+    std::shuffle(dims.begin(), dims.end(), *PRNG);
+    for (size_t j = 0; j < 10; ++j) {
+      descriptors_float(i, dims[j]) = 1.0f;
     }
   }
-  L2NormalizeFeatureDescriptors(&descriptors);
-  return FeatureDescriptorsToUnsignedByte(descriptors);
+  L2NormalizeFeatureDescriptors(&descriptors_float);
+  return FeatureDescriptorsToUnsignedByte(descriptors_float);
 }
 
 void CheckEqualMatches(const FeatureMatches& matches1,
@@ -317,96 +354,114 @@ TEST(CreateSiftGPUMatcherCUDA, Nominal) {
 #endif
 }
 
-TEST(SiftCPUFeatureMatcher, Nominal) {
-  const auto empty_descriptors = std::make_shared<FeatureDescriptors>(0, 128);
-  const auto descriptors1 =
-      std::make_shared<FeatureDescriptors>(CreateRandomFeatureDescriptors(2));
-  const auto descriptors2 =
-      std::make_shared<FeatureDescriptors>(descriptors1->colwise().reverse());
+struct FeatureDescriptorIndexCacheHelper {
+  explicit FeatureDescriptorIndexCacheHelper(
+      const std::vector<FeatureMatcher::Image>& images)
+      : index_cache(100, [this](const image_t image_id) {
+          auto index = FeatureDescriptorIndex::Create();
+          index->Build(this->image_descriptors_.at(image_id)->cast<float>());
+          return index;
+        }) {
+    for (const auto& image : images) {
+      image_descriptors_.emplace(image.image_id, image.descriptors);
+    }
+  }
 
-  FeatureMatches matches;
+  ThreadSafeLRUCache<image_t, FeatureDescriptorIndex> index_cache;
+
+ private:
+  std::map<image_t, std::shared_ptr<const FeatureDescriptors>>
+      image_descriptors_;
+};
+
+TEST(SiftCPUFeatureMatcher, Nominal) {
+  const FeatureMatcher::Image image0 = {
+      0, std::make_shared<FeatureDescriptors>(0, 128)};
+  const FeatureMatcher::Image image1 = {
+      1,
+      std::make_shared<FeatureDescriptors>(CreateRandomFeatureDescriptors(2))};
+  const FeatureMatcher::Image image2 = {
+      2,
+      std::make_shared<FeatureDescriptors>(
+          image1.descriptors->colwise().reverse())};
+
+  FeatureDescriptorIndexCacheHelper index_cache_helper(
+      {image0, image1, image2});
 
   SiftMatchingOptions options;
   options.use_gpu = false;
+  options.cpu_brute_force_matcher = false;
+  options.cpu_descriptor_index_cache = &index_cache_helper.index_cache;
   auto matcher = CreateSiftFeatureMatcher(options);
 
-  matcher->Match(descriptors1, descriptors2, &matches);
+  FeatureMatches matches;
+  matcher->Match(image1, image2, &matches);
   EXPECT_EQ(matches.size(), 2);
   EXPECT_EQ(matches[0].point2D_idx1, 0);
   EXPECT_EQ(matches[0].point2D_idx2, 1);
   EXPECT_EQ(matches[1].point2D_idx1, 1);
   EXPECT_EQ(matches[1].point2D_idx2, 0);
 
-  matcher->Match(descriptors1, nullptr, &matches);
+  matcher->Match(image1, image2, &matches);
   EXPECT_EQ(matches.size(), 2);
   EXPECT_EQ(matches[0].point2D_idx1, 0);
   EXPECT_EQ(matches[0].point2D_idx2, 1);
   EXPECT_EQ(matches[1].point2D_idx1, 1);
   EXPECT_EQ(matches[1].point2D_idx2, 0);
 
-  matcher->Match(nullptr, descriptors2, &matches);
-  EXPECT_EQ(matches.size(), 2);
-  EXPECT_EQ(matches[0].point2D_idx1, 0);
-  EXPECT_EQ(matches[0].point2D_idx2, 1);
-  EXPECT_EQ(matches[1].point2D_idx1, 1);
-  EXPECT_EQ(matches[1].point2D_idx2, 0);
-
-  matcher->Match(nullptr, nullptr, &matches);
-  EXPECT_EQ(matches.size(), 2);
-  EXPECT_EQ(matches[0].point2D_idx1, 0);
-  EXPECT_EQ(matches[0].point2D_idx2, 1);
-  EXPECT_EQ(matches[1].point2D_idx1, 1);
-  EXPECT_EQ(matches[1].point2D_idx2, 0);
-
-  matcher->Match(empty_descriptors, descriptors2, &matches);
+  matcher->Match(image0, image2, &matches);
   EXPECT_EQ(matches.size(), 0);
-  matcher->Match(descriptors1, empty_descriptors, &matches);
+  matcher->Match(image1, image0, &matches);
   EXPECT_EQ(matches.size(), 0);
-  matcher->Match(empty_descriptors, empty_descriptors, &matches);
+  matcher->Match(image0, image0, &matches);
   EXPECT_EQ(matches.size(), 0);
 }
 
-TEST(SiftCPUFeatureMatcherFlannVsBruteForce, Nominal) {
+TEST(SiftCPUFeatureMatcherFaissVsBruteForce, Nominal) {
   SiftMatchingOptions match_options;
   match_options.max_num_matches = 1000;
 
-  auto TestFlannVsBruteForce = [](const SiftMatchingOptions& options,
+  auto TestFaissVsBruteForce = [](const SiftMatchingOptions& options,
                                   const FeatureDescriptors& descriptors1,
                                   const FeatureDescriptors& descriptors2) {
-    const auto descriptors1_ptr =
-        std::make_shared<FeatureDescriptors>(descriptors1);
-    const auto descriptors2_ptr =
-        std::make_shared<FeatureDescriptors>(descriptors2);
+    const FeatureMatcher::Image image0 = {
+        0, std::make_shared<FeatureDescriptors>(0, 128)};
+    const FeatureMatcher::Image image1 = {
+        1, std::make_shared<FeatureDescriptors>(descriptors1)};
+    const FeatureMatcher::Image image2 = {
+        2, std::make_shared<FeatureDescriptors>(descriptors2)};
+
+    FeatureDescriptorIndexCacheHelper index_cache_helper(
+        {image0, image1, image2});
 
     FeatureMatches matches_bf;
-    FeatureMatches matches_flann;
+    FeatureMatches matches_faiss;
 
     SiftMatchingOptions custom_options = options;
     custom_options.use_gpu = false;
-    custom_options.brute_force_cpu_matcher = true;
+    custom_options.cpu_brute_force_matcher = true;
     auto bf_matcher = CreateSiftFeatureMatcher(custom_options);
-    custom_options.brute_force_cpu_matcher = false;
-    auto flann_matcher = CreateSiftFeatureMatcher(custom_options);
+    custom_options.cpu_brute_force_matcher = false;
+    custom_options.cpu_descriptor_index_cache = &index_cache_helper.index_cache;
+    auto faiss_matcher = CreateSiftFeatureMatcher(custom_options);
 
-    bf_matcher->Match(descriptors1_ptr, descriptors2_ptr, &matches_bf);
-    flann_matcher->Match(descriptors1_ptr, descriptors2_ptr, &matches_flann);
-    CheckEqualMatches(matches_bf, matches_flann);
+    bf_matcher->Match(image1, image2, &matches_bf);
+    faiss_matcher->Match(image1, image2, &matches_faiss);
+    CheckEqualMatches(matches_bf, matches_faiss);
 
     const size_t num_matches = matches_bf.size();
 
-    const auto empty_descriptors = std::make_shared<FeatureDescriptors>(0, 128);
+    bf_matcher->Match(image0, image2, &matches_bf);
+    faiss_matcher->Match(image0, image2, &matches_faiss);
+    CheckEqualMatches(matches_bf, matches_faiss);
 
-    bf_matcher->Match(empty_descriptors, descriptors2_ptr, &matches_bf);
-    flann_matcher->Match(empty_descriptors, descriptors2_ptr, &matches_flann);
-    CheckEqualMatches(matches_bf, matches_flann);
+    bf_matcher->Match(image1, image0, &matches_bf);
+    faiss_matcher->Match(image1, image0, &matches_faiss);
+    CheckEqualMatches(matches_bf, matches_faiss);
 
-    bf_matcher->Match(descriptors1_ptr, empty_descriptors, &matches_bf);
-    flann_matcher->Match(descriptors1_ptr, empty_descriptors, &matches_flann);
-    CheckEqualMatches(matches_bf, matches_flann);
-
-    bf_matcher->Match(empty_descriptors, empty_descriptors, &matches_bf);
-    flann_matcher->Match(empty_descriptors, empty_descriptors, &matches_flann);
-    CheckEqualMatches(matches_bf, matches_flann);
+    bf_matcher->Match(image0, image0, &matches_bf);
+    faiss_matcher->Match(image0, image0, &matches_faiss);
+    CheckEqualMatches(matches_bf, matches_faiss);
 
     return num_matches;
   };
@@ -415,7 +470,7 @@ TEST(SiftCPUFeatureMatcherFlannVsBruteForce, Nominal) {
     const FeatureDescriptors descriptors1 = CreateRandomFeatureDescriptors(50);
     const FeatureDescriptors descriptors2 = CreateRandomFeatureDescriptors(50);
     SiftMatchingOptions match_options;
-    TestFlannVsBruteForce(match_options, descriptors1, descriptors2);
+    TestFaissVsBruteForce(match_options, descriptors1, descriptors2);
   }
 
   {
@@ -423,7 +478,7 @@ TEST(SiftCPUFeatureMatcherFlannVsBruteForce, Nominal) {
     const FeatureDescriptors descriptors2 = descriptors1.colwise().reverse();
     SiftMatchingOptions match_options;
     const size_t num_matches =
-        TestFlannVsBruteForce(match_options, descriptors1, descriptors2);
+        TestFaissVsBruteForce(match_options, descriptors1, descriptors2);
     EXPECT_EQ(num_matches, 50);
   }
 
@@ -434,7 +489,7 @@ TEST(SiftCPUFeatureMatcherFlannVsBruteForce, Nominal) {
 
     SiftMatchingOptions match_options;
     const size_t num_matches1 =
-        TestFlannVsBruteForce(match_options, descriptors1, descriptors2);
+        TestFaissVsBruteForce(match_options, descriptors1, descriptors2);
     EXPECT_EQ(num_matches1, 50);
 
     descriptors2.row(49) = descriptors2.row(0);
@@ -446,13 +501,13 @@ TEST(SiftCPUFeatureMatcherFlannVsBruteForce, Nominal) {
         descriptors2.row(49).cast<float>().normalized());
 
     match_options.max_ratio = 0.4;
-    const size_t num_matches2 = TestFlannVsBruteForce(
+    const size_t num_matches2 = TestFaissVsBruteForce(
         match_options, descriptors1.topRows(49), descriptors2);
     EXPECT_EQ(num_matches2, 48);
 
-    match_options.max_ratio = 0.5;
+    match_options.max_ratio = 0.6;
     const size_t num_matches3 =
-        TestFlannVsBruteForce(match_options, descriptors1, descriptors2);
+        TestFaissVsBruteForce(match_options, descriptors1, descriptors2);
     EXPECT_EQ(num_matches3, 49);
   }
 
@@ -466,29 +521,40 @@ TEST(SiftCPUFeatureMatcherFlannVsBruteForce, Nominal) {
 
     match_options.cross_check = false;
     const size_t num_matches1 =
-        TestFlannVsBruteForce(match_options, descriptors1, descriptors2);
+        TestFaissVsBruteForce(match_options, descriptors1, descriptors2);
     EXPECT_EQ(num_matches1, 50);
 
     match_options.cross_check = true;
     const size_t num_matches2 =
-        TestFlannVsBruteForce(match_options, descriptors1, descriptors2);
+        TestFaissVsBruteForce(match_options, descriptors1, descriptors2);
     EXPECT_EQ(num_matches2, 48);
   }
 }
 
 TEST(MatchGuidedSiftFeaturesCPU, Nominal) {
-  auto empty_keypoints = std::make_shared<FeatureKeypoints>(0);
-  auto keypoints1 = std::make_shared<FeatureKeypoints>(2);
-  (*keypoints1)[0].x = 1;
-  (*keypoints1)[1].x = 2;
-  auto keypoints2 = std::make_shared<FeatureKeypoints>(2);
-  (*keypoints2)[0].x = 2;
-  (*keypoints2)[1].x = 1;
-  const auto empty_descriptors = std::make_shared<FeatureDescriptors>(0, 128);
-  const auto descriptors1 =
-      std::make_shared<FeatureDescriptors>(CreateRandomFeatureDescriptors(2));
-  const auto descriptors2 =
-      std::make_shared<FeatureDescriptors>(descriptors1->colwise().reverse());
+  const FeatureMatcher::Image image0 = {
+      0,
+      std::make_shared<FeatureDescriptors>(0, 128),
+      std::make_shared<FeatureKeypoints>(0)};
+  const FeatureMatcher::Image image1 = {
+      1,
+      std::make_shared<FeatureDescriptors>(CreateRandomFeatureDescriptors(2)),
+      std::make_shared<FeatureKeypoints>(
+          std::vector<FeatureKeypoint>{{1, 0}, {2, 0}})};
+  const FeatureMatcher::Image image2 = {
+      2,
+      std::make_shared<FeatureDescriptors>(
+          image1.descriptors->colwise().reverse()),
+      std::make_shared<FeatureKeypoints>(
+          std::vector<FeatureKeypoint>{{2, 0}, {1, 0}})};
+  const FeatureMatcher::Image image3 = {
+      3,
+      std::make_shared<FeatureDescriptors>(CreateRandomFeatureDescriptors(2)),
+      std::make_shared<FeatureKeypoints>(
+          std::vector<FeatureKeypoint>{{100, 0}, {2, 0}})};
+
+  FeatureDescriptorIndexCacheHelper index_cache_helper(
+      {image0, image1, image2, image3});
 
   TwoViewGeometry two_view_geometry;
   two_view_geometry.config = TwoViewGeometry::PLANAR_OR_PANORAMIC;
@@ -496,53 +562,28 @@ TEST(MatchGuidedSiftFeaturesCPU, Nominal) {
 
   SiftMatchingOptions options;
   options.use_gpu = false;
+  options.cpu_descriptor_index_cache = &index_cache_helper.index_cache;
   auto matcher = CreateSiftFeatureMatcher(options);
 
   constexpr double kMaxError = 4.0;
 
-  matcher->MatchGuided(kMaxError,
-                       keypoints1,
-                       keypoints2,
-                       descriptors1,
-                       descriptors2,
-                       &two_view_geometry);
+  matcher->MatchGuided(kMaxError, image1, image2, &two_view_geometry);
   EXPECT_EQ(two_view_geometry.inlier_matches.size(), 2);
   EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 0);
   EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 1);
   EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx1, 1);
   EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx2, 0);
 
-  (*keypoints1)[0].x = 100;
-  matcher->MatchGuided(kMaxError,
-                       keypoints1,
-                       keypoints2,
-                       descriptors1,
-                       descriptors2,
-                       &two_view_geometry);
+  matcher->MatchGuided(kMaxError, image3, image2, &two_view_geometry);
   EXPECT_EQ(two_view_geometry.inlier_matches.size(), 1);
   EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 1);
   EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 0);
 
-  matcher->MatchGuided(kMaxError,
-                       empty_keypoints,
-                       keypoints2,
-                       empty_descriptors,
-                       descriptors2,
-                       &two_view_geometry);
+  matcher->MatchGuided(kMaxError, image0, image2, &two_view_geometry);
   EXPECT_EQ(two_view_geometry.inlier_matches.size(), 0);
-  matcher->MatchGuided(kMaxError,
-                       keypoints1,
-                       empty_keypoints,
-                       descriptors1,
-                       empty_descriptors,
-                       &two_view_geometry);
+  matcher->MatchGuided(kMaxError, image1, image0, &two_view_geometry);
   EXPECT_EQ(two_view_geometry.inlier_matches.size(), 0);
-  matcher->MatchGuided(kMaxError,
-                       empty_keypoints,
-                       empty_keypoints,
-                       empty_descriptors,
-                       empty_descriptors,
-                       &two_view_geometry);
+  matcher->MatchGuided(kMaxError, image0, image0, &two_view_geometry);
   EXPECT_EQ(two_view_geometry.inlier_matches.size(), 0);
 }
 
@@ -561,48 +602,38 @@ TEST(MatchSiftFeaturesGPU, Nominal) {
       options.max_num_matches = 1000;
       auto matcher = THROW_CHECK_NOTNULL(CreateSiftFeatureMatcher(options));
 
-      const auto empty_descriptors =
-          std::make_shared<FeatureDescriptors>(0, 128);
-      const auto descriptors1 = std::make_shared<FeatureDescriptors>(
-          CreateRandomFeatureDescriptors(2));
-      const auto descriptors2 = std::make_shared<FeatureDescriptors>(
-          descriptors1->colwise().reverse());
+      const FeatureMatcher::Image image0 = {
+          0, std::make_shared<FeatureDescriptors>(0, 128)};
+      const FeatureMatcher::Image image1 = {
+          1,
+          std::make_shared<FeatureDescriptors>(
+              CreateRandomFeatureDescriptors(2))};
+      const FeatureMatcher::Image image2 = {
+          2,
+          std::make_shared<FeatureDescriptors>(
+              image1.descriptors->colwise().reverse())};
 
       FeatureMatches matches;
 
-      matcher->Match(descriptors1, descriptors2, &matches);
+      matcher->Match(image1, image2, &matches);
       EXPECT_EQ(matches.size(), 2);
       EXPECT_EQ(matches[0].point2D_idx1, 0);
       EXPECT_EQ(matches[0].point2D_idx2, 1);
       EXPECT_EQ(matches[1].point2D_idx1, 1);
       EXPECT_EQ(matches[1].point2D_idx2, 0);
 
-      matcher->Match(nullptr, nullptr, &matches);
+      matcher->Match(image1, image2, &matches);
       EXPECT_EQ(matches.size(), 2);
       EXPECT_EQ(matches[0].point2D_idx1, 0);
       EXPECT_EQ(matches[0].point2D_idx2, 1);
       EXPECT_EQ(matches[1].point2D_idx1, 1);
       EXPECT_EQ(matches[1].point2D_idx2, 0);
 
-      matcher->Match(descriptors1, nullptr, &matches);
-      EXPECT_EQ(matches.size(), 2);
-      EXPECT_EQ(matches[0].point2D_idx1, 0);
-      EXPECT_EQ(matches[0].point2D_idx2, 1);
-      EXPECT_EQ(matches[1].point2D_idx1, 1);
-      EXPECT_EQ(matches[1].point2D_idx2, 0);
-
-      matcher->Match(nullptr, descriptors2, &matches);
-      EXPECT_EQ(matches.size(), 2);
-      EXPECT_EQ(matches[0].point2D_idx1, 0);
-      EXPECT_EQ(matches[0].point2D_idx2, 1);
-      EXPECT_EQ(matches[1].point2D_idx1, 1);
-      EXPECT_EQ(matches[1].point2D_idx2, 0);
-
-      matcher->Match(empty_descriptors, descriptors2, &matches);
+      matcher->Match(image0, image2, &matches);
       EXPECT_EQ(matches.size(), 0);
-      matcher->Match(descriptors1, empty_descriptors, &matches);
+      matcher->Match(image1, image0, &matches);
       EXPECT_EQ(matches.size(), 0);
-      matcher->Match(empty_descriptors, empty_descriptors, &matches);
+      matcher->Match(image0, image0, &matches);
       EXPECT_EQ(matches.size(), 0);
     }
     OpenGLContextManager opengl_context_;
@@ -626,41 +657,45 @@ TEST(MatchSiftFeaturesCPUvsGPU, Nominal) {
       auto TestCPUvsGPU = [](const SiftMatchingOptions& options,
                              const FeatureDescriptors& descriptors1,
                              const FeatureDescriptors& descriptors2) {
+        const FeatureMatcher::Image image0 = {
+            0, std::make_shared<FeatureDescriptors>(0, 128)};
+        const FeatureMatcher::Image image1 = {
+            1, std::make_shared<FeatureDescriptors>(descriptors1)};
+        const FeatureMatcher::Image image2 = {
+            2, std::make_shared<FeatureDescriptors>(descriptors2)};
+
+        FeatureDescriptorIndexCacheHelper index_cache_helper(
+            {image0, image1, image2});
+
         SiftMatchingOptions custom_options = options;
         custom_options.use_gpu = true;
         custom_options.max_num_matches = 1000;
         auto gpu_matcher =
             THROW_CHECK_NOTNULL(CreateSiftFeatureMatcher(custom_options));
         custom_options.use_gpu = false;
+        custom_options.cpu_descriptor_index_cache =
+            &index_cache_helper.index_cache;
         auto cpu_matcher = CreateSiftFeatureMatcher(custom_options);
 
-        const auto descriptors1_ptr =
-            std::make_shared<FeatureDescriptors>(descriptors1);
-        const auto descriptors2_ptr =
-            std::make_shared<FeatureDescriptors>(descriptors2);
-
         FeatureMatches matches_cpu;
         FeatureMatches matches_gpu;
 
-        cpu_matcher->Match(descriptors1_ptr, descriptors2_ptr, &matches_cpu);
-        gpu_matcher->Match(descriptors1_ptr, descriptors2_ptr, &matches_gpu);
+        cpu_matcher->Match(image1, image2, &matches_cpu);
+        gpu_matcher->Match(image1, image2, &matches_gpu);
         CheckEqualMatches(matches_cpu, matches_gpu);
 
         const size_t num_matches = matches_cpu.size();
 
-        const auto empty_descriptors =
-            std::make_shared<FeatureDescriptors>(0, 128);
-
-        cpu_matcher->Match(empty_descriptors, descriptors2_ptr, &matches_cpu);
-        gpu_matcher->Match(empty_descriptors, descriptors2_ptr, &matches_gpu);
+        cpu_matcher->Match(image0, image2, &matches_cpu);
+        gpu_matcher->Match(image0, image2, &matches_gpu);
         CheckEqualMatches(matches_cpu, matches_gpu);
 
-        cpu_matcher->Match(descriptors1_ptr, empty_descriptors, &matches_cpu);
-        gpu_matcher->Match(descriptors1_ptr, empty_descriptors, &matches_gpu);
+        cpu_matcher->Match(image1, image0, &matches_cpu);
+        gpu_matcher->Match(image1, image0, &matches_gpu);
         CheckEqualMatches(matches_cpu, matches_gpu);
 
-        cpu_matcher->Match(empty_descriptors, empty_descriptors, &matches_cpu);
-        gpu_matcher->Match(empty_descriptors, empty_descriptors, &matches_gpu);
+        cpu_matcher->Match(image0, image0, &matches_cpu);
+        gpu_matcher->Match(image0, image0, &matches_gpu);
         CheckEqualMatches(matches_cpu, matches_gpu);
 
         return num_matches;
@@ -709,7 +744,7 @@ TEST(MatchSiftFeaturesCPUvsGPU, Nominal) {
             TestCPUvsGPU(match_options, descriptors1.topRows(49), descriptors2);
         EXPECT_EQ(num_matches2, 48);
 
-        match_options.max_ratio = 0.5;
+        match_options.max_ratio = 0.6;
         const size_t num_matches3 =
             TestCPUvsGPU(match_options, descriptors1, descriptors2);
         EXPECT_EQ(num_matches3, 49);
@@ -750,107 +785,65 @@ TEST(MatchGuidedSiftFeaturesGPU, Nominal) {
   class TestThread : public Thread {
    private:
     void Run() {
+      const FeatureMatcher::Image image0 = {
+          0,
+          std::make_shared<FeatureDescriptors>(0, 128),
+          std::make_shared<FeatureKeypoints>(0)};
+      const FeatureMatcher::Image image1 = {
+          1,
+          std::make_shared<FeatureDescriptors>(
+              CreateRandomFeatureDescriptors(2)),
+          std::make_shared<FeatureKeypoints>(
+              std::vector<FeatureKeypoint>{{1, 0}, {2, 0}})};
+      const FeatureMatcher::Image image2 = {
+          2,
+          std::make_shared<FeatureDescriptors>(
+              image1.descriptors->colwise().reverse()),
+          std::make_shared<FeatureKeypoints>(
+              std::vector<FeatureKeypoint>{{2, 0}, {1, 0}})};
+      const FeatureMatcher::Image image3 = {
+          3,
+          std::make_shared<FeatureDescriptors>(
+              CreateRandomFeatureDescriptors(2)),
+          std::make_shared<FeatureKeypoints>(
+              std::vector<FeatureKeypoint>{{100, 0}, {1, 0}})};
+
       opengl_context_.MakeCurrent();
       SiftMatchingOptions options;
       options.use_gpu = true;
       options.max_num_matches = 1000;
       auto matcher = THROW_CHECK_NOTNULL(CreateSiftFeatureMatcher(options));
 
-      auto empty_keypoints = std::make_shared<FeatureKeypoints>(0);
-      auto keypoints1 = std::make_shared<FeatureKeypoints>(2);
-      (*keypoints1)[0].x = 1;
-      (*keypoints1)[1].x = 2;
-      auto keypoints2 = std::make_shared<FeatureKeypoints>(2);
-      (*keypoints2)[0].x = 2;
-      (*keypoints2)[1].x = 1;
-      const auto empty_descriptors =
-          std::make_shared<FeatureDescriptors>(0, 128);
-      const auto descriptors1 = std::make_shared<FeatureDescriptors>(
-          CreateRandomFeatureDescriptors(2));
-      const auto descriptors2 = std::make_shared<FeatureDescriptors>(
-          descriptors1->colwise().reverse());
-
       TwoViewGeometry two_view_geometry;
       two_view_geometry.config = TwoViewGeometry::PLANAR_OR_PANORAMIC;
       two_view_geometry.H = Eigen::Matrix3d::Identity();
 
       constexpr double kMaxError = 4.0;
 
-      matcher->MatchGuided(kMaxError,
-                           keypoints1,
-                           keypoints2,
-                           descriptors1,
-                           descriptors2,
-                           &two_view_geometry);
+      matcher->MatchGuided(kMaxError, image1, image2, &two_view_geometry);
       EXPECT_EQ(two_view_geometry.inlier_matches.size(), 2);
       EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 0);
       EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 1);
       EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx1, 1);
       EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx2, 0);
 
-      matcher->MatchGuided(
-          kMaxError, nullptr, nullptr, nullptr, nullptr, &two_view_geometry);
+      matcher->MatchGuided(kMaxError, image1, image2, &two_view_geometry);
       EXPECT_EQ(two_view_geometry.inlier_matches.size(), 2);
       EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 0);
       EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 1);
       EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx1, 1);
       EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx2, 0);
 
-      matcher->MatchGuided(kMaxError,
-                           keypoints1,
-                           nullptr,
-                           descriptors1,
-                           nullptr,
-                           &two_view_geometry);
-      EXPECT_EQ(two_view_geometry.inlier_matches.size(), 2);
-      EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 0);
-      EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 1);
-      EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx1, 1);
-      EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx2, 0);
-
-      matcher->MatchGuided(kMaxError,
-                           nullptr,
-                           keypoints2,
-                           nullptr,
-                           descriptors2,
-                           &two_view_geometry);
-      EXPECT_EQ(two_view_geometry.inlier_matches.size(), 2);
-      EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 0);
-      EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 1);
-      EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx1, 1);
-      EXPECT_EQ(two_view_geometry.inlier_matches[1].point2D_idx2, 0);
-
-      (*keypoints1)[0].x = 100;
-      matcher->MatchGuided(kMaxError,
-                           keypoints1,
-                           keypoints2,
-                           descriptors1,
-                           descriptors2,
-                           &two_view_geometry);
+      matcher->MatchGuided(kMaxError, image3, image2, &two_view_geometry);
       EXPECT_EQ(two_view_geometry.inlier_matches.size(), 1);
       EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx1, 1);
       EXPECT_EQ(two_view_geometry.inlier_matches[0].point2D_idx2, 0);
 
-      matcher->MatchGuided(kMaxError,
-                           empty_keypoints,
-                           keypoints2,
-                           empty_descriptors,
-                           descriptors2,
-                           &two_view_geometry);
+      matcher->MatchGuided(kMaxError, image0, image2, &two_view_geometry);
       EXPECT_EQ(two_view_geometry.inlier_matches.size(), 0);
-      matcher->MatchGuided(kMaxError,
-                           keypoints1,
-                           empty_keypoints,
-                           descriptors1,
-                           empty_descriptors,
-                           &two_view_geometry);
+      matcher->MatchGuided(kMaxError, image1, image0, &two_view_geometry);
       EXPECT_EQ(two_view_geometry.inlier_matches.size(), 0);
-      matcher->MatchGuided(kMaxError,
-                           empty_keypoints,
-                           empty_keypoints,
-                           empty_descriptors,
-                           empty_descriptors,
-                           &two_view_geometry);
+      matcher->MatchGuided(kMaxError, image0, image0, &two_view_geometry);
       EXPECT_EQ(two_view_geometry.inlier_matches.size(), 0);
     }
     OpenGLContextManager opengl_context_;
