@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -30,7 +30,9 @@
 #include "colmap/geometry/essential_matrix.h"
 
 #include "colmap/geometry/pose.h"
+#include "colmap/geometry/rigid3_matchers.h"
 #include "colmap/util/eigen_alignment.h"
+#include "colmap/util/eigen_matchers.h"
 
 #include <Eigen/Geometry>
 #include <gtest/gtest.h>
@@ -78,23 +80,21 @@ TEST(PoseFromEssentialMatrix, Nominal) {
   points3D[2] = Eigen::Vector3d(0.1, 0, 1);
   points3D[3] = Eigen::Vector3d(0.1, 0.1, 1);
 
-  std::vector<Eigen::Vector2d> points1(4);
-  std::vector<Eigen::Vector2d> points2(4);
+  std::vector<Eigen::Vector3d> rays1(4);
+  std::vector<Eigen::Vector3d> rays2(4);
   for (size_t i = 0; i < points3D.size(); ++i) {
-    points1[i] = (cam1_from_world * points3D[i]).hnormalized();
-    points2[i] = (cam2_from_world * points3D[i]).hnormalized();
+    rays1[i] = (cam1_from_world * points3D[i]).normalized();
+    rays2[i] = (cam2_from_world * points3D[i]).normalized();
   }
 
   points3D.clear();
 
-  Eigen::Matrix3d R;
-  Eigen::Vector3d t;
-  PoseFromEssentialMatrix(E, points1, points2, &R, &t, &points3D);
+  Rigid3d cam2_from_cam1_est;
+  PoseFromEssentialMatrix(E, rays1, rays2, &cam2_from_cam1_est, &points3D);
 
   EXPECT_EQ(points3D.size(), 4);
 
-  EXPECT_TRUE(R.isApprox(cam2_from_cam1.rotation.toRotationMatrix()));
-  EXPECT_TRUE(t.isApprox(cam2_from_cam1.translation));
+  EXPECT_THAT(cam2_from_cam1_est, Rigid3dNear(cam2_from_cam1));
 }
 
 TEST(FindOptimalImageObservations, Nominal) {
@@ -120,8 +120,8 @@ TEST(FindOptimalImageObservations, Nominal) {
     Eigen::Vector2d optimal_point2;
     FindOptimalImageObservations(
         E, point1, point2, &optimal_point1, &optimal_point2);
-    EXPECT_TRUE(point1.isApprox(optimal_point1));
-    EXPECT_TRUE(point2.isApprox(optimal_point2));
+    EXPECT_THAT(point1, EigenMatrixNear(optimal_point1));
+    EXPECT_THAT(point2, EigenMatrixNear(optimal_point2));
   }
 }
 
@@ -132,8 +132,8 @@ TEST(EpipoleFromEssentialMatrix, Nominal) {
 
   const Eigen::Vector3d left_epipole = EpipoleFromEssentialMatrix(E, true);
   const Eigen::Vector3d right_epipole = EpipoleFromEssentialMatrix(E, false);
-  EXPECT_TRUE(left_epipole.isApprox(Eigen::Vector3d(0, 0, 1)));
-  EXPECT_TRUE(right_epipole.isApprox(Eigen::Vector3d(0, 0, 1)));
+  EXPECT_THAT(left_epipole, EigenMatrixNear(Eigen::Vector3d(0, 0, 1)));
+  EXPECT_THAT(right_epipole, EigenMatrixNear(Eigen::Vector3d(0, 0, 1)));
 }
 
 TEST(InvertEssentialMatrix, Nominal) {
@@ -144,7 +144,7 @@ TEST(InvertEssentialMatrix, Nominal) {
     const Eigen::Matrix3d E = EssentialMatrixFromPose(cam2_from_cam1);
     const Eigen::Matrix3d inv_inv_E =
         InvertEssentialMatrix(InvertEssentialMatrix(E));
-    EXPECT_TRUE(E.isApprox(inv_inv_E));
+    EXPECT_THAT(E, EigenMatrixNear(inv_inv_E));
   }
 }
 
@@ -157,8 +157,22 @@ TEST(FundamentalFromEssentialMatrix, Nominal) {
       (Eigen::Matrix3d() << 3, 0, 2, 0, 4, 1, 0, 0, 1).finished();
   const Eigen::Matrix3d F = FundamentalFromEssentialMatrix(K2, E, K1);
   const Eigen::Vector3d x(3, 2, 1);
-  EXPECT_TRUE((K2.transpose().inverse() * E * x).isApprox(F * K1 * x));
-  EXPECT_TRUE((E * K1.inverse() * x).isApprox(K2.transpose() * F * x));
+  EXPECT_THAT(K2.transpose().inverse() * E * x,
+              EigenMatrixNear(Eigen::Vector3d(F * K1 * x)));
+  EXPECT_THAT(E * K1.inverse() * x,
+              EigenMatrixNear(Eigen::Vector3d(K2.transpose() * F * x)));
+}
+
+TEST(EssentialFromFundamentalMatrix, Nominal) {
+  const Eigen::Matrix3d E = EssentialMatrixFromPose(
+      Rigid3d(Eigen::Quaterniond::UnitRandom(), Eigen::Vector3d::Random()));
+  const Eigen::Matrix3d K1 =
+      (Eigen::Matrix3d() << 2, 0, 1, 0, 3, 2, 0, 0, 1).finished();
+  const Eigen::Matrix3d K2 =
+      (Eigen::Matrix3d() << 3, 0, 2, 0, 4, 1, 0, 0, 1).finished();
+  const Eigen::Matrix3d F = FundamentalFromEssentialMatrix(K2, E, K1);
+  EXPECT_THAT(EssentialFromFundamentalMatrix(K2, F, K1),
+              EigenMatrixNear(E, 1e-6));
 }
 
 }  // namespace
