@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -32,10 +32,15 @@
 #include "colmap/util/eigen_alignment.h"
 #include "colmap/util/types.h"
 
+#include <ostream>
+
 #include <Eigen/Geometry>
 
 namespace colmap {
 
+// Compose the skew symmetric cross product matrix from a vector.
+Eigen::Matrix3d CrossProductMatrix(const Eigen::Vector3d& vector);
+
 // 3D rigid transform with 6 degrees of freedom.
 // Transforms point x from a to b as: x_in_b = R * x_in_a + t.
 struct Rigid3d {
@@ -55,6 +60,13 @@ struct Rigid3d {
     return matrix;
   }
 
+  static inline Rigid3d FromMatrix(const Eigen::Matrix3x4d& matrix) {
+    Rigid3d t;
+    t.rotation = Eigen::Quaterniond(matrix.leftCols<3>()).normalized();
+    t.translation = matrix.rightCols<1>();
+    return t;
+  }
+
   // Adjoint matrix to propagate uncertainty on Rigid3d
   // [Reference] https://gtsam.org/2021/02/23/uncertainties-part3.html
   inline Eigen::Matrix6d Adjoint() const {
@@ -66,6 +78,16 @@ struct Rigid3d {
     adjoint.block<3, 3>(3, 3) = adjoint.block<3, 3>(0, 0);
     return adjoint;
   }
+
+  inline Eigen::Matrix6d AdjointInverse() const {
+    Eigen::Matrix6d adjoint_inv;
+    adjoint_inv.block<3, 3>(0, 0) = rotation.toRotationMatrix().transpose();
+    adjoint_inv.block<3, 3>(0, 3).setZero();
+    adjoint_inv.block<3, 3>(3, 0) =
+        -adjoint_inv.block<3, 3>(0, 0) * CrossProductMatrix(translation);
+    adjoint_inv.block<3, 3>(3, 3) = adjoint_inv.block<3, 3>(0, 0);
+    return adjoint_inv;
+  }
 };
 
 // Return inverse transform.
@@ -76,11 +98,43 @@ inline Rigid3d Inverse(const Rigid3d& b_from_a) {
   return a_from_b;
 }
 
-// Update covariance (6x6) for rigid3d.inverse()
+// Update covariance (6 x 6) for rigid3d.inverse()
+//
+// [Reference] Joan Solà, Jeremie Deray, Dinesh Atchuthan, A micro Lie theory
+// for state estimation in robotics, 2018.
+// In COLMAP we follow the left convention (rather than the right convention as
+// in the reference paper and GTSAM). With the left convention the Jacobian of
+// the inverse is -Ad(X^{-1}). This can be easily derived combining Eqs. (62)
+// and (57) in the paper.
 inline Eigen::Matrix6d GetCovarianceForRigid3dInverse(
     const Rigid3d& rigid3, const Eigen::Matrix6d& covar) {
-  const Eigen::Matrix6d adjoint = rigid3.Adjoint();
-  return adjoint * covar * adjoint.transpose();
+  const Eigen::Matrix6d adjoint_inv = rigid3.AdjointInverse();
+  return adjoint_inv * covar * adjoint_inv.transpose();
+}
+
+// Given a (12 x 12) covariance on two rigid3d objects (a_from_b, b_from_c),
+// this function calculates the (6 x 6) covariance of the composed
+// transformation a_from_c (a_T_b * b_T_c). b_T_c does not contribute to the
+// covariance propagation and is thus not required.
+inline Eigen::Matrix6d GetCovarianceForComposedRigid3d(
+    const Rigid3d& a_from_b, const Eigen::Matrix<double, 12, 12>& covar) {
+  Eigen::Matrix<double, 6, 12> J;
+  J.block<6, 6>(0, 0) = Eigen::Matrix6d::Identity();
+  J.block<6, 6>(0, 6) = a_from_b.Adjoint();
+  return J * covar * J.transpose();
+}
+
+// Given a (12 x 12) covariance on two rigid3d objects (a_from_c, b_from_c),
+// this function calculates the (6 x 6) covariance of the relative
+// transformation b_from_a (b_T_c * a_T_c^{-1}).
+inline Eigen::Matrix6d GetCovarianceForRelativeRigid3d(
+    const Rigid3d& a_from_c,
+    const Rigid3d& b_from_c,
+    const Eigen::Matrix<double, 12, 12>& covar) {
+  Eigen::Matrix<double, 6, 12> J;
+  J.block<6, 6>(0, 0) = -b_from_c.Adjoint() * a_from_c.AdjointInverse();
+  J.block<6, 6>(0, 6) = Eigen::Matrix6d::Identity();
+  return J * covar * J.transpose();
 }
 
 // Apply transform to point such that one can write expressions like:
@@ -110,4 +164,14 @@ inline Rigid3d operator*(const Rigid3d& c_from_b, const Rigid3d& b_from_a) {
   return c_from_a;
 }
 
+inline bool operator==(const Rigid3d& left, const Rigid3d& right) {
+  return left.rotation.coeffs() == right.rotation.coeffs() &&
+         left.translation == right.translation;
+}
+inline bool operator!=(const Rigid3d& left, const Rigid3d& right) {
+  return !(left == right);
+}
+
+std::ostream& operator<<(std::ostream& stream, const Rigid3d& tform);
+
 }  // namespace colmap
