@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -80,13 +80,20 @@ void WarpImageBetweenCameras(const Camera& source_camera,
       image_point.x() = x + 0.5;
 
       // Camera models assume that the upper left pixel center is (0.5, 0.5).
-      const Eigen::Vector2d cam_point =
+      const std::optional<Eigen::Vector2d> cam_point =
           scaled_target_camera.CamFromImg(image_point);
-      const Eigen::Vector2d source_point = source_camera.ImgFromCam(cam_point);
+      if (!cam_point) {
+        target_image->SetPixel(x, y, BitmapColor<uint8_t>(0));
+        continue;
+      }
+
+      const std::optional<Eigen::Vector2d> source_point =
+          source_camera.ImgFromCam(cam_point->homogeneous());
 
       BitmapColor<float> color;
-      if (source_image.InterpolateBilinear(
-              source_point.x() - 0.5, source_point.y() - 0.5, &color)) {
+      if (source_point &&
+          source_image.InterpolateBilinear(
+              source_point->x() - 0.5, source_point->y() - 0.5, &color)) {
         target_image->SetPixel(x, y, color.Cast<uint8_t>());
       } else {
         target_image->SetPixel(x, y, BitmapColor<uint8_t>(0));
@@ -156,13 +163,24 @@ void WarpImageWithHomographyBetweenCameras(const Eigen::Matrix3d& H,
 
       // Camera models assume that the upper left pixel center is (0.5, 0.5).
       const Eigen::Vector3d warped_point = H * image_point;
-      const Eigen::Vector2d cam_point =
+      if (warped_point.z() == 0) {
+        continue;
+      }
+
+      const std::optional<Eigen::Vector2d> cam_point =
           target_camera.CamFromImg(warped_point.hnormalized());
-      const Eigen::Vector2d source_point = source_camera.ImgFromCam(cam_point);
+      if (!cam_point) {
+        target_image->SetPixel(x, y, BitmapColor<uint8_t>(0));
+        continue;
+      }
+
+      const std::optional<Eigen::Vector2d> source_point =
+          source_camera.ImgFromCam(cam_point->homogeneous());
 
       BitmapColor<float> color;
-      if (source_image.InterpolateBilinear(
-              source_point.x() - 0.5, source_point.y() - 0.5, &color)) {
+      if (source_point &&
+          source_image.InterpolateBilinear(
+              source_point->x() - 0.5, source_point->y() - 0.5, &color)) {
         target_image->SetPixel(x, y, color.Cast<uint8_t>());
       } else {
         target_image->SetPixel(x, y, BitmapColor<uint8_t>(0));
