@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -84,29 +84,27 @@ void Shuffle(uint32_t num_to_shuffle, std::vector<T>* elems);
 
 template <typename T>
 T RandomUniformInteger(const T min, const T max) {
-  if (PRNG == nullptr) {
+  if (COLMAP_PREDICT_FALSE(PRNG == nullptr)) {
     SetPRNGSeed();
   }
 
   std::uniform_int_distribution<T> distribution(min, max);
-
   return distribution(*PRNG);
 }
 
 template <typename T>
 T RandomUniformReal(const T min, const T max) {
-  if (PRNG == nullptr) {
+  if (COLMAP_PREDICT_FALSE(PRNG == nullptr)) {
     SetPRNGSeed();
   }
 
   std::uniform_real_distribution<T> distribution(min, max);
-
   return distribution(*PRNG);
 }
 
 template <typename T>
 T RandomGaussian(const T mean, const T stddev) {
-  if (PRNG == nullptr) {
+  if (COLMAP_PREDICT_FALSE(PRNG == nullptr)) {
     SetPRNGSeed();
   }
 
