@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -29,7 +29,9 @@
 
 #include "colmap/mvs/fusion.h"
 
+#include "colmap/math/math.h"
 #include "colmap/util/eigen_alignment.h"
+#include "colmap/util/file.h"
 #include "colmap/util/misc.h"
 #include "colmap/util/threading.h"
 #include "colmap/util/timer.h"
@@ -40,21 +42,6 @@ namespace colmap {
 namespace mvs {
 namespace internal {
 
-template <typename T>
-float Median(std::vector<T>* elems) {
-  THROW_CHECK(!elems->empty());
-  const size_t mid_idx = elems->size() / 2;
-  std::nth_element(elems->begin(), elems->begin() + mid_idx, elems->end());
-  if (elems->size() % 2 == 0) {
-    const float mid_element1 = static_cast<float>((*elems)[mid_idx]);
-    const float mid_element2 = static_cast<float>(
-        *std::max_element(elems->begin(), elems->begin() + mid_idx));
-    return (mid_element1 + mid_element2) / 2.0f;
-  } else {
-    return static_cast<float>((*elems)[mid_idx]);
-  }
-}
-
 // Use the sparse model to find most connected image that has not yet been
 // fused. This is used as a heuristic to ensure that the workspace cache reuses
 // already cached images as efficient as possible.
@@ -84,7 +71,7 @@ int FindNextImage(const std::vector<std::vector<int>>& overlapping_images,
 }  // namespace internal
 
 void StereoFusionOptions::Print() const {
-#define PrintOption(option) LOG(INFO) << #option ": " << (option) << std::endl
+#define PrintOption(option) LOG(INFO) << #option ": " << (option) << '\n';
   PrintHeading2("StereoFusion::Options");
   PrintOption(mask_path);
   PrintOption(max_image_size);
@@ -352,9 +339,13 @@ void StereoFusion::InitFusedPixelMask(int image_idx,
                                       size_t height) {
   Bitmap mask;
   Mat<char>& fused_pixel_mask = fused_pixel_masks_.at(image_idx);
-  const std::string mask_path =
-      JoinPaths(options_.mask_path,
-                workspace_->GetModel().GetImageName(image_idx) + ".png");
+  const std::string mask_image_name =
+      workspace_->GetModel().GetImageName(image_idx);
+  std::string mask_path =
+      JoinPaths(options_.mask_path, mask_image_name + ".png");
+  if (!ExistsFile(mask_path) && HasFileExtension(mask_image_name, ".png")) {
+    mask_path = JoinPaths(options_.mask_path, mask_image_name);
+  }
   fused_pixel_mask = Mat<char>(width, height, 1);
   if (!options_.mask_path.empty() && ExistsFile(mask_path) &&
       mask.Read(mask_path, false)) {
@@ -534,28 +525,28 @@ void StereoFusion::Fuse(const int thread_id,
     PlyPoint fused_point;
 
     Eigen::Vector3f fused_normal;
-    fused_normal.x() = internal::Median(&fused_point_nx);
-    fused_normal.y() = internal::Median(&fused_point_ny);
-    fused_normal.z() = internal::Median(&fused_point_nz);
+    fused_normal.x() = Median(fused_point_nx);
+    fused_normal.y() = Median(fused_point_ny);
+    fused_normal.z() = Median(fused_point_nz);
     const float fused_normal_norm = fused_normal.norm();
     if (fused_normal_norm < std::numeric_limits<float>::epsilon()) {
       return;
     }
 
-    fused_point.x = internal::Median(&fused_point_x);
-    fused_point.y = internal::Median(&fused_point_y);
-    fused_point.z = internal::Median(&fused_point_z);
+    fused_point.x = Median(fused_point_x);
+    fused_point.y = Median(fused_point_y);
+    fused_point.z = Median(fused_point_z);
 
     fused_point.nx = fused_normal.x() / fused_normal_norm;
     fused_point.ny = fused_normal.y() / fused_normal_norm;
     fused_point.nz = fused_normal.z() / fused_normal_norm;
 
-    fused_point.r = TruncateCast<float, uint8_t>(
-        std::round(internal::Median(&fused_point_r)));
-    fused_point.g = TruncateCast<float, uint8_t>(
-        std::round(internal::Median(&fused_point_g)));
-    fused_point.b = TruncateCast<float, uint8_t>(
-        std::round(internal::Median(&fused_point_b)));
+    fused_point.r =
+        TruncateCast<float, uint8_t>(std::round(Median(fused_point_r)));
+    fused_point.g =
+        TruncateCast<float, uint8_t>(std::round(Median(fused_point_g)));
+    fused_point.b =
+        TruncateCast<float, uint8_t>(std::round(Median(fused_point_b)));
 
     task_fused_points_[thread_id].push_back(fused_point);
     task_fused_points_visibility_[thread_id].emplace_back(
