@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -42,6 +42,7 @@
 #include "colmap/math/random.h"
 #include "colmap/scene/reconstruction.h"
 #include "colmap/util/endian.h"
+#include "colmap/util/file.h"
 #include "colmap/util/logging.h"
 #include "colmap/util/misc.h"
 #include "colmap/util/ply.h"
@@ -122,6 +123,10 @@ bool PoissonMeshing(const PoissonMeshingOptions& options,
                     const std::string& input_path,
                     const std::string& output_path) {
   THROW_CHECK(options.Check());
+  THROW_CHECK_HAS_FILE_EXTENSION(input_path, ".ply");
+  THROW_CHECK_FILE_EXISTS(input_path);
+  THROW_CHECK_HAS_FILE_EXTENSION(output_path, ".ply");
+  THROW_CHECK_PATH_OPEN(output_path);
 
   std::vector<std::string> args;
 
@@ -144,12 +149,10 @@ bool PoissonMeshing(const PoissonMeshingOptions& options,
     args.push_back(std::to_string(options.color));
   }
 
-#if defined(COLMAP_OPENMP_ENABLED)
   if (options.num_threads > 0) {
     args.push_back("--threads");
     args.push_back(std::to_string(options.num_threads));
   }
-#endif  // OPENMP_ENABLED
 
   if (options.trim > 0) {
     args.push_back("--density");
@@ -390,14 +393,15 @@ class DelaunayMeshingInput {
           }
 
           // Check reprojection error between the two points.
-          const Eigen::Vector2f point_proj =
-              camera.ImgFromCam(point_local.hnormalized().cast<double>())
-                  .cast<float>();
-          const Eigen::Vector2f cell_point_proj =
-              camera.ImgFromCam(cell_point_local.hnormalized().cast<double>())
-                  .cast<float>();
+          const std::optional<Eigen::Vector2d> point_proj =
+              camera.ImgFromCam(point_local.cast<double>());
+          const std::optional<Eigen::Vector2d> cell_point_proj =
+              camera.ImgFromCam(cell_point_local.cast<double>());
+          if (!point_proj || !cell_point_proj) {
+            continue;
+          }
           const float squared_proj_dist =
-              (point_proj - cell_point_proj).squaredNorm();
+              (*point_proj - *cell_point_proj).squaredNorm();
           if (squared_proj_dist > max_squared_proj_dist) {
             insert_point = true;
             break;
@@ -441,7 +445,7 @@ struct DelaunayMeshingEdgeWeightComputer {
     }
 
     distance_sigma_ = distance_sigma_factor *
-                      std::max(std::sqrt(Percentile(edge_lengths, 25)), 1e-7f);
+                      std::max(std::sqrt(Percentile(edge_lengths, 25)), 1e-7);
     distance_threshold_ = 5 * distance_sigma_;
     distance_normalization_ = -0.5 / (distance_sigma_ * distance_sigma_);
   }
@@ -636,20 +640,18 @@ void WriteDelaunayTriangulationPly(const std::string& path,
   std::fstream file(path, std::ios::out);
   THROW_CHECK_FILE_OPEN(file, path);
 
-  file << "ply" << std::endl;
-  file << "format ascii 1.0" << std::endl;
-  file << "element vertex " << triangulation.number_of_vertices() << std::endl;
-  file << "property float x" << std::endl;
-  file << "property float y" << std::endl;
-  file << "property float z" << std::endl;
-  file << "element edge " << triangulation.number_of_finite_edges()
-       << std::endl;
-  file << "property int vertex1" << std::endl;
-  file << "property int vertex2" << std::endl;
-  file << "element face " << triangulation.number_of_finite_facets()
-       << std::endl;
-  file << "property list uchar int vertex_index" << std::endl;
-  file << "end_header" << std::endl;
+  file << "ply\n";
+  file << "format ascii 1.0\n";
+  file << "element vertex " << triangulation.number_of_vertices() << '\n';
+  file << "property float x\n";
+  file << "property float y\n";
+  file << "property float z\n";
+  file << "element edge " << triangulation.number_of_finite_edges() << '\n';
+  file << "property int vertex1\n";
+  file << "property int vertex2\n";
+  file << "element face " << triangulation.number_of_finite_facets() << '\n';
+  file << "property list uchar int vertex_index\n";
+  file << "end_header\n";
 
   std::unordered_map<const Delaunay::Vertex_handle, size_t> vertex_indices;
   vertex_indices.reserve(triangulation.number_of_vertices());
@@ -658,14 +660,14 @@ void WriteDelaunayTriangulationPly(const std::string& path,
        ++it) {
     vertex_indices.emplace(it, vertex_indices.size());
     file << it->point().x() << " " << it->point().y() << " " << it->point().z()
-         << std::endl;
+         << '\n';
   }
 
   for (auto it = triangulation.finite_edges_begin();
        it != triangulation.finite_edges_end();
        ++it) {
     file << vertex_indices.at(it->first->vertex(it->second)) << " "
-         << vertex_indices.at(it->first->vertex(it->third)) << std::endl;
+         << vertex_indices.at(it->first->vertex(it->third)) << '\n';
   }
 
   for (auto it = triangulation.finite_facets_begin();
@@ -680,7 +682,7 @@ void WriteDelaunayTriangulationPly(const std::string& path,
          << " "
          << vertex_indices.at(it->first->vertex(
                 triangulation.vertex_triple_index(it->second, 2)))
-         << std::endl;
+         << '\n';
   }
 }
 
@@ -989,7 +991,7 @@ PlyMesh DelaunayMeshing(const DelaunayMeshingOptions& options,
 
   const float max_facet_side_length =
       options.max_side_length_factor *
-      Percentile(surface_facet_side_lengths,
+      Percentile(std::vector<float>(surface_facet_side_lengths),
                  options.max_side_length_percentile);
 
   mesh.faces.reserve(surface_facets.size());
