@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -33,6 +33,7 @@
 #include "colmap/mvs/consistency_graph.h"
 #include "colmap/mvs/patch_match_cuda.h"
 #include "colmap/mvs/workspace.h"
+#include "colmap/util/file.h"
 #include "colmap/util/misc.h"
 
 #include <numeric>
@@ -48,38 +49,9 @@ PatchMatch::PatchMatch(const PatchMatchOptions& options, const Problem& problem)
 
 PatchMatch::~PatchMatch() {}
 
-void PatchMatchOptions::Print() const {
-  PrintHeading2("PatchMatchOptions");
-  PrintOption(max_image_size);
-  PrintOption(gpu_index);
-  PrintOption(depth_min);
-  PrintOption(depth_max);
-  PrintOption(window_radius);
-  PrintOption(window_step);
-  PrintOption(sigma_spatial);
-  PrintOption(sigma_color);
-  PrintOption(num_samples);
-  PrintOption(ncc_sigma);
-  PrintOption(min_triangulation_angle);
-  PrintOption(incident_angle_sigma);
-  PrintOption(num_iterations);
-  PrintOption(geom_consistency);
-  PrintOption(geom_consistency_regularizer);
-  PrintOption(geom_consistency_max_cost);
-  PrintOption(filter);
-  PrintOption(filter_min_ncc);
-  PrintOption(filter_min_triangulation_angle);
-  PrintOption(filter_min_num_consistent);
-  PrintOption(filter_geom_consistency_max_cost);
-  PrintOption(write_consistency_graph);
-  PrintOption(allow_missing_files);
-}
-
 void PatchMatch::Problem::Print() const {
   PrintHeading2("PatchMatch::Problem");
-
-  PrintOption(ref_image_idx);
-
+  LOG(INFO) << "ref_image_idx: " << ref_image_idx;
   LOG(INFO) << "src_image_idxs: ";
   if (!src_image_idxs.empty()) {
     for (size_t i = 0; i < src_image_idxs.size() - 1; ++i) {
