@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -35,7 +35,6 @@
 #include "colmap/mvs/normal_map.h"
 #include "colmap/sensor/bitmap.h"
 #include "colmap/util/cache.h"
-#include "colmap/util/misc.h"
 
 #include <memory>
 
@@ -122,6 +121,7 @@ class CachedWorkspace : public Workspace {
     CachedImage& operator=(CachedImage&& other) noexcept;
     inline size_t NumBytes() const { return num_bytes; }
     size_t num_bytes = 0;
+    std::mutex mutex;
     std::unique_ptr<Bitmap> bitmap;
     std::unique_ptr<DepthMap> depth_map;
     std::unique_ptr<NormalMap> normal_map;
