@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -35,6 +35,7 @@
 #include "colmap/util/logging.h"
 
 #include <cfloat>
+#include <optional>
 #include <random>
 #include <stdexcept>
 #include <vector>
@@ -52,7 +53,11 @@ class LORANSAC : public RANSAC<Estimator, SupportMeasurer, Sampler> {
  public:
   using typename RANSAC<Estimator, SupportMeasurer, Sampler>::Report;
 
-  explicit LORANSAC(const RANSACOptions& options);
+  explicit LORANSAC(const RANSACOptions& options,
+                    Estimator estimator = Estimator(),
+                    LocalEstimator local_estimator = LocalEstimator(),
+                    SupportMeasurer support_measurer = SupportMeasurer(),
+                    Sampler sampler = Sampler(Estimator::kMinNumSamples));
 
   // Robustly estimate model with RANSAC (RANdom SAmple Consensus).
   //
@@ -63,11 +68,10 @@ class LORANSAC : public RANSAC<Estimator, SupportMeasurer, Sampler> {
   Report Estimate(const std::vector<typename Estimator::X_t>& X,
                   const std::vector<typename Estimator::Y_t>& Y);
 
-  // Objects used in RANSAC procedure.
   using RANSAC<Estimator, SupportMeasurer, Sampler>::estimator;
-  LocalEstimator local_estimator;
-  using RANSAC<Estimator, SupportMeasurer, Sampler>::sampler;
   using RANSAC<Estimator, SupportMeasurer, Sampler>::support_measurer;
+  using RANSAC<Estimator, SupportMeasurer, Sampler>::sampler;
+  LocalEstimator local_estimator;
 
  private:
   using RANSAC<Estimator, SupportMeasurer, Sampler>::options_;
@@ -82,8 +86,16 @@ template <typename Estimator,
           typename SupportMeasurer,
           typename Sampler>
 LORANSAC<Estimator, LocalEstimator, SupportMeasurer, Sampler>::LORANSAC(
-    const RANSACOptions& options)
-    : RANSAC<Estimator, SupportMeasurer, Sampler>(options) {}
+    const RANSACOptions& options,
+    Estimator estimator,
+    LocalEstimator local_estimator,
+    SupportMeasurer support_measurer,
+    Sampler sampler)
+    : RANSAC<Estimator, SupportMeasurer, Sampler>(options,
+                                                  std::move(estimator),
+                                                  std::move(support_measurer),
+                                                  std::move(sampler)),
+      local_estimator(std::move(local_estimator)) {}
 
 template <typename Estimator,
           typename LocalEstimator,
@@ -106,7 +118,7 @@ LORANSAC<Estimator, LocalEstimator, SupportMeasurer, Sampler>::Estimate(
   }
 
   typename SupportMeasurer::Support best_support;
-  typename Estimator::M_t best_model;
+  std::optional<typename Estimator::M_t> best_model;
   bool best_model_is_local = false;
 
   bool abort = false;
@@ -151,7 +163,7 @@ LORANSAC<Estimator, LocalEstimator, SupportMeasurer, Sampler>::Estimate(
       const auto support = support_measurer.Evaluate(residuals, max_residual);
 
       // Do local optimization if better than all previous subsets.
-      if (support_measurer.Compare(support, best_support)) {
+      if (support_measurer.IsLeftBetter(support, best_support)) {
         best_support = support;
         best_model = sample_model;
         best_model_is_local = false;
@@ -187,7 +199,7 @@ LORANSAC<Estimator, LocalEstimator, SupportMeasurer, Sampler>::Estimate(
                   support_measurer.Evaluate(residuals, max_residual);
 
               // Check if locally optimized model is better.
-              if (support_measurer.Compare(local_support, best_support)) {
+              if (support_measurer.IsLeftBetter(local_support, best_support)) {
                 best_support = local_support;
                 best_model = local_model;
                 best_model_is_local = true;
@@ -223,8 +235,12 @@ LORANSAC<Estimator, LocalEstimator, SupportMeasurer, Sampler>::Estimate(
     }
   }
 
+  if (!best_model.has_value()) {
+    return report;
+  }
+
   report.support = best_support;
-  report.model = best_model;
+  report.model = best_model.value();
 
   // No valid model was found
   if (report.support.num_inliers < estimator.kMinNumSamples) {
