@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -30,10 +30,11 @@
 #pragma once
 
 #include "colmap/feature/types.h"
-#include "colmap/geometry/gps.h"
+#include "colmap/geometry/pose_prior.h"
 #include "colmap/scene/camera.h"
 #include "colmap/scene/image.h"
 #include "colmap/scene/two_view_geometry.h"
+#include "colmap/sensor/rig.h"
 #include "colmap/util/eigen_alignment.h"
 #include "colmap/util/types.h"
 
@@ -82,7 +83,9 @@ class Database {
 
   // Check if entry already exists in database. For image pairs, the order of
   // `image_id1` and `image_id2` does not matter.
+  bool ExistsRig(rig_t rig_id) const;
   bool ExistsCamera(camera_t camera_id) const;
+  bool ExistsFrame(frame_t frame_id) const;
   bool ExistsImage(image_t image_id) const;
   bool ExistsImageWithName(const std::string& name) const;
   bool ExistsPosePrior(image_t image_id) const;
@@ -91,9 +94,15 @@ class Database {
   bool ExistsMatches(image_t image_id1, image_t image_id2) const;
   bool ExistsInlierMatches(image_t image_id1, image_t image_id2) const;
 
+  // Number of rows in `rigs` table.
+  size_t NumRigs() const;
+
   // Number of rows in `cameras` table.
   size_t NumCameras() const;
 
+  //  Number of rows in `frames` table.
+  size_t NumFrames() const;
+
   //  Number of rows in `images` table.
   size_t NumImages() const;
 
@@ -150,11 +159,19 @@ class Database {
   // Read an existing entry in the database. The user is responsible for making
   // sure that the entry actually exists. For image pairs, the order of
   // `image_id1` and `image_id2` does not matter.
+
+  Rig ReadRig(rig_t rig_id) const;
+  std::optional<Rig> ReadRigWithSensor(sensor_t sensor_id) const;
+  std::vector<Rig> ReadAllRigs() const;
+
   Camera ReadCamera(camera_t camera_id) const;
   std::vector<Camera> ReadAllCameras() const;
 
+  Frame ReadFrame(frame_t frame_id) const;
+  std::vector<Frame> ReadAllFrames() const;
+
   Image ReadImage(image_t image_id) const;
-  Image ReadImageWithName(const std::string& name) const;
+  std::optional<Image> ReadImageWithName(const std::string& name) const;
   std::vector<Image> ReadAllImages() const;
 
   PosePrior ReadPosePrior(image_t image_id) const;
@@ -166,24 +183,32 @@ class Database {
   FeatureMatchesBlob ReadMatchesBlob(image_t image_id1,
                                      image_t image_id2) const;
   FeatureMatches ReadMatches(image_t image_id1, image_t image_id2) const;
+  std::vector<std::pair<image_pair_t, FeatureMatchesBlob>> ReadAllMatchesBlob()
+      const;
   std::vector<std::pair<image_pair_t, FeatureMatches>> ReadAllMatches() const;
 
   TwoViewGeometry ReadTwoViewGeometry(image_t image_id1,
                                       image_t image_id2) const;
-  void ReadTwoViewGeometries(
-      std::vector<image_pair_t>* image_pair_ids,
-      std::vector<TwoViewGeometry>* two_view_geometries) const;
+  std::vector<std::pair<image_pair_t, TwoViewGeometry>> ReadTwoViewGeometries()
+      const;
 
-  // Read all image pairs that have an entry in the `NumVerifiedImagePairs`
+  // Read all image pairs that have an entry in the `two_view_geometry`
   // table with at least one inlier match and their number of inlier matches.
-  void ReadTwoViewGeometryNumInliers(
-      std::vector<std::pair<image_t, image_t>>* image_pairs,
-      std::vector<int>* num_inliers) const;
+  std::vector<std::pair<image_pair_t, int>> ReadTwoViewGeometryNumInliers()
+      const;
+
+  // Add new rig and return its database identifier. If `use_rig_id`
+  // is false a new identifier is automatically generated.
+  rig_t WriteRig(const Rig& rig, bool use_rig_id = false) const;
 
   // Add new camera and return its database identifier. If `use_camera_id`
   // is false a new identifier is automatically generated.
   camera_t WriteCamera(const Camera& camera, bool use_camera_id = false) const;
 
+  // Add new frame and return its database identifier. If `use_frame_id`
+  // is false a new identifier is automatically generated.
+  frame_t WriteFrame(const Frame& frame, bool use_frame_id = false) const;
+
   // Add new image and return its database identifier. If `use_image_id`
   // is false a new identifier is automatically generated.
   image_t WriteImage(const Image& image, bool use_image_id = false) const;
@@ -207,14 +232,26 @@ class Database {
                             image_t image_id2,
                             const TwoViewGeometry& two_view_geometry) const;
 
+  // Update an existing rig in the database. The user is responsible for
+  // making sure that the entry already exists.
+  void UpdateRig(const Rig& rig) const;
+
   // Update an existing camera in the database. The user is responsible for
   // making sure that the entry already exists.
   void UpdateCamera(const Camera& camera) const;
 
+  // Update an existing frame in the database. The user is responsible for
+  // making sure that the entry already exists.
+  void UpdateFrame(const Frame& frame) const;
+
   // Update an existing image in the database. The user is responsible for
   // making sure that the entry already exists.
   void UpdateImage(const Image& image) const;
 
+  // Update an existing pose_prior in the database. The user is responsible for
+  // making sure that the entry already exists.
+  void UpdatePosePrior(image_t image_id, const PosePrior& pose_prior) const;
+
   // Delete matches of an image pair.
   void DeleteMatches(image_t image_id1, image_t image_id2) const;
 
@@ -224,9 +261,15 @@ class Database {
   // Clear all database tables
   void ClearAllTables() const;
 
+  // Clear the entire rigs table
+  void ClearRigs() const;
+
   // Clear the entire cameras table
   void ClearCameras() const;
 
+  // Clear the entire frames table
+  void ClearFrames() const;
+
   // Clear the entire images, keypoints, and descriptors tables
   void ClearImages() const;
 
@@ -268,7 +311,11 @@ class Database {
 
   // Create database tables, if not existing, called when opening a database.
   void CreateTables() const;
+  void CreateRigTable() const;
+  void CreateRigSensorsTable() const;
   void CreateCameraTable() const;
+  void CreateFrameTable() const;
+  void CreateFrameDataTable() const;
   void CreateImageTable() const;
   void CreatePosePriorTable() const;
   void CreateKeypointsTable() const;
@@ -295,7 +342,7 @@ class Database {
 
   // Check if elements got removed from the database to only apply
   // the VACUUM command in such case
-  mutable bool database_cleared_ = false;
+  mutable bool database_entry_deleted_ = false;
 
   // Ensure that only one database object at a time updates the schema of a
   // database. Since the schema is updated every time a database is opened, this
@@ -315,7 +362,9 @@ class Database {
   sqlite3_stmt* sql_stmt_num_descriptors_ = nullptr;
 
   // exists_*
+  sqlite3_stmt* sql_stmt_exists_rig_ = nullptr;
   sqlite3_stmt* sql_stmt_exists_camera_ = nullptr;
+  sqlite3_stmt* sql_stmt_exists_frame_ = nullptr;
   sqlite3_stmt* sql_stmt_exists_image_id_ = nullptr;
   sqlite3_stmt* sql_stmt_exists_image_name_ = nullptr;
   sqlite3_stmt* sql_stmt_exists_pose_prior_ = nullptr;
@@ -325,18 +374,31 @@ class Database {
   sqlite3_stmt* sql_stmt_exists_two_view_geometry_ = nullptr;
 
   // add_*
+  sqlite3_stmt* sql_stmt_add_rig_ = nullptr;
+  sqlite3_stmt* sql_stmt_add_rig_sensor_ = nullptr;
   sqlite3_stmt* sql_stmt_add_camera_ = nullptr;
+  sqlite3_stmt* sql_stmt_add_frame_ = nullptr;
+  sqlite3_stmt* sql_stmt_add_frame_data_ = nullptr;
   sqlite3_stmt* sql_stmt_add_image_ = nullptr;
 
   // update_*
+  sqlite3_stmt* sql_stmt_update_rig_ = nullptr;
   sqlite3_stmt* sql_stmt_update_camera_ = nullptr;
+  sqlite3_stmt* sql_stmt_update_frame_ = nullptr;
   sqlite3_stmt* sql_stmt_update_image_ = nullptr;
+  sqlite3_stmt* sql_stmt_update_pose_prior_ = nullptr;
 
   // read_*
+  sqlite3_stmt* sql_stmt_read_rig_ = nullptr;
+  sqlite3_stmt* sql_stmt_read_rigs_ = nullptr;
+  sqlite3_stmt* sql_stmt_read_rig_with_sensor_ = nullptr;
+  sqlite3_stmt* sql_stmt_read_rig_with_ref_sensor_ = nullptr;
   sqlite3_stmt* sql_stmt_read_camera_ = nullptr;
   sqlite3_stmt* sql_stmt_read_cameras_ = nullptr;
+  sqlite3_stmt* sql_stmt_read_frame_ = nullptr;
+  sqlite3_stmt* sql_stmt_read_frames_ = nullptr;
   sqlite3_stmt* sql_stmt_read_image_id_ = nullptr;
-  sqlite3_stmt* sql_stmt_read_image_name_ = nullptr;
+  sqlite3_stmt* sql_stmt_read_image_with_name_ = nullptr;
   sqlite3_stmt* sql_stmt_read_images_ = nullptr;
   sqlite3_stmt* sql_stmt_read_pose_prior_ = nullptr;
   sqlite3_stmt* sql_stmt_read_keypoints_ = nullptr;
@@ -355,11 +417,15 @@ class Database {
   sqlite3_stmt* sql_stmt_write_two_view_geometry_ = nullptr;
 
   // delete_*
+  sqlite3_stmt* sql_stmt_delete_rig_sensors_ = nullptr;
+  sqlite3_stmt* sql_stmt_delete_frame_data_ = nullptr;
   sqlite3_stmt* sql_stmt_delete_matches_ = nullptr;
   sqlite3_stmt* sql_stmt_delete_two_view_geometry_ = nullptr;
 
   // clear_*
+  sqlite3_stmt* sql_stmt_clear_rigs_ = nullptr;
   sqlite3_stmt* sql_stmt_clear_cameras_ = nullptr;
+  sqlite3_stmt* sql_stmt_clear_frames_ = nullptr;
   sqlite3_stmt* sql_stmt_clear_images_ = nullptr;
   sqlite3_stmt* sql_stmt_clear_pose_priors_ = nullptr;
   sqlite3_stmt* sql_stmt_clear_descriptors_ = nullptr;
