@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -157,6 +157,9 @@ class IncrementalTriangulator {
   };
 
  private:
+  friend std::ostream& operator<<(std::ostream& stream,
+                                  const IncrementalTriangulator& triangulator);
+
   // Clear cache of bogus camera parameters and merge trials.
   void ClearCaches();
 
@@ -212,4 +215,7 @@ class IncrementalTriangulator {
   std::unordered_set<point3D_t> modified_point3D_ids_;
 };
 
+std::ostream& operator<<(std::ostream& stream,
+                         const IncrementalTriangulator& triangulator);
+
 }  // namespace colmap
