@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -53,7 +53,7 @@ bool MergeAndFilterReconstructions(const double max_reproj_error,
 const int ObservationManager::kNumPoint3DVisibilityPyramidLevels = 6;
 
 ObservationManager::ObservationManager(
-    Reconstruction& reconstruction,
+    class Reconstruction& reconstruction,
     std::shared_ptr<const CorrespondenceGraph> correspondence_graph)
     : reconstruction_(reconstruction),
       correspondence_graph_(std::move(correspondence_graph)) {
@@ -70,26 +70,25 @@ ObservationManager::ObservationManager(
 
   // Add image stats.
   image_stats_.reserve(reconstruction_.NumImages());
-  for (const auto& id_image : reconstruction_.Images()) {
-    const Image& image = id_image.second;
-    const Camera& camera = reconstruction_.Camera(image.CameraId());
+  for (const auto& [image_id, image] : reconstruction_.Images()) {
+    const Camera& camera = *image.CameraPtr();
     ImageStat image_stat;
     image_stat.point3D_visibility_pyramid = VisibilityPyramid(
         kNumPoint3DVisibilityPyramidLevels, camera.width, camera.height);
     image_stat.num_correspondences_have_point3D.resize(image.NumPoints2D(), 0);
     image_stat.num_visible_points3D = 0;
-    if (correspondence_graph_) {
+    if (correspondence_graph_ && correspondence_graph_->ExistsImage(image_id)) {
       image_stat.num_observations =
-          correspondence_graph_->NumObservationsForImage(id_image.first);
+          correspondence_graph_->NumObservationsForImage(image_id);
       image_stat.num_correspondences =
-          correspondence_graph_->NumCorrespondencesForImage(id_image.first);
+          correspondence_graph_->NumCorrespondencesForImage(image_id);
     }
-    image_stats_.emplace(id_image.first, image_stat);
+    image_stats_.emplace(image_id, image_stat);
   }
 
   // If an existing model was loaded from disk and there were already images
   // registered previously, we need to set observations as triangulated.
-  for (const auto image_id : reconstruction_.RegImageIds()) {
+  for (const image_t image_id : reconstruction_.RegImageIds()) {
     const Image& image = reconstruction_.Image(image_id);
     for (point2D_t point2D_idx = 0; point2D_idx < image.NumPoints2D();
          ++point2D_idx) {
@@ -105,7 +104,7 @@ ObservationManager::ObservationManager(
 void ObservationManager::IncrementCorrespondenceHasPoint3D(
     const image_t image_id, const point2D_t point2D_idx) {
   const Image& image = reconstruction_.Image(image_id);
-  const struct Point2D& point2D = image.Point2D(point2D_idx);
+  const Point2D& point2D = image.Point2D(point2D_idx);
   ImageStat& stats = image_stats_.at(image_id);
 
   stats.num_correspondences_have_point3D[point2D_idx] += 1;
@@ -121,7 +120,7 @@ void ObservationManager::IncrementCorrespondenceHasPoint3D(
 void ObservationManager::DecrementCorrespondenceHasPoint3D(
     const image_t image_id, const point2D_t point2D_idx) {
   const Image& image = reconstruction_.Image(image_id);
-  const struct Point2D& point2D = image.Point2D(point2D_idx);
+  const Point2D& point2D = image.Point2D(point2D_idx);
   ImageStat& stats = image_stats_.at(image_id);
 
   stats.num_correspondences_have_point3D[point2D_idx] -= 1;
@@ -142,7 +141,7 @@ void ObservationManager::SetObservationAsTriangulated(
     return;
   }
   const Image& image = reconstruction_.Image(image_id);
-  THROW_CHECK(image.IsRegistered());
+  THROW_CHECK(image.HasPose());
 
   const Point2D& point2D = image.Point2D(point2D_idx);
   THROW_CHECK(point2D.HasPoint3D());
@@ -175,7 +174,7 @@ void ObservationManager::ResetTriObservations(const image_t image_id,
     return;
   }
   const Image& image = reconstruction_.Image(image_id);
-  THROW_CHECK(image.IsRegistered());
+  THROW_CHECK(image.HasPose());
   const Point2D& point2D = image.Point2D(point2D_idx);
   THROW_CHECK(point2D.HasPoint3D());
 
@@ -281,12 +280,12 @@ size_t ObservationManager::FilterPoints3D(
     const double max_reproj_error,
     const double min_tri_angle,
     const std::unordered_set<point3D_t>& point3D_ids) {
-  size_t num_filtered = 0;
-  num_filtered +=
+  size_t num_filtered_observations = 0;
+  num_filtered_observations +=
       FilterPoints3DWithLargeReprojectionError(max_reproj_error, point3D_ids);
-  num_filtered +=
+  num_filtered_observations +=
       FilterPoints3DWithSmallTriangulationAngle(min_tri_angle, point3D_ids);
-  return num_filtered;
+  return num_filtered_observations;
 }
 
 size_t ObservationManager::FilterPoints3DInImages(
@@ -312,28 +311,30 @@ size_t ObservationManager::FilterAllPoints3D(const double max_reproj_error,
   // a point stable through a large triangulation angle.
   const std::unordered_set<point3D_t>& point3D_ids =
       reconstruction_.Point3DIds();
-  size_t num_filtered = 0;
-  num_filtered +=
+  size_t num_filtered_observations = 0;
+  num_filtered_observations +=
       FilterPoints3DWithLargeReprojectionError(max_reproj_error, point3D_ids);
-  num_filtered +=
+  num_filtered_observations +=
       FilterPoints3DWithSmallTriangulationAngle(min_tri_angle, point3D_ids);
-  return num_filtered;
+  return num_filtered_observations;
 }
 
 size_t ObservationManager::FilterObservationsWithNegativeDepth() {
   size_t num_filtered = 0;
-  for (const auto image_id : reconstruction_.RegImageIds()) {
-    const Image& image = reconstruction_.Image(image_id);
-    const Eigen::Matrix3x4d cam_from_world = image.CamFromWorld().ToMatrix();
-    for (point2D_t point2D_idx = 0; point2D_idx < image.NumPoints2D();
-         ++point2D_idx) {
-      const Point2D& point2D = image.Point2D(point2D_idx);
-      if (point2D.HasPoint3D()) {
-        const struct Point3D& point3D =
-            reconstruction_.Point3D(point2D.point3D_id);
-        if (!HasPointPositiveDepth(cam_from_world, point3D.xyz)) {
-          DeleteObservation(image_id, point2D_idx);
-          num_filtered += 1;
+  for (const frame_t frame_id : reconstruction_.RegFrameIds()) {
+    for (const data_t& data_id : reconstruction_.Frame(frame_id).ImageIds()) {
+      const Image& image = reconstruction_.Image(data_id.id);
+      const Eigen::Matrix3x4d cam_from_world = image.CamFromWorld().ToMatrix();
+      for (point2D_t point2D_idx = 0; point2D_idx < image.NumPoints2D();
+           ++point2D_idx) {
+        const Point2D& point2D = image.Point2D(point2D_idx);
+        if (point2D.HasPoint3D()) {
+          const struct Point3D& point3D =
+              reconstruction_.Point3D(point2D.point3D_id);
+          if (!HasPointPositiveDepth(cam_from_world, point3D.xyz)) {
+            DeleteObservation(data_id.id, point2D_idx);
+            num_filtered += 1;
+          }
         }
       }
     }
@@ -344,8 +345,8 @@ size_t ObservationManager::FilterObservationsWithNegativeDepth() {
 size_t ObservationManager::FilterPoints3DWithSmallTriangulationAngle(
     const double min_tri_angle,
     const std::unordered_set<point3D_t>& point3D_ids) {
-  // Number of filtered points.
-  size_t num_filtered = 0;
+  // Number of filtered observations.
+  size_t num_filtered_observations = 0;
 
   // Minimum triangulation angle in radians.
   const double min_tri_angle_rad = DegToRad(min_tri_angle);
@@ -395,12 +396,12 @@ size_t ObservationManager::FilterPoints3DWithSmallTriangulationAngle(
     }
 
     if (!keep_point) {
-      num_filtered += 1;
+      num_filtered_observations += point3D.track.Length();
       DeletePoint3D(point3D_id);
     }
   }
 
-  return num_filtered;
+  return num_filtered_observations;
 }
 
 size_t ObservationManager::FilterPoints3DWithLargeReprojectionError(
@@ -408,8 +409,8 @@ size_t ObservationManager::FilterPoints3DWithLargeReprojectionError(
     const std::unordered_set<point3D_t>& point3D_ids) {
   const double max_squared_reproj_error = max_reproj_error * max_reproj_error;
 
-  // Number of filtered points.
-  size_t num_filtered = 0;
+  // Number of filtered observations.
+  size_t num_filtered_observations = 0;
 
   for (const auto point3D_id : point3D_ids) {
     if (!reconstruction_.ExistsPoint3D(point3D_id)) {
@@ -419,7 +420,7 @@ size_t ObservationManager::FilterPoints3DWithLargeReprojectionError(
     struct Point3D& point3D = reconstruction_.Point3D(point3D_id);
 
     if (point3D.track.Length() < 2) {
-      num_filtered += point3D.track.Length();
+      num_filtered_observations += point3D.track.Length();
       DeletePoint3D(point3D_id);
       continue;
     }
@@ -430,7 +431,7 @@ size_t ObservationManager::FilterPoints3DWithLargeReprojectionError(
 
     for (const auto& track_el : point3D.track.Elements()) {
       const Image& image = reconstruction_.Image(track_el.image_id);
-      const struct Camera& camera = reconstruction_.Camera(image.CameraId());
+      const struct Camera& camera = *image.CameraPtr();
       const Point2D& point2D = image.Point2D(track_el.point2D_idx);
       const double squared_reproj_error = CalculateSquaredReprojectionError(
           point2D.xy, point3D.xyz, image.CamFromWorld(), camera);
@@ -442,10 +443,10 @@ size_t ObservationManager::FilterPoints3DWithLargeReprojectionError(
     }
 
     if (track_els_to_delete.size() >= point3D.track.Length() - 1) {
-      num_filtered += point3D.track.Length();
+      num_filtered_observations += point3D.track.Length();
       DeletePoint3D(point3D_id);
     } else {
-      num_filtered += track_els_to_delete.size();
+      num_filtered_observations += track_els_to_delete.size();
       for (const auto& track_el : track_els_to_delete) {
         DeleteObservation(track_el.image_id, track_el.point2D_idx);
       }
@@ -453,42 +454,67 @@ size_t ObservationManager::FilterPoints3DWithLargeReprojectionError(
     }
   }
 
-  return num_filtered;
+  return num_filtered_observations;
 }
 
-void ObservationManager::DeRegisterImage(const image_t image_id) {
-  Image& image = reconstruction_.Image(image_id);
-  const auto num_points2D = image.NumPoints2D();
-  for (point2D_t point2D_idx = 0; point2D_idx < num_points2D; ++point2D_idx) {
-    if (image.Point2D(point2D_idx).HasPoint3D()) {
-      DeleteObservation(image_id, point2D_idx);
+void ObservationManager::DeRegisterFrame(const frame_t frame_id) {
+  const Frame& frame = reconstruction_.Frame(frame_id);
+  for (const data_t& data_id : frame.DataIds()) {
+    Image& image = reconstruction_.Image(data_id.id);
+    const auto num_points2D = image.NumPoints2D();
+    for (point2D_t point2D_idx = 0; point2D_idx < num_points2D; ++point2D_idx) {
+      if (image.Point2D(point2D_idx).HasPoint3D()) {
+        DeleteObservation(data_id.id, point2D_idx);
+      }
     }
   }
-  reconstruction_.DeRegisterImage(image_id);
+  reconstruction_.DeRegisterFrame(frame_id);
 }
 
-std::vector<image_t> ObservationManager::FilterImages(
+std::vector<frame_t> ObservationManager::FilterFrames(
     const double min_focal_length_ratio,
     const double max_focal_length_ratio,
     const double max_extra_param) {
-  std::vector<image_t> filtered_image_ids;
-  for (const image_t image_id : reconstruction_.RegImageIds()) {
-    const Image& image = reconstruction_.Image(image_id);
-    if (image.NumPoints3D() == 0 || reconstruction_.Camera(image.CameraId())
-                                        .HasBogusParams(min_focal_length_ratio,
-                                                        max_focal_length_ratio,
-                                                        max_extra_param)) {
-      filtered_image_ids.push_back(image_id);
+  std::vector<frame_t> filtered_frame_ids;
+  for (const frame_t frame_id : reconstruction_.RegFrameIds()) {
+    const Frame& frame = reconstruction_.Frame(frame_id);
+    int num_points3D = 0;
+    for (const data_t& data_id : frame.ImageIds()) {
+      const Image& image = reconstruction_.Image(data_id.id);
+      num_points3D += image.NumPoints3D();
+      if (image.CameraPtr()->HasBogusParams(min_focal_length_ratio,
+                                            max_focal_length_ratio,
+                                            max_extra_param)) {
+        // Flag the frame for filtering.
+        num_points3D = 0;
+        break;
+      }
+    }
+    if (num_points3D == 0) {
+      filtered_frame_ids.push_back(frame_id);
     }
   }
 
-  // Only de-register after iterating over reg_image_ids_ to avoid
+  // Only de-register after iterating over reg_frame_ids_ to avoid
   // simultaneous iteration and modification of the vector.
-  for (const image_t image_id : filtered_image_ids) {
-    DeRegisterImage(image_id);
+  for (const frame_t frame_id : filtered_frame_ids) {
+    DeRegisterFrame(frame_id);
   }
 
-  return filtered_image_ids;
+  return filtered_frame_ids;
+}
+
+std::ostream& operator<<(std::ostream& stream,
+                         const ObservationManager& obs_manager) {
+  stream << "ObservationManager(reconstruction=" << obs_manager.reconstruction_
+         << ", correspondence_graph=";
+  if (obs_manager.correspondence_graph_ == nullptr) {
+    stream << "null";
+  } else {
+    stream << obs_manager.correspondence_graph_;
+  }
+  stream << ")";
+  return stream;
 }
 
 }  // namespace colmap
