@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -42,6 +42,8 @@ BundleAdjustmentWidget::BundleAdjustmentWidget(MainWindow* main_window,
       options_(options),
       reconstruction_(nullptr),
       thread_control_widget_(new ThreadControlWidget(this)) {
+  setWindowFlags(Qt::Dialog);
+  setWindowModality(Qt::ApplicationModal);
   setWindowTitle("Bundle adjustment");
 
   AddOptionInt(&options->bundle_adjustment->solver_options.max_num_iterations,
@@ -72,8 +74,10 @@ BundleAdjustmentWidget::BundleAdjustmentWidget(MainWindow* main_window,
                 "refine_principal_point");
   AddOptionBool(&options->bundle_adjustment->refine_extra_params,
                 "refine_extra_params");
-  AddOptionBool(&options->bundle_adjustment->refine_extrinsics,
-                "refine_extrinsics");
+  AddOptionBool(&options->bundle_adjustment->refine_rig_from_world,
+                "refine_rig_from_world");
+  AddOptionBool(&options->bundle_adjustment->refine_sensor_from_rig,
+                "refine_sensor_from_rig");
 
   QPushButton* run_button = new QPushButton(tr("Run"), this);
   grid_layout_->addWidget(run_button, grid_layout_->rowCount(), 1);
