@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -29,7 +29,7 @@
 
 #pragma once
 
-#include "colmap/controllers/incremental_mapper.h"
+#include "colmap/controllers/incremental_pipeline.h"
 #include "colmap/scene/reconstruction.h"
 #include "colmap/sensor/bitmap.h"
 #include "colmap/ui/automatic_reconstruction_widget.h"
@@ -130,6 +130,7 @@ class MainWindow : public QMainWindow {
 
   void SetOptions();
   void ResetOptions();
+  void SetLogLevel();
 
   void About();
   void Documentation();
@@ -146,8 +147,7 @@ class MainWindow : public QMainWindow {
   OptionManager options_;
 
   std::shared_ptr<ReconstructionManager> reconstruction_manager_;
-  std::unique_ptr<ControllerThread<IncrementalMapperController>>
-      mapper_controller_;
+  std::unique_ptr<ControllerThread<IncrementalPipeline>> mapper_controller_;
 
   Timer timer_;
 
@@ -225,6 +225,7 @@ class MainWindow : public QMainWindow {
   QAction* action_extract_colors_;
   QAction* action_set_options_;
   QAction* action_reset_options_;
+  QAction* action_set_log_level_;
 
   QAction* action_about_;
   QAction* action_documentation_;
