@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -32,21 +32,24 @@
 #include "colmap/util/logging.h"
 
 #include <functional>
+#include <future>
 #include <iostream>
 #include <limits>
 #include <list>
+#include <memory>
+#include <shared_mutex>
 #include <unordered_map>
 
 namespace colmap {
 
 // Least Recently Used cache implementation. Whenever the cache size is
-// exceeded, the least recently used (by Get and GetMutable) is deleted.
+// exceeded, the least recently used (by Get) is deleted.
 template <typename key_t, typename value_t>
 class LRUCache {
  public:
-  LRUCache(size_t max_num_elems,
-           const std::function<value_t(const key_t&)>& getter_func);
-  virtual ~LRUCache() = default;
+  using LoadFn = std::function<std::shared_ptr<value_t>(const key_t&)>;
+
+  LRUCache(size_t max_num_elems, LoadFn load_fn);
 
   // The number of elements in the cache.
   size_t NumElems() const;
@@ -56,20 +59,20 @@ class LRUCache {
   bool Exists(const key_t& key) const;
 
   // Get the value of an element either from the cache or compute the new value.
-  const value_t& Get(const key_t& key);
-  value_t& GetMutable(const key_t& key);
+  std::shared_ptr<value_t> Get(const key_t& key);
 
-  // Manually set the value of an element.
-  virtual void Set(const key_t& key, value_t value);
+  // Manually evict an element from the cache.
+  // Returns true if the element was evicted.
+  bool Evict(const key_t& key);
 
   // Pop least recently used element from cache.
-  virtual void Pop();
+  void Pop();
 
   // Clear all elements from cache.
-  virtual void Clear();
+  void Clear();
 
- protected:
-  typedef typename std::pair<key_t, value_t> key_value_pair_t;
+ private:
+  typedef typename std::pair<key_t, std::shared_ptr<value_t>> key_value_pair_t;
   typedef typename std::list<key_value_pair_t>::iterator list_iterator_t;
 
   // Maximum number of least-recently-used elements the cache remembers.
@@ -82,39 +85,103 @@ class LRUCache {
   std::unordered_map<key_t, list_iterator_t> elems_map_;
 
   // Function to compute new values if not in the cache.
-  const std::function<value_t(const key_t&)> getter_func_;
+  const LoadFn load_fn_;
+};
+
+// Thread-safe Least Recently Used cache implementation.
+template <typename key_t, typename value_t>
+class ThreadSafeLRUCache {
+ public:
+  using LoadFn = std::function<std::shared_ptr<value_t>(const key_t&)>;
+
+  ThreadSafeLRUCache(size_t max_num_elems, LoadFn load_fn);
+
+  // The number of elements in the cache.
+  size_t NumElems() const;
+  size_t MaxNumElems() const;
+
+  // Check whether the element with the given key exists.
+  bool Exists(const key_t& key) const;
+
+  // Get the value of an element either from the cache or compute the new value.
+  std::shared_ptr<value_t> Get(const key_t& key);
+
+  // Manually evict an element from the cache.
+  // Returns true if the element was evicted.
+  bool Evict(const key_t& key);
+
+  // Pop least recently used element from cache.
+  void Pop();
+
+  // Clear all elements from cache.
+  void Clear();
+
+ protected:
+  struct Entry {
+    Entry() : future(promise.get_future()) {}
+    std::promise<std::shared_ptr<value_t>> promise;
+    std::shared_future<std::shared_ptr<value_t>> future;
+    bool is_loading = false;
+  };
+
+  // Function to compute new values if not in the cache.
+  const LoadFn load_fn_;
+
+  mutable std::shared_mutex cache_mutex_;
+  LRUCache<key_t, Entry> cache_;
 };
 
 // Least Recently Used cache implementation that is constrained by a maximum
 // memory limitation of its elements. Whenever the memory limit is exceeded, the
-// least recently used (by Get and GetMutable) is deleted. Each element must
-// implement a `size_t NumBytes()` method that returns its size in memory.
+// least recently used (by Get) is deleted. Each element must implement a
+// `size_t NumBytes()` method that returns its size in memory.
 template <typename key_t, typename value_t>
-class MemoryConstrainedLRUCache : public LRUCache<key_t, value_t> {
+class MemoryConstrainedLRUCache {
  public:
-  MemoryConstrainedLRUCache(
-      size_t max_num_bytes,
-      const std::function<value_t(const key_t&)>& getter_func);
+  using LoadFn = std::function<std::shared_ptr<value_t>(const key_t&)>;
 
+  MemoryConstrainedLRUCache(size_t max_num_bytes, LoadFn load_fn);
+
+  // The size in bytes of the elements in the cache.
   size_t NumBytes() const;
   size_t MaxNumBytes() const;
   void UpdateNumBytes(const key_t& key);
 
-  void Set(const key_t& key, value_t value) override;
-  void Pop() override;
-  void Clear() override;
+  // The number of elements in the cache.
+  size_t NumElems() const;
+  size_t MaxNumElems() const;
+
+  // Check whether the element with the given key exists.
+  bool Exists(const key_t& key) const;
+
+  // Get the value of an element either from the cache or compute the new value.
+  std::shared_ptr<value_t> Get(const key_t& key);
+
+  // Manually evict an element from the cache.
+  // Returns true if the element was evicted.
+  bool Evict(const key_t& key);
+
+  // Pop least recently used element from cache.
+  void Pop();
+
+  // Clear all elements from cache.
+  void Clear();
 
  private:
-  using typename LRUCache<key_t, value_t>::key_value_pair_t;
-  using typename LRUCache<key_t, value_t>::list_iterator_t;
-  using LRUCache<key_t, value_t>::max_num_elems_;
-  using LRUCache<key_t, value_t>::elems_list_;
-  using LRUCache<key_t, value_t>::elems_map_;
-  using LRUCache<key_t, value_t>::getter_func_;
+  typedef typename std::pair<key_t, std::shared_ptr<value_t>> key_value_pair_t;
+  typedef typename std::list<key_value_pair_t>::iterator list_iterator_t;
 
   const size_t max_num_bytes_;
   size_t num_bytes_;
-  std::unordered_map<key_t, size_t> elems_num_bytes_;
+
+  // List to keep track of the least-recently-used elements.
+  std::list<key_value_pair_t> elems_list_;
+
+  // Mapping from key to (location in list, num_bytes).
+  std::unordered_map<key_t, std::pair<list_iterator_t, size_t>> elems_map_;
+
+  // Function to compute new values if not in the cache.
+  const LoadFn load_fn_;
 };
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -122,11 +189,9 @@ class MemoryConstrainedLRUCache : public LRUCache<key_t, value_t> {
 ////////////////////////////////////////////////////////////////////////////////
 
 template <typename key_t, typename value_t>
-LRUCache<key_t, value_t>::LRUCache(
-    const size_t max_num_elems,
-    const std::function<value_t(const key_t&)>& getter_func)
-    : max_num_elems_(max_num_elems), getter_func_(getter_func) {
-  THROW_CHECK(getter_func);
+LRUCache<key_t, value_t>::LRUCache(const size_t max_num_elems, LoadFn load_fn)
+    : max_num_elems_(max_num_elems), load_fn_(std::move(load_fn)) {
+  THROW_CHECK_NOTNULL(load_fn_);
   THROW_CHECK_GT(max_num_elems, 0);
 }
 
@@ -146,16 +211,20 @@ bool LRUCache<key_t, value_t>::Exists(const key_t& key) const {
 }
 
 template <typename key_t, typename value_t>
-const value_t& LRUCache<key_t, value_t>::Get(const key_t& key) {
-  return GetMutable(key);
-}
-
-template <typename key_t, typename value_t>
-value_t& LRUCache<key_t, value_t>::GetMutable(const key_t& key) {
+std::shared_ptr<value_t> LRUCache<key_t, value_t>::Get(const key_t& key) {
   const auto it = elems_map_.find(key);
   if (it == elems_map_.end()) {
-    Set(key, std::move(getter_func_(key)));
-    return elems_map_[key]->second;
+    auto it = elems_map_.find(key);
+    elems_list_.emplace_front(key, load_fn_(key));
+    if (it != elems_map_.end()) {
+      elems_list_.erase(it->second);
+      elems_map_.erase(it);
+    }
+    it = elems_map_.emplace_hint(it, key, elems_list_.begin());
+    if (elems_map_.size() > max_num_elems_) {
+      Pop();
+    }
+    return it->second->second;
   } else {
     elems_list_.splice(elems_list_.begin(), elems_list_, it->second);
     return it->second->second;
@@ -163,17 +232,14 @@ value_t& LRUCache<key_t, value_t>::GetMutable(const key_t& key) {
 }
 
 template <typename key_t, typename value_t>
-void LRUCache<key_t, value_t>::Set(const key_t& key, value_t value) {
-  auto it = elems_map_.find(key);
-  elems_list_.emplace_front(key, std::move(value));
+bool LRUCache<key_t, value_t>::Evict(const key_t& key) {
+  const auto it = elems_map_.find(key);
   if (it != elems_map_.end()) {
     elems_list_.erase(it->second);
     elems_map_.erase(it);
+    return true;
   }
-  elems_map_[key] = elems_list_.begin();
-  if (elems_map_.size() > max_num_elems_) {
-    Pop();
-  }
+  return false;
 }
 
 template <typename key_t, typename value_t>
@@ -192,16 +258,100 @@ void LRUCache<key_t, value_t>::Clear() {
   elems_map_.clear();
 }
 
+template <typename key_t, typename value_t>
+ThreadSafeLRUCache<key_t, value_t>::ThreadSafeLRUCache(
+    const size_t max_num_elems, LoadFn load_fn)
+    : load_fn_(load_fn), cache_(max_num_elems, [](const key_t&) {
+        return std::make_shared<Entry>();
+      }) {
+  THROW_CHECK_NOTNULL(load_fn_);
+}
+
+template <typename key_t, typename value_t>
+size_t ThreadSafeLRUCache<key_t, value_t>::NumElems() const {
+  std::shared_lock lock(cache_mutex_);
+  return cache_.NumElems();
+}
+
+template <typename key_t, typename value_t>
+size_t ThreadSafeLRUCache<key_t, value_t>::MaxNumElems() const {
+  std::shared_lock lock(cache_mutex_);
+  return cache_.MaxNumElems();
+}
+
+template <typename key_t, typename value_t>
+bool ThreadSafeLRUCache<key_t, value_t>::Exists(const key_t& key) const {
+  std::shared_lock lock(cache_mutex_);
+  return cache_.Exists(key);
+}
+
+template <typename key_t, typename value_t>
+std::shared_ptr<value_t> ThreadSafeLRUCache<key_t, value_t>::Get(
+    const key_t& key) {
+  bool should_load = false;
+  std::shared_ptr<Entry> entry;
+  std::shared_future<std::shared_ptr<value_t>> shared_future;
+
+  {
+    std::unique_lock lock(cache_mutex_);
+    entry = cache_.Get(key);
+    if (entry == nullptr) {
+      return nullptr;
+    }
+    shared_future = entry->future;
+    if (!entry->is_loading) {
+      should_load = true;
+      entry->is_loading = true;
+    }
+  }
+
+  if (should_load) {
+    try {
+      entry->promise.set_value(THROW_CHECK_NOTNULL(load_fn_(key)));
+    } catch (...) {
+      // Evict the cache entry after load failed and set the exception.
+      std::unique_lock lock(cache_mutex_);
+      cache_.Evict(key);
+      entry->promise.set_exception(std::current_exception());
+    }
+  }
+
+  return shared_future.get();
+}
+
+template <typename key_t, typename value_t>
+bool ThreadSafeLRUCache<key_t, value_t>::Evict(const key_t& key) {
+  std::unique_lock lock(cache_mutex_);
+  return cache_.Evict(key);
+}
+
+template <typename key_t, typename value_t>
+void ThreadSafeLRUCache<key_t, value_t>::Pop() {
+  std::unique_lock lock(cache_mutex_);
+  return cache_.Pop();
+}
+
+template <typename key_t, typename value_t>
+void ThreadSafeLRUCache<key_t, value_t>::Clear() {
+  std::unique_lock lock(cache_mutex_);
+  return cache_.Clear();
+}
+
 template <typename key_t, typename value_t>
 MemoryConstrainedLRUCache<key_t, value_t>::MemoryConstrainedLRUCache(
-    const size_t max_num_bytes,
-    const std::function<value_t(const key_t&)>& getter_func)
-    : LRUCache<key_t, value_t>(std::numeric_limits<size_t>::max(), getter_func),
-      max_num_bytes_(max_num_bytes),
-      num_bytes_(0) {
+    const size_t max_num_bytes, LoadFn load_fn)
+    : max_num_bytes_(max_num_bytes),
+      num_bytes_(0),
+      load_fn_(std::move(load_fn)) {
+  THROW_CHECK_NOTNULL(load_fn_);
   THROW_CHECK_GT(max_num_bytes, 0);
 }
 
+template <typename key_t, typename value_t>
+size_t MemoryConstrainedLRUCache<key_t, value_t>::NumElems() const {
+  return elems_map_.size();
+}
+
 template <typename key_t, typename value_t>
 size_t MemoryConstrainedLRUCache<key_t, value_t>::NumBytes() const {
   return num_bytes_;
@@ -213,23 +363,48 @@ size_t MemoryConstrainedLRUCache<key_t, value_t>::MaxNumBytes() const {
 }
 
 template <typename key_t, typename value_t>
-void MemoryConstrainedLRUCache<key_t, value_t>::Set(const key_t& key,
-                                                    value_t value) {
-  const size_t num_bytes = value.NumBytes();
-  auto it = elems_map_.find(key);
-  elems_list_.emplace_front(key, std::move(value));
+bool MemoryConstrainedLRUCache<key_t, value_t>::Exists(const key_t& key) const {
+  return elems_map_.find(key) != elems_map_.end();
+}
+
+template <typename key_t, typename value_t>
+std::shared_ptr<value_t> MemoryConstrainedLRUCache<key_t, value_t>::Get(
+    const key_t& key) {
+  const auto it = elems_map_.find(key);
+  if (it == elems_map_.end()) {
+    std::shared_ptr<value_t> value = load_fn_(key);
+    const size_t num_bytes = value->NumBytes();
+    auto it = elems_map_.find(key);
+    elems_list_.emplace_front(key, std::move(value));
+    if (it != elems_map_.end()) {
+      elems_list_.erase(it->second.first);
+      elems_map_.erase(it);
+    }
+    it = elems_map_.emplace_hint(
+        it, key, std::make_pair(elems_list_.begin(), num_bytes));
+
+    num_bytes_ += num_bytes;
+    while (num_bytes_ > max_num_bytes_ && elems_map_.size() > 1) {
+      Pop();
+    }
+
+    return it->second.first->second;
+  } else {
+    elems_list_.splice(elems_list_.begin(), elems_list_, it->second.first);
+    return it->second.first->second;
+  }
+}
+
+template <typename key_t, typename value_t>
+bool MemoryConstrainedLRUCache<key_t, value_t>::Evict(const key_t& key) {
+  const auto it = elems_map_.find(key);
   if (it != elems_map_.end()) {
-    elems_list_.erase(it->second);
+    num_bytes_ -= it->second.second;
+    elems_list_.erase(it->second.first);
     elems_map_.erase(it);
+    return true;
   }
-  elems_map_[key] = elems_list_.begin();
-
-  num_bytes_ += num_bytes;
-  elems_num_bytes_.emplace(key, num_bytes);
-
-  while (num_bytes_ > max_num_bytes_ && elems_map_.size() > 1) {
-    Pop();
-  }
+  return false;
 }
 
 template <typename key_t, typename value_t>
@@ -237,10 +412,10 @@ void MemoryConstrainedLRUCache<key_t, value_t>::Pop() {
   if (!elems_list_.empty()) {
     auto last = elems_list_.end();
     --last;
-    num_bytes_ -= elems_num_bytes_.at(last->first);
+    const auto it = elems_map_.find(last->first);
+    num_bytes_ -= it->second.second;
     THROW_CHECK_GE(num_bytes_, 0);
-    elems_num_bytes_.erase(last->first);
-    elems_map_.erase(last->first);
+    elems_map_.erase(it);
     elems_list_.pop_back();
   }
 }
@@ -248,10 +423,10 @@ void MemoryConstrainedLRUCache<key_t, value_t>::Pop() {
 template <typename key_t, typename value_t>
 void MemoryConstrainedLRUCache<key_t, value_t>::UpdateNumBytes(
     const key_t& key) {
-  auto& num_bytes = elems_num_bytes_.at(key);
+  size_t& num_bytes = elems_map_.at(key).second;
   num_bytes_ -= num_bytes;
   THROW_CHECK_GE(num_bytes_, 0);
-  num_bytes = LRUCache<key_t, value_t>::Get(key).NumBytes();
+  num_bytes = Get(key)->NumBytes();
   num_bytes_ += num_bytes;
 
   while (num_bytes_ > max_num_bytes_ && elems_map_.size() > 1) {
@@ -261,9 +436,9 @@ void MemoryConstrainedLRUCache<key_t, value_t>::UpdateNumBytes(
 
 template <typename key_t, typename value_t>
 void MemoryConstrainedLRUCache<key_t, value_t>::Clear() {
-  LRUCache<key_t, value_t>::Clear();
+  elems_list_.clear();
+  elems_map_.clear();
   num_bytes_ = 0;
-  elems_num_bytes_.clear();
 }
 
 }  // namespace colmap
