@@ -1,4 +1,4 @@
-// Copyright (c) 2023, ETH Zurich and UNC Chapel Hill.
+// Copyright (c), ETH Zurich and UNC Chapel Hill.
 // All rights reserved.
 //
 // Redistribution and use in source and binary forms, with or without
@@ -54,7 +54,7 @@ inline int SQLite3CallHelper(int result_code,
   }
 }
 
-#define SQLITE3_CALL(func) SQLite3CallHelper(func, __FILE__, __LINE__)
+#define SQLITE3_CALL(func) colmap::SQLite3CallHelper(func, __FILE__, __LINE__)
 
 #define SQLITE3_EXEC(database, sql, callback)                             \
   {                                                                       \
