@@ -9,7 +9,7 @@ namespace py = pybind11;
 
 void BindOptim(py::module& m) {
   auto PyRANSACOptions =
-      py::class_<RANSACOptions>(m, "RANSACOptions")
+      py::classh<RANSACOptions>(m, "RANSACOptions")
           .def(py::init<>([]() {
             RANSACOptions options;
             options.max_error = 4.0;
@@ -25,6 +25,7 @@ void BindOptim(py::module& m) {
           .def_readwrite("dyn_num_trials_multiplier",
                          &RANSACOptions::dyn_num_trials_multiplier)
           .def_readwrite("min_num_trials", &RANSACOptions::min_num_trials)
-          .def_readwrite("max_num_trials", &RANSACOptions::max_num_trials);
+          .def_readwrite("max_num_trials", &RANSACOptions::max_num_trials)
+          .def("check", &RANSACOptions::Check);
   MakeDataclass(PyRANSACOptions);
 }
