@@ -6,7 +6,7 @@ void BindObservationManager(py::module& m);
 void BindIncrementalTriangulator(py::module& m);
 void BindIncrementalMapper(py::module& m);
 
-void BindSfMObjects(py::module& m) {
+void BindSfm(py::module& m) {
   BindObservationManager(m);
   BindIncrementalTriangulator(m);
   BindIncrementalMapper(m);
