@@ -1,7 +1,7 @@
-#include "colmap/controllers/incremental_mapper.h"
-
 #include "colmap/sfm/incremental_mapper.h"
 
+#include "colmap/controllers/incremental_pipeline.h"
+
 #include "pycolmap/helpers.h"
 #include "pycolmap/pybind11_extension.h"
 
@@ -11,11 +11,188 @@
 #include <pybind11/stl_bind.h>
 
 using namespace colmap;
+using namespace pybind11::literals;
 namespace py = pybind11;
 
-void BindIncrementalMapperController(py::module& m) {
-  // bind callbacks
-  using CallbackType = IncrementalMapperController::CallbackType;
+void BindIncrementalPipeline(py::module& m) {
+  using Opts = IncrementalPipelineOptions;
+  auto PyOpts = py::classh<Opts>(m, "IncrementalPipelineOptions");
+  PyOpts.def(py::init<>())
+      .def_readwrite(
+          "min_num_matches",
+          &Opts::min_num_matches,
+          "The minimum number of matches for inlier matches to be considered.")
+      .def_readwrite(
+          "ignore_watermarks",
+          &Opts::ignore_watermarks,
+          "Whether to ignore the inlier matches of watermark image pairs.")
+      .def_readwrite("multiple_models",
+                     &Opts::multiple_models,
+                     "Whether to reconstruct multiple sub-models.")
+      .def_readwrite("max_num_models",
+                     &Opts::max_num_models,
+                     "The number of sub-models to reconstruct.")
+      .def_readwrite(
+          "max_model_overlap",
+          &Opts::max_model_overlap,
+          "The maximum number of overlapping images between sub-models. If the "
+          "current sub-models shares more than this number of images with "
+          "another model, then the reconstruction is stopped.")
+      .def_readwrite("min_model_size",
+                     &Opts::min_model_size,
+                     "The minimum number of registered images of a sub-model, "
+                     "otherwise the sub-model is discarded. Note that the "
+                     "first sub-model is always kept independent of size.")
+      .def_readwrite("init_image_id1",
+                     &Opts::init_image_id1,
+                     "The image identifier of the first image used to "
+                     "initialize the reconstruction.")
+      .def_readwrite(
+          "init_image_id2",
+          &Opts::init_image_id2,
+          "The image identifier of the second image used to initialize the "
+          "reconstruction. Determined automatically if left unspecified.")
+      .def_readwrite("init_num_trials",
+                     &Opts::init_num_trials,
+                     "The number of trials to initialize the reconstruction.")
+      .def_readwrite("extract_colors",
+                     &Opts::extract_colors,
+                     "Whether to extract colors for reconstructed points.")
+      .def_readwrite("num_threads",
+                     &Opts::num_threads,
+                     "The number of threads to use during reconstruction.")
+      .def_readwrite("min_focal_length_ratio",
+                     &Opts::min_focal_length_ratio,
+                     "The threshold used to filter and ignore images with "
+                     "degenerate intrinsics.")
+      .def_readwrite("max_focal_length_ratio",
+                     &Opts::max_focal_length_ratio,
+                     "The threshold used to filter and ignore images with "
+                     "degenerate intrinsics.")
+      .def_readwrite("max_extra_param",
+                     &Opts::max_extra_param,
+                     "The threshold used to filter and ignore images with "
+                     "degenerate intrinsics.")
+      .def_readwrite(
+          "ba_refine_focal_length",
+          &Opts::ba_refine_focal_length,
+          "Whether to refine the focal length during the reconstruction.")
+      .def_readwrite(
+          "ba_refine_principal_point",
+          &Opts::ba_refine_principal_point,
+          "Whether to refine the principal point during the reconstruction.")
+      .def_readwrite(
+          "ba_refine_extra_params",
+          &Opts::ba_refine_extra_params,
+          "Whether to refine extra parameters during the reconstruction.")
+      .def_readwrite("ba_refine_sensor_from_rig",
+                     &Opts::ba_refine_sensor_from_rig,
+                     "Whether to refine rig poses during the reconstruction.")
+      .def_readwrite(
+          "ba_min_num_residuals_for_cpu_multi_threading",
+          &Opts::ba_min_num_residuals_for_cpu_multi_threading,
+          "The minimum number of residuals per bundle adjustment problem to "
+          "enable multi-threading solving of the problems.")
+      .def_readwrite(
+          "ba_local_num_images",
+          &Opts::ba_local_num_images,
+          "The number of images to optimize in local bundle adjustment.")
+      .def_readwrite(
+          "ba_local_function_tolerance",
+          &Opts::ba_local_function_tolerance,
+          "Ceres solver function tolerance for local bundle adjustment.")
+      .def_readwrite(
+          "ba_local_max_num_iterations",
+          &Opts::ba_local_max_num_iterations,
+          "The maximum number of local bundle adjustment iterations.")
+      .def_readwrite(
+          "ba_global_frames_ratio",
+          &Opts::ba_global_frames_ratio,
+          "The growth rates after which to perform global bundle adjustment.")
+      .def_readwrite(
+          "ba_global_points_ratio",
+          &Opts::ba_global_points_ratio,
+          "The growth rates after which to perform global bundle adjustment.")
+      .def_readwrite(
+          "ba_global_frames_freq",
+          &Opts::ba_global_frames_freq,
+          "The growth rates after which to perform global bundle adjustment.")
+      .def_readwrite(
+          "ba_global_points_freq",
+          &Opts::ba_global_points_freq,
+          "The growth rates after which to perform global bundle adjustment.")
+      .def_readwrite(
+          "ba_global_function_tolerance",
+          &Opts::ba_global_function_tolerance,
+          "Ceres solver function tolerance for global bundle adjustment.")
+      .def_readwrite(
+          "ba_global_max_num_iterations",
+          &Opts::ba_global_max_num_iterations,
+          "The maximum number of global bundle adjustment iterations.")
+      .def_readwrite(
+          "ba_local_max_refinements",
+          &Opts::ba_local_max_refinements,
+          "The thresholds for iterative bundle adjustment refinements.")
+      .def_readwrite(
+          "ba_local_max_refinement_change",
+          &Opts::ba_local_max_refinement_change,
+          "The thresholds for iterative bundle adjustment refinements.")
+      .def_readwrite(
+          "ba_global_max_refinements",
+          &Opts::ba_global_max_refinements,
+          "The thresholds for iterative bundle adjustment refinements.")
+      .def_readwrite(
+          "ba_global_max_refinement_change",
+          &Opts::ba_global_max_refinement_change,
+          "The thresholds for iterative bundle adjustment refinements.")
+      .def_readwrite("ba_use_gpu",
+                     &IncrementalPipelineOptions::ba_use_gpu,
+                     "Whether to use Ceres' CUDA sparse linear algebra "
+                     "library, if available.")
+      .def_readwrite("ba_gpu_index",
+                     &IncrementalPipelineOptions::ba_gpu_index,
+                     "Index of CUDA GPU to use for BA, if available.")
+      .def_readwrite("use_prior_position",
+                     &Opts::use_prior_position,
+                     "Whether to use priors on the camera positions.")
+      .def_readwrite("use_robust_loss_on_prior_position",
+                     &Opts::use_robust_loss_on_prior_position,
+                     "Whether to use a robust loss on prior camera positions.")
+      .def_readwrite("prior_position_loss_scale",
+                     &Opts::prior_position_loss_scale,
+                     "Threshold on the residual for the robust position prior "
+                     "loss (chi2 for 3DOF at 95% = 7.815).")
+      .def_readwrite("snapshot_path",
+                     &Opts::snapshot_path,
+                     "Path to a folder in which reconstruction snapshots will "
+                     "be saved during incremental reconstruction.")
+      .def_readwrite("snapshot_frames_freq",
+                     &Opts::snapshot_frames_freq,
+                     "Frequency of registered images according to which "
+                     "reconstruction snapshots will be saved.")
+      .def_readwrite(
+          "image_names",
+          &Opts::image_names,
+          "Optional list of image names to reconstruct. If no images are "
+          "specified, all images will be reconstructed by default.")
+      .def_readwrite("fix_existing_frames",
+                     &Opts::fix_existing_frames,
+                     "If reconstruction is provided as input, fix the existing "
+                     "frame poses.")
+      .def_readwrite(
+          "mapper", &Opts::mapper, "Options of the IncrementalMapper.")
+      .def_readwrite("triangulation",
+                     &Opts::triangulation,
+                     "Options of the IncrementalTriangulator.")
+      .def("get_mapper", &Opts::Mapper)
+      .def("get_triangulation", &Opts::Triangulation)
+      .def("get_local_bundle_adjustment", &Opts::LocalBundleAdjustment)
+      .def("get_global_bundle_adjustment", &Opts::GlobalBundleAdjustment)
+      .def("is_initial_pair_provided", &Opts::IsInitialPairProvided)
+      .def("check", &Opts::Check);
+  MakeDataclass(PyOpts);
+
+  using CallbackType = IncrementalPipeline::CallbackType;
   auto PyCallbackType =
       py::enum_<CallbackType>(m, "IncrementalMapperCallback")
           .value("INITIAL_IMAGE_PAIR_REG_CALLBACK",
@@ -26,8 +203,7 @@ void BindIncrementalMapperController(py::module& m) {
                  CallbackType::LAST_IMAGE_REG_CALLBACK);
   AddStringToEnumConstructor(PyCallbackType);
 
-  // bind status
-  using Status = IncrementalMapperController::Status;
+  using Status = IncrementalPipeline::Status;
   auto PyStatus = py::enum_<Status>(m, "IncrementalMapperStatus")
                       .value("NO_INITIAL_PAIR", Status::NO_INITIAL_PAIR)
                       .value("BAD_INITIAL_PAIR", Status::BAD_INITIAL_PAIR)
@@ -35,11 +211,8 @@ void BindIncrementalMapperController(py::module& m) {
                       .value("INTERRUPTED", Status::INTERRUPTED);
   AddStringToEnumConstructor(PyStatus);
 
-  // bind controller
-  py::class_<IncrementalMapperController,
-             std::shared_ptr<IncrementalMapperController>>(
-      m, "IncrementalMapperController")
-      .def(py::init<std::shared_ptr<const IncrementalMapperOptions>,
+  py::classh<IncrementalPipeline>(m, "IncrementalPipeline")
+      .def(py::init<std::shared_ptr<const IncrementalPipelineOptions>,
                     const std::string&,
                     const std::string&,
                     std::shared_ptr<ReconstructionManager>>(),
@@ -47,47 +220,53 @@ void BindIncrementalMapperController(py::module& m) {
            "image_path"_a,
            "database_path"_a,
            "reconstruction_manager"_a)
-      .def_property_readonly("options", &IncrementalMapperController::Options)
-      .def_property_readonly("image_path",
-                             &IncrementalMapperController::ImagePath)
+      .def_property_readonly("options", &IncrementalPipeline::Options)
+      .def_property_readonly("image_path", &IncrementalPipeline::ImagePath)
       .def_property_readonly("database_path",
-                             &IncrementalMapperController::DatabasePath)
-      .def_property_readonly(
-          "reconstruction_manager",
-          &IncrementalMapperController::ReconstructionManager)
+                             &IncrementalPipeline::DatabasePath)
+      .def_property_readonly("reconstruction_manager",
+                             &IncrementalPipeline::ReconstructionManager)
       .def_property_readonly("database_cache",
-                             &IncrementalMapperController::DatabaseCache)
-      .def("add_callback",
-           &IncrementalMapperController::AddCallback,
-           "id"_a,
-           "func"_a)
-      .def("callback", &IncrementalMapperController::Callback, "id"_a)
-      .def("load_database", &IncrementalMapperController::LoadDatabase)
+                             &IncrementalPipeline::DatabaseCache)
+      .def("add_callback", &IncrementalPipeline::AddCallback, "id"_a, "func"_a)
+      .def("callback", &IncrementalPipeline::Callback, "id"_a)
+      .def("load_database", &IncrementalPipeline::LoadDatabase)
       .def("check_run_global_refinement",
-           &IncrementalMapperController::CheckRunGlobalRefinement,
+           &IncrementalPipeline::CheckRunGlobalRefinement,
            "reconstruction"_a,
            "ba_prev_num_reg_images"_a,
            "ba_prev_num_points"_a)
       .def("reconstruct",
-           &IncrementalMapperController::Reconstruct,
-           "mapper_options"_a)
+           &IncrementalPipeline::Reconstruct,
+           "mapper"_a,
+           "mapper_options"_a,
+           "continue_reconstruction"_a)
       .def("reconstruct_sub_model",
-           &IncrementalMapperController::ReconstructSubModel,
-           "core_mapper"_a,
+           &IncrementalPipeline::ReconstructSubModel,
+           "mapper"_a,
            "mapper_options"_a,
            "reconstruction"_a)
       .def("initialize_reconstruction",
-           &IncrementalMapperController::InitializeReconstruction,
-           "core_mapper"_a,
+           &IncrementalPipeline::InitializeReconstruction,
+           "mapper"_a,
            "mapper_options"_a,
            "reconstruction"_a)
-      .def("run", &IncrementalMapperController::Run);
+      .def("run", &IncrementalPipeline::Run);
 }
 
 void BindIncrementalMapperOptions(py::module& m) {
+  using ImageSelection = IncrementalMapper::Options::ImageSelectionMethod;
+  auto PyImageSelectionMethod =
+      py::enum_<ImageSelection>(m, "ImageSelectionMethod")
+          .value("MAX_VISIBLE_POINTS_NUM",
+                 ImageSelection::MAX_VISIBLE_POINTS_NUM)
+          .value("MAX_VISIBLE_POINTS_RATIO",
+                 ImageSelection::MAX_VISIBLE_POINTS_RATIO)
+          .value("MIN_UNCERTAINTY", ImageSelection::MIN_UNCERTAINTY);
+  AddStringToEnumConstructor(PyImageSelectionMethod);
+
   using Opts = IncrementalMapper::Options;
-  auto PyOpts =
-      py::class_<Opts, std::shared_ptr<Opts>>(m, "IncrementalMapperOptions");
+  auto PyOpts = py::classh<Opts>(m, "IncrementalMapperOptions");
   PyOpts.def(py::init<>())
       .def_readwrite("init_min_num_inliers",
                      &Opts::init_min_num_inliers,
@@ -152,36 +331,25 @@ void BindIncrementalMapperOptions(py::module& m) {
       .def_readwrite("max_reg_trials",
                      &Opts::max_reg_trials,
                      "Maximum number of trials to register an image.")
-      .def_readwrite("fix_existing_images",
-                     &Opts::fix_existing_images,
+      .def_readwrite("fix_existing_frames",
+                     &Opts::fix_existing_frames,
                      "If reconstruction is provided as input, fix the existing "
-                     "image poses.")
+                     "frame poses.")
       .def_readwrite("num_threads", &Opts::num_threads, "Number of threads.")
       .def_readwrite("image_selection_method",
                      &Opts::image_selection_method,
-                     "Method to find and select next best image to register.");
+                     "Method to find and select next best image to register.")
+      .def("check", &Opts::Check);
   MakeDataclass(PyOpts);
 }
 
 void BindIncrementalMapperImpl(py::module& m) {
-  // bind options
   BindIncrementalMapperOptions(m);
 
-  // bind ImageSelectionMethod
-  using ImageSelection = IncrementalMapper::Options::ImageSelectionMethod;
-  auto PyImageSelectionMethod =
-      py::enum_<ImageSelection>(m, "ImageSelectionMethod")
-          .value("MAX_VISIBLE_POINTS_NUM",
-                 ImageSelection::MAX_VISIBLE_POINTS_NUM)
-          .value("MAX_VISIBLE_POINTS_RATIO",
-                 ImageSelection::MAX_VISIBLE_POINTS_RATIO)
-          .value("MIN_UNCERTAINTY", ImageSelection::MIN_UNCERTAINTY);
-  AddStringToEnumConstructor(PyImageSelectionMethod);
-
   // bind local bundle adjustment report
   using LocalBAReport = IncrementalMapper::LocalBundleAdjustmentReport;
   auto PyLocalBAReport =
-      py::class_<LocalBAReport>(m, "LocalBundleAdjustmentReport");
+      py::classh<LocalBAReport>(m, "LocalBundleAdjustmentReport");
   PyLocalBAReport.def(py::init<>())
       .def_readwrite("num_merged_observations",
                      &LocalBAReport::num_merged_observations)
@@ -195,9 +363,8 @@ void BindIncrementalMapperImpl(py::module& m) {
 
   // bind incremental mapper
   // TODO: migrate comments. improve formatting
-  py::class_<IncrementalMapper, std::shared_ptr<IncrementalMapper>>(
-      m, "IncrementalMapper")
-      .def(py::init<std::shared_ptr<const DatabaseCache>>())
+  py::classh<IncrementalMapper>(m, "IncrementalMapper")
+      .def(py::init<std::shared_ptr<const DatabaseCache>>(), "database_cache"_a)
       .def("begin_reconstruction",
            &IncrementalMapper::BeginReconstruction,
            "reconstruction"_a)
@@ -209,17 +376,20 @@ void BindIncrementalMapperImpl(py::module& m) {
           [](IncrementalMapper& self,
              const IncrementalMapper::Options& options,
              int image_id1,
-             int image_id2) -> py::object {
+             int image_id2)
+              -> py::typing::Optional<
+                  py::typing::Tuple<py::typing::Tuple<image_t, image_t>,
+                                    Rigid3d>> {
             // Explicitly handle the conversion
             // from -1 (int) to kInvalidImageId (uint32_t).
             image_t image_id1_cast = image_id1;
             image_t image_id2_cast = image_id2;
-            TwoViewGeometry two_view_geometry;
+            Rigid3d cam2_from_cam1;
             const bool success = self.FindInitialImagePair(
-                options, two_view_geometry, image_id1_cast, image_id2_cast);
+                options, image_id1_cast, image_id2_cast, cam2_from_cam1);
             if (success) {
               const auto pair = std::make_pair(image_id1_cast, image_id2_cast);
-              return py::cast(std::make_pair(pair, two_view_geometry));
+              return py::cast(std::make_pair(pair, cam2_from_cam1));
             } else {
               return py::none();
             }
@@ -232,12 +402,12 @@ void BindIncrementalMapperImpl(py::module& m) {
           [](IncrementalMapper& self,
              const IncrementalMapper::Options& options,
              const image_t image_id1,
-             const image_t image_id2) -> py::object {
-            TwoViewGeometry two_view_geometry;
+             const image_t image_id2) -> py::typing::Optional<Rigid3d> {
+            Rigid3d cam2_from_cam1;
             const bool success = self.EstimateInitialTwoViewGeometry(
-                options, two_view_geometry, image_id1, image_id2);
+                options, image_id1, image_id2, cam2_from_cam1);
             if (success)
-              return py::cast(two_view_geometry);
+              return py::cast(cam2_from_cam1);
             else
               return py::none();
           },
@@ -298,17 +468,21 @@ void BindIncrementalMapperImpl(py::module& m) {
            "ba_options"_a,
            "tri_options"_a,
            "normalize_reconstruction"_a = true)
-      .def("filter_images", &IncrementalMapper::FilterImages, "options"_a)
+      .def("filter_frames", &IncrementalMapper::FilterFrames, "options"_a)
       .def("filter_points", &IncrementalMapper::FilterPoints, "options"_a)
       .def_property_readonly("reconstruction",
                              &IncrementalMapper::Reconstruction)
       .def_property_readonly("observation_manager",
                              &IncrementalMapper::ObservationManager)
       .def_property_readonly("triangulator", &IncrementalMapper::Triangulator)
-      .def_property_readonly("filtered_images",
-                             &IncrementalMapper::FilteredImages)
-      .def_property_readonly("existing_image_ids",
-                             &IncrementalMapper::ExistingImageIds)
+      .def_property_readonly("filtered_frames",
+                             &IncrementalMapper::FilteredFrames)
+      .def_property_readonly("existing_frame_ids",
+                             &IncrementalMapper::ExistingFrameIds)
+      .def("reset_initialization_stats",
+           &IncrementalMapper::ResetInitializationStats)
+      .def_property_readonly("num_reg_frames_per_rig",
+                             &IncrementalMapper::NumRegFramesPerRig)
       .def_property_readonly("num_reg_images_per_camera",
                              &IncrementalMapper::NumRegImagesPerCamera)
       .def("num_total_reg_images", &IncrementalMapper::NumTotalRegImages)
@@ -319,6 +493,6 @@ void BindIncrementalMapperImpl(py::module& m) {
 }
 
 void BindIncrementalMapper(py::module& m) {
-  BindIncrementalMapperController(m);
   BindIncrementalMapperImpl(m);
+  BindIncrementalPipeline(m);
 }
